/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.ide.java.ui.deliver;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.TypeNameMatch;
import org.eclipse.jdt.core.search.TypeNameMatchRequestor;
import org.eclipse.jdt.junit.launcher.JUnitLaunchShortcut;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.osgi.util.NLS;
import org.eclipse.pde.core.plugin.PluginRegistry;
import org.eclipse.pde.ui.launcher.JUnitWorkbenchLaunchShortcut;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

import com.ibm.team.filesystem.internal.ide.java.ui.DefinitionsIdeUIPlugin;
import com.ibm.team.process.client.IAdvisorProblemResolutionDelegate;
import com.ibm.team.process.common.advice.IAdvisorInfo;

public class FailedTestsProblemResolutionDelegate implements IAdvisorProblemResolutionDelegate {

	private static void errorDialog(final String msg) {
		Display.getDefault().asyncExec(new Runnable() {
			public void run() {
				Display display = Display.getCurrent();
				if (display == null) {
					return;
				}
				Shell shell = display.getActiveShell();
				if (shell == null) {
					return;
				}
				
				MessageDialog.openError(shell, Messages.FailedTestsProblemResolutionDelegate_ERROR_DIALOG_TITLE, msg);
			}
		});
	}
	
	/**
	 * Searches for and returns a type with the given name in the workspace,
	 * or <code>null</code> if none.
	 * 
	 * @param typeName fully qualified type name
	 * @return type or <code>null</code>
	 * @throws JavaModelException
	 */
	private static IType findTypeInWorkspace(String typeName) throws CoreException {
		int dot= typeName.lastIndexOf('.');
		char[][] qualifications;
		String simpleName;
		if (dot != -1) {
			qualifications= new char[][] { typeName.substring(0, dot).toCharArray() };
			simpleName= typeName.substring(dot + 1);
		} else {
			qualifications= null;
			simpleName= typeName;
		}
		char[][] typeNames= new char[][] { simpleName.toCharArray() };
		
		class ResultException extends RuntimeException {
			private static final long serialVersionUID= 1L;
			private final IType fType;
			public ResultException(IType type) {
				fType= type;
			}
		}
		TypeNameMatchRequestor requestor= new TypeNameMatchRequestor() {
			public void acceptTypeNameMatch(TypeNameMatch match) {
				throw new ResultException(match.getType());
			}
		};
		try {
			new SearchEngine().searchAllTypeNames(qualifications, typeNames, SearchEngine.createWorkspaceScope(), requestor, IJavaSearchConstants.WAIT_UNTIL_READY_TO_SEARCH, null);
		} catch (ResultException e) {
			return e.fType;
		}
		return null;
	}

	public boolean run(IAdvisorInfo problem, IProgressMonitor monitor) {
		// Find the test class
		String className = problem.getData();
		final IType type;
		try {
			type = findTypeInWorkspace(className);
		} catch (CoreException e) {
			errorDialog(NLS.bind(Messages.FailedTestsProblemResolutionDelegate_EXCEPTION_LOCATING_CLASS_NAME, className));
			DefinitionsIdeUIPlugin.getDefault().log(e);
			return false;
		}
		
		if (type == null) {
			errorDialog(NLS.bind(Messages.FailedTestsProblemResolutionDelegate_UNABLE_TO_FIND_TEST_CLASS, className));
			return false;
		}
		
		IResource typeResource = type.getResource();
		if (typeResource == null) {
			errorDialog(NLS.bind(Messages.FailedTestsProblemResolutionDelegate_UNABLE_TO_FIND_RESOURCE_FOR_CLASS, className));
			return false;
		}
		
		IProject project = typeResource.getProject();
		if (project == null) {
			errorDialog(NLS.bind(Messages.FailedTestsProblemResolutionDelegate_UNABLE_TO_FIND_LAUNCHER_FOR_CLASS, className));
			return false;
		}
		
		// Choose the right gender for our launcher
		final JUnitLaunchShortcut shortcut;
		if (null == PluginRegistry.findModel(project)) {
			shortcut = new JUnitLaunchShortcut();
		} else {
			shortcut = new JUnitWorkbenchLaunchShortcut();
		}
		
		Display.getDefault().asyncExec(new Runnable() {
			public void run() {
				shortcut.launch(new StructuredSelection(type), "run"); //$NON-NLS-1$
			}
		});
		
		return true;
	}
	
	/*
	 * @see com.ibm.team.process.client.IAdvisorProblemResolutionDelegate#mayBeApplicable(com.ibm.team.process.common.advice.IAdvisorInfo)
	 */
	public boolean mayBeApplicable(IAdvisorInfo info) {
		return true;
	}
	
}
