/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2009. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.ide.java.ui.deliver;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ibm.team.filesystem.ide.ui.process.AbstractFileAdvisor;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.repository.common.TeamRepositoryException;

/**
 * Advisor which checks that there are no compilation errors in projects
 * affected by delivered change sets.
 */
public class ProhibitCompilationErrorsAdvisor extends AbstractFileAdvisor {

	/**
	 * XML-formatted problem data constants
	 */
	public static final String DATA_TAG_PROJECT = "project"; //$NON-NLS-1$
	public static final String DATA_ATTR_PROJECT_NAME = "name"; //$NON-NLS-1$
	
	public static final String ID_PROHIBIT_COMPILATION_ERRORS_PROBLEM = "com.ibm.team.process.deliver.prohibitCompilationErrorsProblem"; //$NON-NLS-1$
	public static final String ID_PROHIBIT_COMPILATION_ERRORS = "com.ibm.team.process.deliver.prohibitCompilationErrors"; //$NON-NLS-1$
	
	private List fCheckedProjects;
	private List fErrorProjects;
	
	/*
	 * This advisor scans all files for errors and records which projects contain
	 * them. If any errors are found, the advisor will create a single problem
	 * which lists all projects containing errors.
	 * 
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#run(com.ibm.team.process.common.advice.AdvisableOperation, com.ibm.team.process.common.IProcessConfigurationElement, int, com.ibm.team.process.common.advice.IAdvisorInfoCollector, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		fCheckedProjects = new ArrayList();
		fErrorProjects = new ArrayList();
		super.run(operation, advisorConfiguration, collector, monitor); // the super class processes all files, calling getProblem(...)
		if (fErrorProjects.size() > 0) {
			IAdvisorInfo problem = collector.createProblemInfo(Messages.ProhibitCompilationErrorsAdvisor_0, Messages.ProhibitCompilationErrorsAdvisor_1, ID_PROHIBIT_COMPILATION_ERRORS_PROBLEM);
			problem.setProblemObject(fErrorProjects);
			collector.addInfo(problem);
		}
	}
	
	/*
	 * Looks for a compilation error in the given file, but does not return any problems.
	 * Instead, this method records which projects contain errors.
	 * 
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#getProblem(org.eclipse.core.resources.IFile, org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IReportInfo getProblem(IFile file, IAdvisorInfoCollector collector, IProgressMonitor monitor) {
    	IProject project = file.getProject();
    	if (!fCheckedProjects.contains(project)) {
    		fCheckedProjects.add(project);
            try {
                IMarker[] markers = project.findMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
                for (int i = 0; i < markers.length; i++) {
					if (markers[i].getAttribute(IMarker.SEVERITY, IStatus.OK) == IMarker.SEVERITY_ERROR) {
						fErrorProjects.add(project);
						break;
					}
				}
            } catch (CoreException e) {
            	return collector.createExceptionInfo(Messages.ProhibitCompilationErrorsAdvisor_2, e);
            }
    	}
        return null;
	}
	
	/*
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#appendToProblemData(org.w3c.dom.Element, java.lang.Object, org.w3c.dom.Document)
	 */
	protected void appendToProblemData(Element dataElement, Object custom, Document doc) {
	}

}
