/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.ide.java.ui.deliver;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaModelMarker;
import org.eclipse.jdt.core.compiler.IProblem;
import org.eclipse.osgi.util.NLS;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ibm.team.filesystem.ide.ui.process.AbstractFileAdvisor;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.repository.common.TeamRepositoryException;

/**
 * Advisor which checks that there are no non-externalized strings in files
 * being delivered.
 * <p>
 * Note that this is only an approximation: it is implemented by looking for the
 * file in the local Eclipse workspace and checking for existing compilation
 * problems.
 */
public class ProhibitNonexternalizedStringsAdvisor extends AbstractFileAdvisor {
	
	public static final String ID_PROHIBIT_NONEXTERNALIZED_STRINGS_PROBLEM = "com.ibm.team.process.definitions.prohibitNonexternalizedStringsProblem"; //$NON-NLS-1$
	public static final String ID_PROHIBIT_NONEXTERNALIZED_STRINGS = "com.ibm.team.process.deliver.prohibitNonexternalizedStrings"; //$NON-NLS-1$
	
	private List fFilesWithNonexternalizedStrings;
	
	/*
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#run(com.ibm.team.process.common.advice.AdvisableOperation, com.ibm.team.process.common.IProcessConfigurationElement, com.ibm.team.process.common.advice.IAdvisorInfoCollector, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		fFilesWithNonexternalizedStrings = new ArrayList();
		super.run(operation, advisorConfiguration, collector, monitor); // the super class processes all files, calling getProblem(...)
		if (fFilesWithNonexternalizedStrings.size() > 0) {
			IAdvisorInfo problem = collector.createProblemInfo(Messages.ProhibitNonexternalizedStringsAdvisor_0, Messages.ProhibitNonexternalizedStringsAdvisor_1, ID_PROHIBIT_NONEXTERNALIZED_STRINGS_PROBLEM);
			problem.setProblemObject(fFilesWithNonexternalizedStrings);
			collector.addInfo(problem);
		}
	}

	/*
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#getProblem(org.eclipse.core.resources.IFile, com.ibm.team.process.common.advice.IAdvisorInfoCollector, org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IReportInfo getProblem(IFile file, IAdvisorInfoCollector collector, IProgressMonitor monitor) {
        monitor.beginTask(Messages.ProhibitNonexternalizedStringsAdvisor_2, 1);
        try {
            IMarker[] markers = file.findMarkers(IJavaModelMarker.JAVA_MODEL_PROBLEM_MARKER, false, IResource.DEPTH_INFINITE);
            for (int i = 0; i < markers.length; i++) {
                IMarker marker = markers[i];
                int id = marker.getAttribute(IJavaModelMarker.ID, -1);
                if (id == IProblem.NonExternalizedStringLiteral) {
                	fFilesWithNonexternalizedStrings.add(file);
                	break;
                }
            }
            return null;
        } catch (CoreException e) {
        	return collector.createExceptionInfo(NLS.bind(Messages.ProhibitNonexternalizedStringsAdvisor_3, file.getFullPath().toString()), e);
        } finally {
        	monitor.done();
        }
	}

	/*
	 * @see com.ibm.team.process.internal.definitions.ide.ui.deliver.AbstractFileAdvisor#appendToProblemData(org.w3c.dom.Element)
	 */
	protected void appendToProblemData(Element dataElement, Object custom, Document doc) {
	}
	
}
