/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2011. All Rights Reserved.
 *  
 *  Note to U.S. Government Users Restricted Rights:
 *  Use, duplication or disclosure restricted by GSA ADP Schedule
 *  Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.internal.ide.java.ui.deliver;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.jdt.core.ElementChangedEvent;
import org.eclipse.jdt.core.IElementChangedListener;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaElementDelta;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.junit.JUnitCore;
import org.eclipse.jdt.junit.TestRunListener;
import org.eclipse.jdt.junit.model.ITestCaseElement;
import org.eclipse.jdt.junit.model.ITestElement;
import org.eclipse.jdt.junit.model.ITestElementContainer;
import org.eclipse.jdt.junit.model.ITestRunSession;
import org.eclipse.jdt.junit.model.ITestSuiteElement;

public class TestRunManager {
	
	private static class TestRun {
		
		private Map fFailedTests;
		private Set fRunTestClasses;
		
		public void testFinished(String testClassName, String testMethodName, boolean success, boolean overwrite) {
			if (!success) {
				if (fFailedTests == null) {
					fFailedTests = new HashMap();
				}
				fFailedTests.put(testClassName, testMethodName);
			} else if (fFailedTests != null && overwrite) {
				fFailedTests.remove(testClassName);
				if (fFailedTests.isEmpty()) {
					fFailedTests = null;
				}
			} else if (success) {
				if (testMethodName != null) {
					if (fRunTestClasses == null) {
						fRunTestClasses = new HashSet();
					}
					fRunTestClasses.add(testClassName);
				}
			}
		}
		
		public void suiteRan(String suiteName) {
			if (fRunTestClasses == null) {
				return;
			}
			
			fRunTestClasses.add(suiteName);
		}
		
		public Map getFailedTests() {
			return fFailedTests;
		}
		
		public Set getRunTestClasses() {
			return fRunTestClasses;
		}
	}
	
	private class ProcessTestRunListener extends TestRunListener {
		
		private TestRun fCurrentRun;
		private boolean fIgnore;

		public void sessionStarted(ITestRunSession session) {
			fCurrentRun = new TestRun();
			fIgnore = false;
		}
		
		public void sessionFinished(ITestRunSession session) {
			if (!fIgnore) {
				if (fCurrentRun != null) {
					recordSuites(session);
				}
				testsRan(fCurrentRun);
			}
			cleanup();
		}
		
		private void recordSuites(ITestElementContainer ctr) {
			for (ITestElement el : ctr.getChildren()) {
				if (el instanceof ITestSuiteElement) {
					fCurrentRun.suiteRan((((ITestSuiteElement) el).getSuiteTypeName()));
				}
				
				if (el instanceof ITestElementContainer) {
					for (ITestElement child : ((ITestElementContainer)el).getChildren()) {
						if (child instanceof ITestElementContainer) {
							recordSuites((ITestElementContainer)child);
						}
					}
				}
			}
		}

		public void testCaseFinished(ITestCaseElement testCaseElement) {
			if (fCurrentRun != null) {
				ITestElement.Result testResult = testCaseElement.getTestResult(false);
				boolean success = testResult.equals(ITestElement.Result.OK); 
				fCurrentRun.testFinished(testCaseElement.getTestClassName(), testCaseElement.getTestMethodName(), success, false);
			}
		}
		
		public void ignoreCurrentRun() {
			fIgnore = true;
		}
		
		private void cleanup() {
			fCurrentRun = null;
			fIgnore = false;
		}
	}
		
	private class ElementChangedListener implements IElementChangedListener {
		
		public void elementChanged(ElementChangedEvent event) {
			processDelta(event.getDelta());				
		}
		
		private boolean processDelta(IJavaElementDelta delta) {
			int kind = delta.getKind();
			int details = delta.getFlags();
			int type = delta.getElement().getElementType();
			
			switch (type) {
				case IJavaElement.JAVA_MODEL:
				case IJavaElement.JAVA_PROJECT:
				case IJavaElement.PACKAGE_FRAGMENT_ROOT:
				case IJavaElement.PACKAGE_FRAGMENT:
					// If we did some different than changing a child we flush the the undo / redo stack.
					if (kind != IJavaElementDelta.CHANGED || details != IJavaElementDelta.F_CHILDREN) {
						codeChanged();
						return false;
					}
					break;
				case IJavaElement.COMPILATION_UNIT:
					// if we have changed a primary working copy (e.g created, removed, ...)
					// then we do nothing.
					if ((details & IJavaElementDelta.F_PRIMARY_WORKING_COPY) != 0) {
						return true;
					}
					codeChanged();
					return false;
					
				case IJavaElement.CLASS_FILE:
					// Don't examine children of a class file but keep on examining siblings.
					return true;
				default:
					codeChanged();
					return false;	
			}
				
			IJavaElementDelta[] affectedChildren = delta.getAffectedChildren();
			if (affectedChildren == null) {
				return true;
			}
	
			for (int i= 0; i < affectedChildren.length; i++) {
				if (!processDelta(affectedChildren[i])) {
					return false;
				}
			}
			return true;			
		}
	}
	
	private ProcessTestRunListener fTestRunListener = new ProcessTestRunListener();
	private ElementChangedListener fElementChangedListener = new ElementChangedListener();
	private TestRun fRecordedTestRun;
	
	
	public void start() {
		JUnitCore.addTestRunListener(fTestRunListener);
		JavaCore.addElementChangedListener(fElementChangedListener);
	}
	
	public void stop() {
		JUnitCore.removeTestRunListener(fTestRunListener);
		JavaCore.removeElementChangedListener(fElementChangedListener);
	}
	
	public synchronized boolean wereTestsRun(String testClassName) {
		if (fRecordedTestRun != null) {
			Set runTestClasses = fRecordedTestRun.getRunTestClasses();
			if (runTestClasses != null) {
				return testClassName == null ? true : runTestClasses.contains(testClassName);
			}
		}
		return false;
	}
	
	public synchronized Map getFailedTests(String testClassName) {
		if (fRecordedTestRun != null) {
			Map result = new HashMap();
			Map failedTests = fRecordedTestRun.getFailedTests();
			if (failedTests != null) {
				for (Iterator e = failedTests.entrySet().iterator(); e.hasNext();) {
					Map.Entry entry = (Entry) e.next();
					String failedTestClassName = (String) entry.getKey();
					if (testClassName.equalsIgnoreCase(failedTestClassName)) {
						result.put(entry.getKey(), entry.getValue());
					}
				}
				if (!result.isEmpty()) {
					return result;
				}
			}
		}
		return null;
	}
	
	private synchronized void testsRan(TestRun run) {
		fRecordedTestRun = run;			
	}
	
	private synchronized void codeChanged() {
		fRecordedTestRun = null;
		fTestRunListener.ignoreCurrentRun();
	}
}
