/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PlatformUI;

import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.rcp.core.internal.FileSystemResourcesPlugin;
import com.ibm.team.filesystem.rcp.core.internal.changes.IActiveWorkspaceManager;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.IBaselineGroup;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.IComponentSyncContext;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.IComponentSyncModel;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.IFileSystemWorkItem;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.IRemoteActivity;
import com.ibm.team.filesystem.rcp.core.internal.changes.model.ISuspendedRemoteActivity;
import com.ibm.team.filesystem.ui.changes.actions.DeliverAction;
import com.ibm.team.filesystem.ui.changes.actions.UnSuspendAction;
import com.ibm.team.filesystem.ui.item.AdapterUtil;
import com.ibm.team.filesystem.ui.wrapper.ChangeSetInContextWrapper;
import com.ibm.team.internal.filesystem.ui.views.history.actions.SuspendWrapperAction;
import com.ibm.team.links.common.ILink;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.TeamPlatform;
import com.ibm.team.repository.client.ITeamRepository.ILoginParticipant;
import com.ibm.team.repository.client.util.IEvent;
import com.ibm.team.repository.client.util.IListener;
import com.ibm.team.repository.client.util.PropertyChangeEvent;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.ui.internal.utils.JFaceUtils;
import com.ibm.team.repository.rcp.ui.utils.UIContext;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.client.events.IChangeHistoryEvent;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.internal.util.ItemId;
import com.ibm.team.scm.common.internal.util.ItemLists;
import com.ibm.team.scm.common.internal.util.NewCollection;
import com.ibm.team.workitem.client.WorkflowUtilities;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;
import com.ibm.team.workitem.rcp.core.ClientModel;
import com.ibm.team.workitem.rcp.core.activation.WorkItemActivationEvent;
import com.ibm.team.workitem.rcp.core.activation.WorkItemActivationParticipant;

/**
 * This class manages the work item to change set related workflows. More specifically 
 * it handles:
 * <ul>
 *   <li>Ensures that new change sets are auto-linked to the active work item.
 *   <li>Prompts to suspend/resume when moving between tasks.
 *   <li>Ensures that work item state changes may cause changes to be suspended or delivered.
 * </ul>
 * @since 0.6
 */   
public class WorkItemActivationListener extends WorkItemActivationParticipant implements IListener, ILoginParticipant {

	// Do we want to only link new change sets in workspaces/component tuples which don't already have a linked change set to the same work item?
	private static boolean onlyLinkFirstNewChangeSet = false;
	
	// Do we want to create a work item link if other work item links already exist (for un-suspended change sets)
	private static boolean allowLinkCreationIfOthersExist = false;
	
	// Repositories and workspaces being tracked
	private Map<ITeamRepository, Set<ItemId<IWorkspace>>> trackedRepositories = 
		new HashMap<ITeamRepository, Set<ItemId<IWorkspace>>>();
	
	// Tracking repo/workspaces which weren't completely tracked because it wasn't possible to
	// to resolve the workspace handle at startup because the repository wasn't logged-in. Keep a 
	// queue for when the repository is logged in again.
	private Map<ITeamRepository, Set<ItemId<IWorkspace>>> pendingTrackedRepositories = 
		new HashMap<ITeamRepository, Set<ItemId<IWorkspace>>>();
	
	private final Object mutex = new Object();
	
	private final boolean TRACE = false;
	

	WorkItemActivationListener() {		
		initialize();
	}
	
	protected void initialize() {
		Job job = new Job("WorkItemActivationListener (system)") { //$NON-NLS-1$
			@Override
			protected IStatus run(IProgressMonitor monitor) {
				// Mutual Exclusion - We want to ensure all workspaces have a chance to "startListening" 
				// before we allow the handleLogin() method to proceed.
				synchronized(mutex) {
					IActiveWorkspaceManager manager = FileSystemResourcesPlugin.getActiveWorkspaceManager();        
					ClientModel.getWorkItemActivationManager().addActivationParticipant(WorkItemActivationListener.this);
					ITeamRepository[] teamRepositories = TeamPlatform.getTeamRepositoryService().getTeamRepositories();
					// Moved the LoginParticipant to the top because previously 
					// it was possible to have the connection established before the listener was set in place
					for (ITeamRepository repo : teamRepositories) {
						repo.addLoginParticipant(WorkItemActivationListener.this);
					}
					manager.addGenericListener(IActiveWorkspaceManager.ACTIVE_WORKSPACES, WorkItemActivationListener.this);
					IWorkspaceHandle[] contexts = manager.getActiveWorkspaces();
					SubMonitor progress = SubMonitor.convert(monitor, contexts.length);
					for (int i = 0; i < contexts.length; i++) {
						IWorkspaceHandle wh = contexts[i];
						startListening(wh, pendingTrackedRepositories, progress.newChild(1));
					}

				}
				return Status.OK_STATUS;
			}
		};
		job.setUser(false);
		job.setSystem(true);
		job.schedule();	
	}
	
	@Override
	public IStatus handleActivationEvent(WorkItemActivationEvent event,
			IProgressMonitor monitor) {
		int type = event.getType();
		IWorkItemHandle wi = event.getWorkItem();
		if(type == WorkItemActivationEvent.WORKITEM_ACTIVATE) {
			workItemActivated(wi, monitor);
		} else if(type == WorkItemActivationEvent.WORKITEM_DEACTIVATE) {
			workItemDeactivated(wi, monitor);
		}
		
		return super.handleActivationEvent(event, monitor);
	}

	private void workItemDeactivated(IWorkItemHandle wi, IProgressMonitor monitor) {
		IComponentSyncModel syncModel = FileSystemResourcesPlugin.getComponentSyncModel();
		List<IRemoteActivity> outgoing = getAllActivitiesFor(syncModel, monitor);
		List<IRemoteActivity> changeSets = new ArrayList<IRemoteActivity>();
		for (IRemoteActivity remoteActivity : outgoing) {
			List<IFileSystemWorkItem> links = remoteActivity.getWorkItems();
			for (IFileSystemWorkItem link : links) {
				Object obj = link.getLink().getTargetRef().resolve();
				if(obj instanceof IWorkItemHandle) {					
					if(wi.sameItemId((IItemHandle) obj)) {
						changeSets.add(remoteActivity);
					}
				}
			}
		}
		if(changeSets.size() > 0) {
			boolean resolved = false;
			ITeamRepository repo = (ITeamRepository)wi.getOrigin();
			IWorkItem workItem = null;
			try {
				workItem = (IWorkItem)repo.itemManager().fetchCompleteItem(wi, IItemManager.DEFAULT, monitor);
			} catch (TeamRepositoryException e) {
				return; // Silent return
			}
			final IWorkflowInfo info= WorkflowUtilities.findCachedWorkflowInfo(workItem);
			if (info != null && info.getStateGroup(workItem.getState2()) == IWorkflowInfo.CLOSED_STATES) {
				resolved = true;
			}
			if(resolved) {
				deliverChanges(changeSets, monitor);
			} else {			
				suspendChanges(changeSets);
			}
		}
	}

	private void suspendChanges(List<IRemoteActivity> toSuspend) {
		if(JFaceUtils.showMessageBlocking(
			Messages.WorkItemActivationListener_2, 
			Messages.WorkItemActivationListener_3, 
			new String[] { IDialogConstants.YES_LABEL, IDialogConstants.NO_LABEL }, 
			MessageDialog.WARNING)) {
			IWorkbenchPage page = JFaceUtils.getWorkbenchPage();
		    if (page != null) {
		        final Shell shell = page.getWorkbenchWindow().getShell();
		        final List<ChangeSetInContextWrapper> changeSets = AdapterUtil.adaptList(toSuspend, ChangeSetInContextWrapper.class);
		        SuspendWrapperAction.doSuspend(UIContext.createPageContext(shell.getDisplay(), shell, page), changeSets);
		    }
		}
	}

	private void workItemActivated(IWorkItemHandle wi, IProgressMonitor monitor) {
		IComponentSyncModel syncModel = FileSystemResourcesPlugin.getComponentSyncModel();
		List<ISuspendedRemoteActivity> suspended = getSuspendedActivitiesFor(syncModel);
		List<IRemoteActivity> outgoing = getAllActivitiesFor(syncModel, monitor);
		ArrayList<ISuspendedRemoteActivity> toResume = new ArrayList<ISuspendedRemoteActivity>();
		ArrayList<IRemoteActivity> toLink = new ArrayList<IRemoteActivity>();
		for (ISuspendedRemoteActivity remoteActivity : suspended) {
			List<IFileSystemWorkItem> links = remoteActivity.getWorkItems();
			for (IFileSystemWorkItem link : links) {
				Object obj = link.getLink().getTargetRef().resolve();
				if (obj instanceof IWorkItemHandle) {
					if (wi.sameItemId((IItemHandle) obj)) {
						toResume.add(remoteActivity);
					}
				}
			}
		}
		for (IRemoteActivity activity : outgoing) {
			List<IFileSystemWorkItem> links = activity.getWorkItems();
			if(links.isEmpty())
				toLink.add(activity);
		}
		
		if(toResume.size() > 0) {
			if(JFaceUtils.showMessageBlocking(
					Messages.WorkItemActivationListener_6, 
					Messages.WorkItemActivationListener_7, 
					new String[] { IDialogConstants.YES_LABEL, IDialogConstants.NO_LABEL },
					MessageDialog.WARNING)) {
				IWorkbenchPage page = JFaceUtils.getWorkbenchPage();
		        if (page != null) {
		        	IWorkbench workbench = PlatformUI.getWorkbench();
		        	if (workbench != null) {
			            UIContext context = UIContext.createPageContext(workbench.getDisplay(), 
			            		page.getWorkbenchWindow().getShell(), page, Messages.WorkItemActivationListener_11);
			            UnSuspendAction.resumeRemoteActivity(context, toResume);
		        	}
		        }
			}
		}
	}

	/**
	 * Detect when a link from a change set to a work item is created and activate
	 * the work item if needed.
	 */
	public void handleEvents(List events) {
		final List<IWorkspaceHandle> startListening = new ArrayList<IWorkspaceHandle>();
		final List<IWorkspaceHandle> stopListening = new ArrayList<IWorkspaceHandle>();
		final List<IChangeHistoryEvent> linkChangeSet = new ArrayList<IChangeHistoryEvent>();
		for (Iterator it = events.iterator(); it.hasNext();) {
			IEvent event = (IEvent) it.next();
			if(event instanceof PropertyChangeEvent) {
				PropertyChangeEvent propEvent = (PropertyChangeEvent)event;
				if(propEvent.getProperty().equals(IActiveWorkspaceManager.ACTIVE_WORKSPACES)) {
					IWorkspaceHandle beforeContext = (IWorkspaceHandle) propEvent.getOldValue();
					IWorkspaceHandle afterContext = (IWorkspaceHandle) propEvent.getNewValue();
					if(beforeContext == null && afterContext != null) {
						startListening.add(afterContext); // workspace was added					
					} if(beforeContext != null && afterContext == null) {
						stopListening.add(beforeContext); // workspace was removed	
					}
				}
			} else if(event.getEventType() == IChangeHistoryEvent.CREATE_CHANGE_SET) {
				IChangeHistoryEvent chEvent = (IChangeHistoryEvent) event;
				linkChangeSet.add(chEvent);
			}
		}
		if(startListening.size() > 0 || stopListening.size() > 0 || linkChangeSet.size() > 0) {
			Job job = new Job("") { //$NON-NLS-1$
				@Override
				protected IStatus run(IProgressMonitor monitor) {
					try {
						for (IChangeHistoryEvent event : linkChangeSet) {
							IWorkItemHandle wih = ClientModel.getWorkItemActivationManager().getActiveWorkItem();
							if(wih != null) {
								ITeamRepository repo = (ITeamRepository) wih.getOrigin();
								IFileSystemWorkItemManager workItemManager = (IFileSystemWorkItemManager) repo.getClientLibrary(IFileSystemWorkItemManager.class);
								IWorkspaceConnection connection = (IWorkspaceConnection) event.getEventSource();
		                        List<IChangeSetHandle> chHandles = event.getChangeSets();		                        
		                        List<ItemId<IChangeSet>> changeSets = NewCollection.arrayList(chHandles.size());
		                        for (IChangeSetHandle changeSetHandle : chHandles) {
									changeSets.add(new ItemId<IChangeSet>(changeSetHandle));
								}
		                        List<IChangeSetHandle> changeSetsToBeLinked = getNeedsToBeLinked(
		                        		FileSystemResourcesPlugin.getComponentSyncModel(), 
		                        		wih,
		                        		changeSets, monitor);								
		                        for (IChangeSetHandle handle : changeSetsToBeLinked) {									
	                        		workItemManager.createLink(
	                        				connection.getResolvedWorkspace(), 
	                        				handle, new IWorkItemHandle[] {wih}, monitor);
		                        }
							}
						}
						synchronized (mutex) { //Run either before or after handleLogin()
							for (IWorkspaceHandle workspaceHandle : startListening) {
								startListening(workspaceHandle, pendingTrackedRepositories, monitor);
							}
							
							for (IWorkspaceHandle workspaceHandle : stopListening) {
								stopListening(workspaceHandle, true);
							}
						}
					} catch(TeamRepositoryException e) {
						
					}
					return Status.OK_STATUS;
				}
			};
			job.schedule();
		}
	}

	private void stopListening(IWorkspaceHandle workspaceHandle, boolean contactServer) {
		ITeamRepository repo = (ITeamRepository) workspaceHandle.getOrigin();
		Set<ItemId<IWorkspace>> trackedWorkspaces = trackedRepositories.get(repo);
		if(trackedWorkspaces != null) {
    		ItemId workspace = new ItemId(workspaceHandle);
    		trackedWorkspaces.remove(workspace);
    		// Connection scoped, always removed    	
    		IWorkspaceConnection connection = getConnectionFor(workspaceHandle, repo, contactServer, null);
			// The connection has already been garbage collected, no need to remove the listener
			if(connection != null) {
				connection.removeGenericListener(IChangeHistoryEvent.CREATE_CHANGE_SET, this);
				TRACE("stopped listening to workspace " + workspace.getItemUUID().toString() + " on connection " + connection.getName()); //$NON-NLS-1$ //$NON-NLS-2$
			}
			// Repository scoped, only removed when all workspace in repo are no longer tracked
    		if(trackedWorkspaces.isEmpty()) {
    			trackedRepositories.remove(repo);
    		}
		}
	}

	/**
	 * The connection scoped listener has to be registered once per workspace.
	 */
	private void startListening(IWorkspaceHandle workspaceHandle, Map<ITeamRepository, Set<ItemId<IWorkspace>>> pendingWorkspaces, IProgressMonitor monitor) {
		ITeamRepository repo = (ITeamRepository) workspaceHandle.getOrigin();
		Set<ItemId<IWorkspace>> trackedWorkspaces = trackedRepositories.get(repo);
		
		if(trackedWorkspaces == null) {         
        	trackedWorkspaces = new HashSet<ItemId<IWorkspace>>();
			trackedRepositories.put(repo, trackedWorkspaces);
		}
		ItemId workspace = new ItemId(workspaceHandle);
	
		if(! trackedWorkspaces.contains(workspace)) {
			IWorkspaceConnection connection = getConnectionFor(workspaceHandle, repo, true, monitor);
			// not known and can't login just yet
			if(connection == null) {
				Set<ItemId<IWorkspace>> pending = pendingWorkspaces.get(repo);
				if(pending == null) {
					pending = new HashSet<ItemId<IWorkspace>>();
					pendingWorkspaces.put(repo, pending);
				}
				if (!pending.contains(workspace)){
					TRACE("added to pending listening set for workspace " + workspace.getItemUUID().toString()); //$NON-NLS-1$
					pending.add(workspace);
				}
			} else {			
				connection.addGenericListener(IChangeHistoryEvent.CREATE_CHANGE_SET, this);
				trackedWorkspaces.add(workspace);
				TRACE("started listening to workspace " + workspace.getItemUUID().toString() + " on connection " + connection.getName()); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}		
	}
	
	private void deliverChanges(final List<IRemoteActivity> outgoing, IProgressMonitor monitor) {
		if(JFaceUtils.showMessageBlocking(
				Messages.WorkItemActivationListener_18, 
				Messages.WorkItemActivationListener_19, 
				new String[] { IDialogConstants.YES_LABEL, IDialogConstants.NO_LABEL },
				MessageDialog.WARNING)) {
			IWorkbenchPage page = JFaceUtils.getWorkbenchPage();
		    if (page != null) {
		        final Shell shell = page.getWorkbenchWindow().getShell();
				JFaceUtils.syncExec(new Runnable() {
					public void run() {
						DeliverAction deliverAction = new DeliverAction();
						deliverAction.deliver(shell, outgoing);	
					}
				});
			}
		}
	}

	private IWorkspaceConnection getConnectionFor(
			IWorkspaceHandle workspaceHandle, ITeamRepository repo, boolean contactServer, IProgressMonitor monitor) {
		List<IWorkspaceConnection> knownConnections = SCMPlatform.getWorkspaceManager(repo).
			knownWorkspaceConnections();
		IWorkspaceConnection connection = null;
		for (IWorkspaceConnection workspaceConnection : knownConnections) {
			if(workspaceConnection.getResolvedWorkspace().sameItemId(workspaceHandle)) {
				connection = workspaceConnection;
				TRACE("Connection found through knownWorkspaceConnections()"); //$NON-NLS-1$
			}
		}	
		if(repo.loggedIn()) {
			TRACE("Repo Logged In"); //$NON-NLS-1$
			try {
				if(contactServer) {
					connection = SCMPlatform.getWorkspaceManager(repo).getWorkspaceConnection(workspaceHandle, monitor);
					if (connection != null) {
						TRACE("REPO logged in and connection found"); //$NON-NLS-1$
					}
				}
			} catch(TeamRepositoryException e) {
				// don't log, just don't return a connection
				TRACE("REPO logged in and TeamRepositoryException exception caught"); //$NON-NLS-1$
				connection = null;
			}
		}
		if (connection == null) {
			TRACE("Connection is null"); //$NON-NLS-1$
		}
		return connection;
	}
	
	public static List<IRemoteActivity> getAllActivitiesFor(IComponentSyncModel model, IProgressMonitor monitor) {
		try {
			model.refresh(true, monitor);
		} catch(TeamRepositoryException e) {
		}
        List<IRemoteActivity> outgoingActivities = new ArrayList();
        IComponentSyncContext[] contexts = model.getComponentSyncContexts();
        for (int i = 0; i < contexts.length; i++) {
            IComponentSyncContext context = contexts[i];
            // Collect all the remote activities known to the model, including the ones inside baseline groups
            outgoingActivities.addAll(context.getOutgoingActivitySource().getActivities());
            for (IBaselineGroup bg : (List<IBaselineGroup>) context.getOutgoingActivitySource().getBaselines())
                outgoingActivities.addAll(bg.getActivities());            
        }
        return outgoingActivities;
    }
	
	public static List<ISuspendedRemoteActivity> getSuspendedActivitiesFor(IComponentSyncModel model) {
        List<ISuspendedRemoteActivity> outgoingActivities = new ArrayList();
        IComponentSyncContext[] contexts = model.getComponentSyncContexts();
        for (int i = 0; i < contexts.length; i++) {
            IComponentSyncContext context = contexts[i];
            // Collect all the remote activities known to the model, including the ones inside baseline groups
            outgoingActivities.addAll(context.getSuspendedActivitySource().getActivities());
        }
        return outgoingActivities;
    }
	
	public static List<IRemoteActivity> getOutgoingFor(IComponentSyncModel model, IWorkItemHandle wi, IProgressMonitor monitor) {
		if(wi == null)
			return Collections.EMPTY_LIST;
		List<IRemoteActivity> outgoing = getAllActivitiesFor(model, monitor);
		ArrayList<IRemoteActivity> outgoingRelated = new ArrayList<IRemoteActivity>();
		for (IRemoteActivity remoteActivity : outgoing) {
			List<IFileSystemWorkItem> links = remoteActivity.getWorkItems();
			for (IFileSystemWorkItem link : links) {
				// link can be null if the model is not completely resolved yet
				ILink linkObj = link.getLink();
				if(linkObj != null) {
					Object obj = linkObj.getTargetRef().resolve();
					if (obj != null && obj instanceof IWorkItemHandle) {
						if (wi.sameItemId((IItemHandle) obj)) {
							outgoingRelated.add(remoteActivity);
						}
					}
				}
			}
		}
		return outgoingRelated;
	}
	
	public static List<IChangeSetHandle> getNeedsToBeLinked(IComponentSyncModel model, IWorkItemHandle wi, List<ItemId<IChangeSet>> changeSets, IProgressMonitor monitor) {
        if(wi == null) {
            return Collections.EMPTY_LIST;
        }
        List<IRemoteActivity> outgoing = getAllActivitiesFor(model, monitor);
        List<IRemoteActivity> outgoingNewRemoteActivities = new ArrayList<IRemoteActivity>();
        ArrayList<IWorkspace> outgoingRelatedWorkspaces = new ArrayList<IWorkspace>();
        ArrayList<IComponent> outgoingRelatedComponents = new ArrayList<IComponent>();
        // find the workspaces with change sets already linked to the given work item
        for (IRemoteActivity remoteActivity : outgoing) {
        	ItemId<IChangeSet> changeSetInCSM = ItemId.create(remoteActivity.getChangeSetHandle());
			if(changeSets.contains(changeSetInCSM)) {
        		outgoingNewRemoteActivities.add(remoteActivity);
        		changeSets.remove(changeSetInCSM);
        	}
			boolean isComplete = false;
			IChangeSet fullChangeSet = remoteActivity.getChangeSet();
			if(fullChangeSet != null) {
				isComplete = fullChangeSet.isActive();
			}
            List<IFileSystemWorkItem> links = remoteActivity.getWorkItems();
            for (IFileSystemWorkItem link : links) {
                // link can be null if the model is not completely resolved yet
                ILink linkObj = link.getLink();
                if(linkObj != null) {
                    Object obj = linkObj.getTargetRef().resolve();
                    if (obj != null && obj instanceof IWorkItemHandle) {
                        if (wi.sameItemId((IItemHandle) obj) && isComplete) {
                        	IWorkspace workspace = remoteActivity.getActivitySource().getWorkspace();
                        	if (workspace != null)
                        		outgoingRelatedWorkspaces.add(workspace);
                        	outgoingRelatedComponents.add(remoteActivity.getActivitySource().getModel().getComponent());
                        }
                    }
                }
            }
        }
        
        for (Iterator it = outgoingNewRemoteActivities.iterator(); it
				.hasNext();) {
			IRemoteActivity remoteActivity = (IRemoteActivity) it.next();
			if(onlyLinkFirstNewChangeSet && // only link the change sets in workspaces/component tuples which don't already have a linked change set
			   outgoingRelatedWorkspaces.contains(remoteActivity.getActivitySource().getWorkspace()) &&
			   outgoingRelatedComponents.contains(remoteActivity.getActivitySource().getModel().getComponent())) {
				it.remove();
			} else {
				List<IFileSystemWorkItem> linkedWorkItems = remoteActivity.getWorkItems();
				if (!allowLinkCreationIfOthersExist && linkedWorkItems != null && linkedWorkItems.size() > 0) { 
					// We don't want to link to a work item if it is already linked to other work items
					it.remove();
				} else {
					ItemId<IChangeSet> changeSetInCSM = ItemId.create(remoteActivity.getChangeSetHandle());
					changeSets.add(changeSetInCSM);
				}
			}
		} 
        return ItemLists.idsToHandles(changeSets);
    }
	
	public void handleLogin(final ITeamRepository repository) {
		Job job = new Job("") { //$NON-NLS-1$
			@Override
			protected IStatus run(IProgressMonitor monitor) {
				synchronized (mutex) { // Run after the initialize() job
					if (pendingTrackedRepositories.size() > 0) {
						TRACE("handling login for pendings "); //$NON-NLS-1$
						Map<ITeamRepository, Set<ItemId<IWorkspace>>> failedPendingWorkspaces = new HashMap<ITeamRepository, Set<ItemId<IWorkspace>>>();
						for (ITeamRepository repo : pendingTrackedRepositories.keySet()) {
							for (ItemId<IWorkspace> workspaceId : pendingTrackedRepositories.get(repo)) {
								IWorkspaceHandle handle = workspaceId.toHandle(repo);
								startListening(handle, failedPendingWorkspaces, monitor);
							}
						}
						pendingTrackedRepositories.clear();
						if (failedPendingWorkspaces.size() > 0) {
							pendingTrackedRepositories = failedPendingWorkspaces;
							// short circuit - don't remove listener and try again next time
							return Status.OK_STATUS;
						}
					}
					repository.removeLoginParticipant(WorkItemActivationListener.this);
					return Status.OK_STATUS;
				}
			}
			};
			job.setUser(false);
			job.setSystem(true);
			job.schedule();
	}

	public void handleLogout(ITeamRepository repository) {
		// do nothing
	}
	
	public void dispose() {
		IActiveWorkspaceManager manager = FileSystemResourcesPlugin.getActiveWorkspaceManager();
		manager.removeGenericListener(IActiveWorkspaceManager.ACTIVE_WORKSPACES, this);
		ClientModel.getWorkItemActivationManager().removeActivationParticipant(this);
		IWorkspaceHandle[] contexts = manager.getActiveWorkspaces();
		for (int i = 0; i < contexts.length; i++) {
			IWorkspaceHandle workspaceSyncContext = contexts[i];
			stopListening(workspaceSyncContext, false);
		}
	}
	
	private void TRACE(String message) {
		if(TRACE)
			System.out.println("WorkItemActivationListener: " + message); //$NON-NLS-1$
	}
}
