/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems.actions;

import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchPage;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.snapshot.SnapshotId;
import com.ibm.team.filesystem.client.internal.snapshot.WorkspaceId;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.internal.rcp.ui.workitems.Messages;
import com.ibm.team.filesystem.ui.UiPlugin;
import com.ibm.team.filesystem.ui.item.AdapterUtil;
import com.ibm.team.filesystem.ui.wrapper.ChangeSetInContextWrapper;
import com.ibm.team.filesystem.ui.wrapper.ChangeSetWrapper;
import com.ibm.team.internal.filesystem.ui.util.RepositoryUtils;
import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.rcp.ui.teamnavigator.ConnectedProjectAreaRegistry;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.ui.operations.IStatusCollector;
import com.ibm.team.repository.rcp.ui.operations.OperationFailedException;
import com.ibm.team.repository.rcp.ui.operations.RepositoryOperation;
import com.ibm.team.repository.rcp.ui.parts.impl.MessageDialogFactory;
import com.ibm.team.repository.rcp.ui.utils.AbstractActionDelegate;
import com.ibm.team.repository.rcp.ui.utils.UIContext;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IFlowEntry;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.internal.util.NewCollection;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.rcp.ui.internal.ConnectedProjectAreas;
import com.ibm.team.workitem.rcp.ui.workitempicker.WorkItemSelectionDialog;

public class AssociateWorkItemToChangeSetAction extends AbstractActionDelegate {

    @Override
    public void run(Shell shell, IWorkbenchPage page, IStructuredSelection selection) {
        final Set<ChangeSetWrapper> changeSetWrappers = NewCollection.hashSet();
        changeSetWrappers.addAll(AdapterUtil.adaptList(selection.toList(), ChangeSetWrapper.class));

        ITeamRepository repository = null;
        /**
         * Figure out if all the change sets know about and agree upon which team area they're in.
         * If so, then use that workspace for picking work items.
         */
        SnapshotId workspace = null;
        for (ChangeSetWrapper wrapper : changeSetWrappers) {
        	if (repository == null) {
        		repository = wrapper.getRepository();
        	} else {
        		if (repository != wrapper.getRepository()) {
        			MessageDialogFactory.showMessage(shell, IStatus.ERROR, Messages.AssociateWorkItemToChangeSetAction_0, Messages.AssociateWorkItemToChangeSetAction_1);
        			return;
        		}
        	}
        	
        	if (wrapper instanceof ChangeSetInContextWrapper) {
        		ChangeSetInContextWrapper inContext = (ChangeSetInContextWrapper)wrapper;
        		
        		SnapshotId nextWorkspace = inContext.getSnapshot();
        		if (workspace == null) {
        			workspace = nextWorkspace;
        		} else {
        			if (!workspace.equals(nextWorkspace)) {
        				workspace = null;
        				break;
        			}
        		}
        	}
        }

        final SnapshotId finalWorkspace = workspace;
        final UIContext context = getContext();
        final ITeamRepository server = repository;
        context.getUserOperationRunner().enqueue(Messages.AssociateWorkItemToChangeSetAction_2, new RepositoryOperation() {
            public void repositoryRun(IProgressMonitor monitor, IStatusCollector problems) throws TeamRepositoryException, FileSystemException {
            	SubMonitor progress = SubMonitor.convert(monitor, 100);
				IProjectAreaHandle projectAreaHandle = null;
				IWorkspace workspace = null;
				
				// This detection code only works with workspaces. For change sets associated with a baseline, snapshot,
				// or era, we'll fall back to using no project area for scoping. This handles 99.99% of cases <g>
				if (finalWorkspace != null && (finalWorkspace instanceof WorkspaceId)) {
					IWorkspaceConnection workspaceConnection = (IWorkspaceConnection) finalWorkspace.getConnection(progress.newChild(30));
					workspace = workspaceConnection.getResolvedWorkspace();
					
					IFlowEntry currentDeliverFlow = workspaceConnection.getFlowTable().getCurrentDeliverFlow();
					ITeamRepository repo = workspaceConnection.teamRepository();
					
					if(currentDeliverFlow != null) {
						if (currentDeliverFlow.getRemoteRepositoryIdentifier() == null) {
							
							IWorkspaceConnection streamConnection = SCMPlatform.getWorkspaceManager(repo).getWorkspaceConnection((IWorkspaceHandle)currentDeliverFlow.getFlowNode(), progress.newChild(10));
							IProcessAreaHandle processAreaHandle = streamConnection.getProcessArea(progress.newChild(30));
							
							if (processAreaHandle != null) {
								IProcessArea area = (IProcessArea) streamConnection.teamRepository().itemManager().fetchCompleteItem(processAreaHandle, IItemManager.DEFAULT,
										progress.newChild(30));
								projectAreaHandle = area.getProjectArea();
							}
						}
					}
					
					if(UiPlugin.getBooleanPreference(UiPlugin.USE_CURRENT_PROJECTAREA_FORWORKITEMPROMPTS) || projectAreaHandle == null) {
						IProjectAreaHandle defaultProject = ConnectedProjectAreas.getInstance().getDefaultProjectAreaFallback(repo, true);
						if(defaultProject != null) {
							projectAreaHandle = defaultProject;
						}
					}
				}
				
				if (projectAreaHandle == null) {
					List connectedProjectAreas = ConnectedProjectAreaRegistry.getDefault().getConnectedProjectAreas(server);
					if (connectedProjectAreas.size() == 0) {
						context.getDisplay().asyncExec(new Runnable() {
							public void run() {
								MessageDialogFactory.showMessage(context.getShell(), IStatus.ERROR, Messages.AssociateWorkItemToChangeSetAction_0, NLS.bind(Messages.AssociateWorkItemToChangeSetAction_5, RepositoryUtils.getLabel(server)));
							}
						});
						return;
					} else {
						projectAreaHandle = (IProjectAreaHandle) connectedProjectAreas.get(0);
					}
				}
            	
				final IProjectAreaHandle finalProjectArea = projectAreaHandle;
				final IWorkspace finalWorkspace = workspace;
            	context.getDisplay().asyncExec(new Runnable() {
					public void run() {
				        final IWorkItemHandle[] workItems = AssociateWorkItemToChangeSetAction.openWorkItemPicker(context.getShell(), context.getPage(), server, finalProjectArea);
				        if (workItems.length > 0) {
				        	context.getUserOperationRunner().enqueue(Messages.AssociateWorkItemToChangeSetAction_3, new RepositoryOperation() {
			                    public void repositoryRun(IProgressMonitor monitor, IStatusCollector problems) throws TeamRepositoryException, FileSystemException,
			                            OperationFailedException {                        
			                    	for (ChangeSetWrapper next : changeSetWrappers) {
				                        IChangeSetHandle changeSet = next.getChangeSet();
				                        IFileSystemWorkItemManager workItemManager = (IFileSystemWorkItemManager) server.getClientLibrary(IFileSystemWorkItemManager.class);
				                        workItemManager.createLink(
				                        		finalWorkspace, changeSet, workItems, monitor);
			                    	}
			                    }
				        	});
				        }
					}
            	});
            }
        });        
    }

    /**
     * User can select work items and click ok, to create a new activity with
     * the selected work items. Or user can select cancel and a new activity is
     * created without work items.
     * 
     * @param projectAreaHandle
     */
    public static IWorkItemHandle[] openWorkItemPicker(Shell parent, IWorkbenchPage page, ITeamRepository server, IProjectAreaHandle projectAreaHandle) {
        WorkItemSelectionDialog dialog = null;
        if (projectAreaHandle == null) {
            dialog = new WorkItemSelectionDialog(parent, server, WorkItemSelectionDialog.MULTIPLE_SELECTION);
        } else {
            dialog = new WorkItemSelectionDialog(parent, projectAreaHandle, WorkItemSelectionDialog.MULTIPLE_SELECTION);
        }
        dialog.setTitle(Messages.AssociateWorkItemToChangeSetAction_4);
        if (Window.OK == dialog.open())
            return (IWorkItemHandle[]) dialog.getResult();
        return new IWorkItemHandle[0];
    }

    /**
     * User can select work items and click ok, to create a new activity with
     * the selected work items. Or user can select cancel and a new activity is
     * created without work items.
     */
    public static IWorkItemHandle[] openWorkItemPicker(Shell parent, IWorkbenchPage page, ITeamRepository server) {
        WorkItemSelectionDialog dialog = new WorkItemSelectionDialog(parent, server, WorkItemSelectionDialog.MULTIPLE_SELECTION);
        dialog.setTitle(Messages.AssociateWorkItemToChangeSetAction_4);
        if (Window.OK == dialog.open())
            return (IWorkItemHandle[]) dialog.getResult();
        return new IWorkItemHandle[0];
    }
}
