/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems.actions;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.dialogs.ListSelectionDialog;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.calm.OslcLinkUtil;
import com.ibm.team.filesystem.client.internal.calm.OslcLinkUtil.DetailsInfo;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.internal.rcp.ui.workitems.Messages;
import com.ibm.team.filesystem.ui.actions.RemoveChangeSetChangeRequestsAction;
import com.ibm.team.filesystem.ui.item.AdapterUtil;
import com.ibm.team.filesystem.ui.wrapper.ChangeSetWrapper;
import com.ibm.team.internal.filesystem.ui.util.ItemIdComparator;
import com.ibm.team.jface.labelProviders.ElementRemovedNotifierImpl;
import com.ibm.team.links.common.IItemReference;
import com.ibm.team.links.common.ILink;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.IURIReference;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.rcp.ui.internal.utils.JFaceUtils;
import com.ibm.team.repository.rcp.ui.labelproviders.LabelProviders;
import com.ibm.team.repository.rcp.ui.operations.IOperationRunner;
import com.ibm.team.repository.rcp.ui.operations.IStatusCollector;
import com.ibm.team.repository.rcp.ui.operations.OperationFailedException;
import com.ibm.team.repository.rcp.ui.operations.RepositoryOperation;
import com.ibm.team.repository.rcp.ui.utils.AbstractActionDelegate;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.links.ChangeSetLinks;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.scm.common.providers.ProviderFactory;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

/**
 * Remove a list of work items from their change sets.
 */
public class RemoveChangeSetReasonsActions extends AbstractActionDelegate {

	/**
	 * Allow user to pick up from two types of link targets - IWorkItem or a
	 * resolvable IReference representing an OSLC link target. These objects can
	 * be rendered by the standard label provider and thus displayed in the
	 * dialog.
	 * 
	 * User can then map back to the original links and change sets e.g. to
	 * perform the actual deletion of the links (and oslc back links).
	 */
	class RelatedArtifacts {
		/**
		 * Repo that includes the change sets and links (but not necessarily
		 * their target)
		 */
		ITeamRepository repo;
		/**
		 * Links the user can decide to remove for the given change sets (change
		 * sets and links (but not their target) all belong to the given
		 * repository
		 * 
		 * UI doesn't know how to display links. Instead we must resolve the
		 * targets as IWorkItem for the
		 * {@link ILinkConstants#CHANGESET_WORKITEM_LINKTYPE_ID} type, and
		 * through the batch compact renderer for
		 * {@link ILinkConstants#CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID}
		 * which returns IReference's.
		 */
		Map<IChangeSetHandle, Collection<ILink>> changeSet2Links = new HashMap<IChangeSetHandle, Collection<ILink>>();

		/**
		 * Resolved targets for
		 * {@link ILinkConstants#CHANGESET_WORKITEM_LINKTYPE_ID} link type. UI
		 * can display those and user can select those.
		 * 
		 * @see #resolveArtifacts(IProgressMonitor)
		 */
		List<IWorkItem> workItems = new ArrayList<IWorkItem>();
		
		/**
		 * Unresolved targets. Try to provide as much info as possible for UI.
		 */
		List<LinkInfo> linkInfos = new ArrayList<LinkInfo>();

		/**
		 * Resolved targets for
		 * {@link ILinkConstants#CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID}
		 * link type. UI can display those and user can select those. This is
		 * obtained through the batch oslc compact renderer.
		 * 
		 * @see #resolveArtifacts(IProgressMonitor)
		 */
		Collection<IReference> resolvableOslcReferences = new ArrayList<IReference>();

		/**
		 * @see #repo
		 */
		RelatedArtifacts(ITeamRepository repo) {
			this.repo = repo;
		}

		void add(IChangeSetHandle changeSet, ILink link) {
			Collection<ILink> links = changeSet2Links.get(changeSet);
			if (links == null) {
				links = new ArrayList<ILink>();
				changeSet2Links.put(changeSet, links);
			}
			links.add(link);
		}

		/**
		 * Resolve the links - fetch the work items for
		 * {@link ILinkConstants#CHANGESET_WORKITEM_LINKTYPE_ID} and invoke
		 * compact renderer for
		 * {@link ILinkConstants#CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID}
		 * 
		 * @see #changeSet2Links
		 * @see #workItems
		 * @see #resolvableOslcReferences
		 */
		void resolveArtifacts(IProgressMonitor monitor) throws TeamRepositoryException {
			SubMonitor progress = SubMonitor.convert(monitor, 2);
			List<ILink> links = new ArrayList<ILink>();
			Map<UUID, IWorkItemHandle> uuidToWorkItemHandles = new HashMap<UUID, IWorkItemHandle>();
			Map<UUID, IURIReference> uuidToOslcReferences = new HashMap<UUID, IURIReference>();
			for (Entry<IChangeSetHandle, Collection<ILink>> entry : changeSet2Links
					.entrySet()) {
				for (ILink link : entry.getValue()) {
					String linkTypeId = link.getLinkTypeId();
					if (linkTypeId.equals(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID)) {
						if (link.getTargetRef().isItemReference()) {
							IItemHandle handle = ((IItemReference)link.getTargetRef()).getReferencedItem();
							if (handle instanceof IWorkItemHandle) {
								UUID workItemUUID = handle.getItemId();
								if (!uuidToWorkItemHandles.containsKey(workItemUUID)) {
									uuidToWorkItemHandles.put(workItemUUID, (IWorkItemHandle)handle);
									links.add(link);
								}
							}
						}
					} else if (linkTypeId.equals(ILinkConstants.CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID)) {
						if (link.getTargetRef().isURIReference()) {
							UUID oslcUUID = link.getItemId();
							if (!uuidToOslcReferences.containsKey(oslcUUID)) {
								uuidToOslcReferences.put(oslcUUID, (IURIReference)link.getTargetRef());
							}

						}
					}
				}
			}

			// keep track of deleted/inaccessible link targets to allow user to remove them
			if (!uuidToWorkItemHandles.isEmpty()) {
				workItems = repo.itemManager().fetchCompleteItems(new ArrayList<IWorkItemHandle>(uuidToWorkItemHandles.values()), IItemManager.DEFAULT, progress.newChild(1));
				int j = 0;
				for (Iterator i = workItems.iterator(); i.hasNext();) {
					Object wi = i.next();

					if (wi == null) {
						i.remove();
						LinkInfo linkInfo = new LinkInfo(links.get(j));
						linkInfos.add(linkInfo);
					}
					
					j++;
				}
			}
			if (!uuidToOslcReferences.isEmpty()) {
				List<DetailsInfo> infos;
					infos = OslcLinkUtil.fetchDetails(
							repo, new ArrayList<IURIReference>(uuidToOslcReferences.values()), true, progress.newChild(1));
				for (DetailsInfo info : infos) {
					if (info.resolvable != null)
						resolvableOslcReferences.add(info.resolvable);
				}				
			}
		}

		Collection<ILink> getAllLinks() {
			Collection<ILink> result = new ArrayList<ILink>();
			for (Collection<ILink> links : changeSet2Links.values()) {
				result.addAll(links);
			}
			return result;
		}
		
		Collection getAllResolvedArtifacts() {
			List result = new ArrayList();
			result.addAll(workItems);
			result.addAll(resolvableOslcReferences);
			result.addAll(linkInfos);
			return result;
		}

		/**
		 * Extract the work items from the user selection of work items and oslc
		 * resolved references
		 */
		Collection<IWorkItem> getWorkItems(Collection resolvedArtifacts) {
			List<IWorkItem> result = new ArrayList<IWorkItem>();
			for (Object artifact : resolvedArtifacts) {
				if (artifact instanceof IWorkItem)
					result.add((IWorkItem) artifact);
			}
			return result;
		}
		
		Collection<ILink> getLinkInfoLinks(Collection artifacts) {
			List<ILink> result = new ArrayList<ILink>();
			for (Object artifact : artifacts) {
				if (artifact instanceof LinkInfo) {
					LinkInfo linkInfo = (LinkInfo) artifact;
					result.add(linkInfo.link);
				}
			}
			return result;
		}

		/**
		 * Extract the oslc resolved references from the user selection of work
		 * items and oslc resolved references
		 */
		Collection<IReference> getResolvableOslcReferences(
				Collection resolvedArtifacts) {
			List<IReference> result = new ArrayList<IReference>();
			for (Object artifact : resolvedArtifacts) {
				if (artifact instanceof IReference)
					result.add((IReference) artifact);
			}
			return result;
		}

		/**
		 * Map the UI selection of work items back to the actual links (so
		 * caller can delete those links)
		 * 
		 * Note. Work items all belong to the same repo
		 * 
		 * @see #repo
		 */
		List<ILink> getWorkItemLinks(Collection<IWorkItem> workItems) {
			List<ILink> result = new ArrayList<ILink>();
			Map<UUID, IWorkItem> uuid2WorkItem = new HashMap<UUID, IWorkItem>();
			for (IWorkItem workItem : workItems)
				uuid2WorkItem.put(workItem.getItemId(), workItem);
			for (Entry<IChangeSetHandle, Collection<ILink>> entry : changeSet2Links.entrySet()) {
				for (ILink link : entry.getValue()) {
					if (link.getLinkTypeId().equals(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID)) {
						if (link.getTargetRef().isItemReference()) {
							IItemReference ref = (IItemReference)link.getTargetRef();
							IWorkItem workItem = null;
							if (ref.getReferencedItem() != null)
								workItem = uuid2WorkItem.get(ref.getReferencedItem().getItemId());
							if (workItem != null)
								result.add(link);
						}
					}
				}
			}
			return result;
		}

		/**
		 * Map the UI selection of resolvable oslc references back to the actual
		 * links and change set (so caller can delete those links)
		 */
		Map<IChangeSetHandle, List<ILink>> getOslcLinks(
				Collection<IReference> resolvableOslcReferences) {
			Map<IChangeSetHandle, List<ILink>> result = new HashMap<IChangeSetHandle, List<ILink>>();
			Map<UUID, ILink> uuid2Link = new HashMap<UUID, ILink>();
			for (IReference ref : resolvableOslcReferences) {
				ILink link = ref.getLink();
				if (link == null)
					throw new IllegalArgumentException();
				uuid2Link.put(link.getItemId(), link);
			}
			for (Entry<IChangeSetHandle, Collection<ILink>> entry : changeSet2Links.entrySet()) {
				for (ILink link : entry.getValue()) {
					if (link.getLinkTypeId().equals(ILinkConstants.CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID)) {
						ILink found = uuid2Link.get(link.getItemId());
						if (found != null) {
							List<ILink> links = result.get(entry.getKey());
							if (links == null) {
								links = new ArrayList<ILink>();
								result.put(entry.getKey(), links);
							}
							links.add(link);
						}
					}
				}
			}
			return result;
		}

	}

	private class WorkItemsToRemoveSelectionDialog extends ListSelectionDialog {
		public WorkItemsToRemoveSelectionDialog(Shell parentShell,
				RelatedArtifacts artifacts) {
			super(
					parentShell,
					artifacts.getAllResolvedArtifacts(),
					new ArrayContentProvider(),
					LabelProviders
							.createLabelProvider(new ElementRemovedNotifierImpl()),
					Messages.RemoveChangeSetReasonsActions_0);
			setTitle(Messages.RemoveChangeSetReasonsActions_1);
		}
	}
	
	public class LinkInfo {
		ILink link;
		
		public LinkInfo(ILink link) {
			this.link = link;
		}

		@Override
		public String toString() {
			String comment = link.getTargetRef().getComment();
			if (comment == null || comment.length() == 0)
				comment = Messages.RemoveChangeSetReasonsActions_7;
			return comment;
		}
	}

	@Override
	public void run(Shell shell, IWorkbenchPage page,
			IStructuredSelection selection) {
		Set<IChangeSetHandle> extractWrapperSelection = extractWrapperSelection(
				shell, selection);
		if (extractWrapperSelection.size() > 0) {
			deleteWorkItems(shell, getOperationRunner(),
					extractWrapperSelection);
		}
	}

	public void selectionChanged(IAction action, ISelection selection) {
		super.selectionChanged(action, selection);
		action.setEnabled(true);
	}

	private Set<IChangeSetHandle> extractWrapperSelection(Shell shell,
			IStructuredSelection selection) {
		final TreeSet<IChangeSetHandle> changeSetHandles = new TreeSet<IChangeSetHandle>(
				new ItemIdComparator());
		changeSetHandles.addAll(AdapterUtil.adaptList(selection.toList(),
				IChangeSetHandle.class));
		List<ChangeSetWrapper> wrappers = AdapterUtil.adaptList(selection
				.toList(), ChangeSetWrapper.class);
		for (ChangeSetWrapper changeSetWrapper : wrappers) {
			changeSetHandles.add(changeSetWrapper.getChangeSet());
		}
		return changeSetHandles;
	}

	/**
	 * Note. User selection is only allowed for change sets that belong to the same repository
	 */
	private ITeamRepository getOrigin(Collection<IChangeSetHandle> changeSets) {
		ITeamRepository result = null;
		for (IChangeSetHandle changeSet : changeSets) {
			ITeamRepository repo = (ITeamRepository)changeSet.getOrigin();
			if (repo == null)
				throw new IllegalArgumentException();
			if (result != null && repo != result)
				throw new IllegalArgumentException();
			result = repo;
		}
		return result;
	}
	
	private void deleteWorkItems(final Shell shell, IOperationRunner runner,
			final Set<IChangeSetHandle> list) {
		final ITeamRepository repo = getOrigin(list);
		runner.enqueue(Messages.RemoveChangeSetReasonsActions_4,
				new RepositoryOperation(repo) {
					public void repositoryRun(IProgressMonitor monitor,
							IStatusCollector problems)
							throws TeamRepositoryException,
							FileSystemException, OperationFailedException {
						try {
							SubMonitor progress = SubMonitor.convert(monitor, 4);
							final RelatedArtifacts artifacts = new RelatedArtifacts(repo);
							for (IChangeSetHandle csw : list) {
								ProviderFactory factory = (ProviderFactory) SCMPlatform
										.getClientLibrary(csw,
												ProviderFactory.class);
								List<ILink> links = ChangeSetLinks
										.findLinks(
												factory,
												csw,
												new String[] {
														ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID,
														ILinkConstants.CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID },
														progress.newChild(1));
								for (ILink link : links) {
									String linkTypeId = link.getLinkTypeId();
									if (linkTypeId.equals(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID) || linkTypeId.equals(ILinkConstants.CHANGESET_OSLC_CM_CHANGEREQUEST_LINKTYPE_ID))
										artifacts.add(csw, link);
								}
							}

							if (artifacts.getAllLinks().isEmpty()) {
								JFaceUtils
										.showMessage(
												Messages.RemoveChangeSetReasonsActions_8,
												Messages.RemoveChangeSetReasonsActions_9,
												MessageDialog.INFORMATION);
								return;
							}

							artifacts.resolveArtifacts(progress.newChild(1));
							final Collection selectedResolvedArtifacts = new ArrayList();
							JFaceUtils.syncExec(new Runnable() {
								public void run() {
									WorkItemsToRemoveSelectionDialog d = new WorkItemsToRemoveSelectionDialog(
											shell, artifacts);
									if (d.open() == Window.OK) {
										selectedResolvedArtifacts.addAll((Arrays
												.asList(d.getResult())));
									}
								}
							});
							Collection<IReference> resolvableOslcReferences = artifacts.getResolvableOslcReferences(selectedResolvedArtifacts);
							Collection<IWorkItem> workItems = artifacts.getWorkItems(selectedResolvedArtifacts);
							Collection<ILink> linkInfoLinks = artifacts.getLinkInfoLinks(selectedResolvedArtifacts);
							
							if (resolvableOslcReferences.size() > 0) {
								SubMonitor oslcProgress = progress.newChild(1);
								oslcProgress.setWorkRemaining(2 * resolvableOslcReferences.size());
								List<ILink> allFailures = new ArrayList<ILink>();
								List<ILink> allLinks = new ArrayList<ILink>();
								Map<IChangeSetHandle, List<ILink>> changeSet2Links = artifacts.getOslcLinks(resolvableOslcReferences);
								for (Entry<IChangeSetHandle, List<ILink>> entry : changeSet2Links.entrySet()) {
									List<ILink> failures = RemoveChangeSetChangeRequestsAction.removeOslcCMBackLinks(repo, entry.getValue(), entry.getKey(), oslcProgress.newChild(1));
									allFailures.addAll(failures);
									allLinks.addAll(entry.getValue());
								}
								if (allFailures.size() > 0) {
									if(!JFaceUtils.showMessageBlocking(
											Messages.RemoveChangeSetReasonsActions_5, 
											Messages.RemoveChangeSetReasonsActions_6, 
											new String[] { IDialogConstants.YES_LABEL, IDialogConstants.NO_LABEL },
											MessageDialog.WARNING)) {
										allLinks.removeAll(allFailures);
									}
								}
								if (allLinks.size() > 0)
									OslcLinkUtil.deleteScmOslcLinks(repo, allLinks, oslcProgress.newChild(resolvableOslcReferences.size()));
							}
							if (workItems.size() > 0 || linkInfoLinks.size() > 0) {
								List<ILink> links = artifacts.getWorkItemLinks(workItems);
								links.addAll(linkInfoLinks);
								IFileSystemWorkItemManager linkService = (IFileSystemWorkItemManager) repo
								.getClientLibrary(IFileSystemWorkItemManager.class);
								linkService.deleteLinks(links,
										progress.newChild(1));
							}
						} finally {
							monitor.done();
						}
					}
				});
	}
}
