/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2013. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems.deliveradvisor;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.ListSelectionDialog;

import com.ibm.team.filesystem.client.internal.LoggingHelper;
import com.ibm.team.filesystem.client.internal.utils.RepositoryUtils;
import com.ibm.team.filesystem.common.workitems.internal.process.RequireWorkItemProblemObject;
import com.ibm.team.filesystem.ide.ui.process.providers.ParsedUnifiedDescriptionRegistry;
import com.ibm.team.filesystem.internal.rcp.ui.workitems.ScmWorkItemUi;
import com.ibm.team.jface.labelProviders.ElementRemovedNotifierImpl;
import com.ibm.team.process.client.IAdvisorProblemResolutionDelegate;
import com.ibm.team.process.client.IProcessItemService;
import com.ibm.team.process.common.IDevelopmentLine;
import com.ibm.team.process.common.IDevelopmentLineHandle;
import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProjectArea;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.common.ITeamArea;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.core.utils.StatusUtil;
import com.ibm.team.repository.rcp.ui.labelproviders.LabelProviders;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.AdvisorData;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.IScmAdvisorMarshallable;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.MarshallerException;
import com.ibm.team.workitem.client.IWorkItemClient;
import com.ibm.team.workitem.client.IWorkItemWorkingCopyManager;
import com.ibm.team.workitem.client.WorkItemWorkingCopy;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.ItemProfile;
import com.ibm.team.workitem.rcp.ui.WorkItemUI;

public abstract class AbstractWorkItemResolutionDelegate implements IAdvisorProblemResolutionDelegate {
	
	public abstract boolean mayBeApplicable(RequireWorkItemProblemObject problemObject);

	public abstract void modifyWorkItem(IWorkItem item, IDevelopmentLine devLine, ITeamRepository teamRepository, IProgressMonitor monitor) throws TeamRepositoryException;

	/*
	 * @see com.ibm.team.process.client.IAdvisorProblemResolutionDelegate#mayBeApplicable(com.ibm.team.process.common.advice.IAdvisorInfo)
	 */
	public boolean mayBeApplicable(IAdvisorInfo info) {
		AdvisorData<IScmAdvisorMarshallable> unified;
		try {
			unified = ParsedUnifiedDescriptionRegistry.getInstance().parse(info);
		} catch (MarshallerException e) {
			return false;
		}
		
		if (unified != null && unified.data instanceof RequireWorkItemProblemObject) {
			return mayBeApplicable((RequireWorkItemProblemObject) unified.data);
		}
		return false;
	}

	/*
	 * @see com.ibm.team.process.client.IAdvisorProblemResolutionDelegate#run(com.ibm.team.process.common.advice.IAdvisorInfo, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public boolean run(IAdvisorInfo info, IProgressMonitor progress) {
		SubMonitor monitor = SubMonitor.convert(progress, Messages.AbstractWorkItemResolutionDelegate_0, 1000);
		
		Object problemObject;
		try {
			AdvisorData<IScmAdvisorMarshallable> data = ParsedUnifiedDescriptionRegistry.getInstance().parse(info);

			if (data == null) {
				return false;
			}
			
			problemObject = data.data;
		} catch (MarshallerException e) {
			LoggingHelper.log(ScmWorkItemUi.PLUGIN_ID, new TeamRepositoryException("Could not parse problem object", e)); //$NON-NLS-1$
			return false;
		}
		
		if (problemObject instanceof RequireWorkItemProblemObject) {
			final Display display = Display.getDefault();
			RequireWorkItemProblemObject data = (RequireWorkItemProblemObject) problemObject;
			IProjectAreaHandle projectArea = data.getProjectArea();
			ITeamRepository teamRepository = RepositoryUtils.getTeamRepositoryById(data.getRepoUuid());
			List workItems = ((RequireWorkItemProblemObject) problemObject).getWorkItems();
			int size = workItems.size();
			if (size == 1) {
				try {
					modifyWorkItem(display, teamRepository, projectArea, (IWorkItemHandle) workItems.get(0), monitor);
				} catch (TeamRepositoryException e) {
					reportException(e);
				}
			} else if (size > 1) {
				try {
					// multiple work items are around, let the user pick which ones to change
					final List fetchedWorkItems = teamRepository.itemManager().fetchCompleteItems(workItems, IItemManager.DEFAULT, monitor.newChild(100));
					final Object[] result = { null };
					display.syncExec(new Runnable() {
						public void run() {
							ListSelectionDialog dialog = new ListSelectionDialog(null, fetchedWorkItems, ArrayContentProvider.getInstance(), LabelProviders
									.createLabelProvider(new ElementRemovedNotifierImpl()), Messages.AbstractWorkItemResolutionDelegate_4);
							dialog.setTitle(Messages.AbstractWorkItemResolutionDelegate_3);
							if (dialog.open() == Window.OK) {
								result[0] = dialog.getResult();
							}
						}
					});

					Object[] selectedWorkItems = (Object[]) result[0];
					if (selectedWorkItems == null || selectedWorkItems.length == 0) {
						// user cancelled or pressed OK while selecting nothing, return now
						return false;
					}

					monitor.setWorkRemaining(selectedWorkItems.length);
					// modify the work items that the user selected
					for (Object selected : selectedWorkItems) {
						modifyWorkItem(display, teamRepository, projectArea, (IWorkItemHandle) selected, monitor.newChild(1));
					}
					return true;
				} catch (TeamRepositoryException e) {
					reportException(e);
				}
			}
		}
		return false;
	}

	/**
	 * Modifies the work item so that it will now satisfy one of the
	 * requirements of the advisor.
	 * 
	 * @param display
	 *            the SWT display for interacting with the user interface, must
	 *            not be <tt>null</tt>
	 * @param repository
	 *            the repository to interact with, must not be <tt>null</tt>
	 * @param projectArea
	 *            the project area that the operation was run in, must not be
	 *            <tt>null</tt>
	 * @param workItem
	 *            the work item to modify, must not be <tt>null</tt>
	 * @param progress
	 *            a progress monitor for reporting progress or supporting
	 *            cancellation, or <tt>null</tt> if progress monitoring or
	 *            cancellation support is not required
	 * @throws TeamRepositoryException
	 *             if an error occurs while trying to manipulate the work item
	 *             or retrieve information from the server
	 */
	private void modifyWorkItem(final Display display, ITeamRepository repository, IProjectAreaHandle projectArea, IWorkItemHandle workItem, IProgressMonitor progress)
			throws TeamRepositoryException {
		SubMonitor monitor = SubMonitor.convert(progress, 1000);
		IWorkItemClient client = (IWorkItemClient) repository.getClientLibrary(IWorkItemClient.class);
		IWorkItemWorkingCopyManager workingCopyManager = client.getWorkItemWorkingCopyManager();
		workingCopyManager.connectCurrent(workItem, ItemProfile.createFullProfile(IWorkItem.ITEM_TYPE), monitor.newChild(125));
		WorkItemWorkingCopy workingCopy = workingCopyManager.getWorkingCopy(workItem);
		boolean wasDirty = workingCopy.isDirty();
		IWorkItem item = workingCopy.getWorkItem();
		IProcessArea processArea = (IProcessArea) repository.itemManager().fetchCompleteItem(projectArea, IItemManager.DEFAULT, monitor.newChild(100));
		IDevelopmentLine developmentLine = null;
		if (processArea instanceof ITeamArea) {
			IProcessItemService processService = (IProcessItemService) repository.getClientLibrary(IProcessItemService.class);
			developmentLine = processService.getDevelopmentLine(processArea, new SubProgressMonitor(monitor, 25));
		} else {
			IDevelopmentLineHandle projectDevLine = ((IProjectArea) processArea).getProjectDevelopmentLine();
			if (projectDevLine != null) {
				developmentLine = (IDevelopmentLine) repository.itemManager().fetchCompleteItem(projectDevLine, IItemManager.DEFAULT, new SubProgressMonitor(monitor, 25));
			}
		}
		modifyWorkItem(item, developmentLine, repository, new SubProgressMonitor(monitor, 250));
		if (!wasDirty) {
			workingCopy.save(new SubProgressMonitor(monitor, 500));
		} else {
			final WorkItemWorkingCopy finalWorkingCopy = workingCopy;
			display.asyncExec(new Runnable() {
				public void run() {
					if (!display.isDisposed()) {
						IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
						if (window != null) {
							IWorkbenchPage page = window.getActivePage();
							if (page != null) {
								WorkItemUI.openEditor(page, finalWorkingCopy.getWorkItem());
							}
						}
					}
				}
			});
		}
	}

	/**
	 * Logs the given exception and opens a dialog to report it to the user.
	 */
	private void reportException(final TeamRepositoryException e) {
		StatusUtil.log(this, e);
		final Display display = Display.getDefault();
		display.asyncExec(new Runnable() {
			public void run() {
				if (!display.isDisposed()) {
					Shell shell = display.getActiveShell();
					if (shell == null) {
						Shell[] shells = display.getShells();
						if (shells.length > 0) {
							shell = shells[0];
						}
					}
					if (shell != null) {
						ErrorDialog.openError(
								shell, Messages.AbstractWorkItemResolutionDelegate_1, null, 
								StatusUtil.newStatus(this, Messages.AbstractWorkItemResolutionDelegate_2, e));
					}
				}
			}
		});
	}
}
