/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems.deliveradvisor;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

import com.ibm.team.filesystem.client.internal.utils.RepositoryUtils;
import com.ibm.team.filesystem.common.workitems.internal.process.RequireWorkItemProblemObject;
import com.ibm.team.filesystem.ide.ui.process.providers.ParsedUnifiedDescriptionRegistry;
import com.ibm.team.filesystem.ui.changes.dialogs.EditCommentDialog;
import com.ibm.team.process.client.IAdvisorProblemResolutionDelegate;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.core.utils.StatusUtil;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.AdvisorData;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.IScmAdvisorMarshallable;
import com.ibm.team.scm.common.internal.process.AdvisorDataMarshaller.MarshallerException;

/**
 * Allows the user to bulk assign comments to change sets.
 */
public class AssignCommentResolutionDelegate implements IAdvisorProblemResolutionDelegate {

	public boolean mayBeApplicable(IAdvisorInfo info) {
		if (RequireWorkItemProblemObject.ID_REQUIRE_WORK_ITEM_PROBLEM.equals(info.getIdentifier())) {
			// This resolution is not applicable if the problem is a missing comment
			Object object;
			try {
				AdvisorData<IScmAdvisorMarshallable> data = ParsedUnifiedDescriptionRegistry.getInstance().parse(info);
				if (data == null) {
					return false;
				}
				
				object = data.data;
			} catch (MarshallerException e) {
				return false;
			}
			
			if (! (object instanceof RequireWorkItemProblemObject)) {
				return false;
			}
			
			return object != null && ((RequireWorkItemProblemObject)object).wantsComment();
		}
		return false;
	}

	public boolean run(IAdvisorInfo info, IProgressMonitor monitor) {
		final Object object;
		try {
			AdvisorData<IScmAdvisorMarshallable> data = ParsedUnifiedDescriptionRegistry.getInstance().parse(info);
			if (data == null) {
				return false;
			}
			
			object = data.data;
		} catch (MarshallerException e) {
			return false;
		}
		
		if (! (object instanceof RequireWorkItemProblemObject)) {
			return false;
		}

		final boolean[] result = new boolean[] { false };

        final Display display = Display.getDefault();
		display.syncExec(new Runnable() {
			public void run() {
		        String newComment = editCommentDialog(display.getActiveShell());
		        if (newComment == null || "".equals(newComment.trim())) { //$NON-NLS-1$
		            result[0] = false;
		            return;
		        }
		        
		        scheduleAssignment((RequireWorkItemProblemObject)object, newComment);
		        result[0] = true;
			}
		});
		
		return result[0];
	}

	/**
	 * Ask the user to set a comment for the change sets. 
	 * 
	 * @return <code>null</code> if the user canceled; otherwise, the edited 
	 *         string.
	 */
	private String editCommentDialog(Shell shell) {
	    final String[] result = new String[] {null};
		EditCommentDialog dialog = new EditCommentDialog(shell, "") { //$NON-NLS-1$
			@Override
			protected void saveOnOk(String value) {
			    result[0] = value;
			}
		};
		
		dialog.open();
		
		return result[0];
	}
	

	/**
	 * Assign the change set comment. 
	 */
    private void scheduleAssignment(final RequireWorkItemProblemObject problem, final String newComment) {
    	Job job = new Job(Messages.AssociateWorkItemResolutionDelegate_0) {
    		protected IStatus run(IProgressMonitor progress) {
    		    SubMonitor monitor = SubMonitor.convert(progress, Messages.AssignCommentResolutionDelegate_PROGRESS_MESSAGE, 1);
    			try {
    			    ITeamRepository repo = RepositoryUtils.getTeamRepositoryById(problem.getRepoUuid());
    			    
    			    SCMPlatform.getWorkspaceManager(repo).setComment(problem.getChangeSet(), newComment, monitor.newChild(1));
    			    
    				return Status.OK_STATUS;
    				
    			} catch (TeamRepositoryException e) {
    				return StatusUtil.newStatus(this, Messages.AssignCommentResolutionDelegate_ERROR_MESSAGE, e);
    				
    			} finally {
    				monitor.done();
    			}
    		}
    	};
    	
    	job.setUser(true);
    	job.schedule();
    }
}
