/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.internal.rcp.ui.workitems.deliveradvisor;

import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_ATTR_MISSING_APPROVAL_COUNT;
import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_ATTR_REQUIRED_APPROVAL_STRING;
import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_ATTR_REQUIRED_ROLE_STRING;
import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_ATTR_WI_NUMBER;
import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_ATTR_WI_UUID;
import static com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants.ERROR_TAG_APPROVAL;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.osgi.util.NLS;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.ibm.team.filesystem.client.internal.utils.RepositoryUtils;
import com.ibm.team.filesystem.common.internal.process.RequireWorkItemApprovalConstants;
import com.ibm.team.filesystem.common.workitems.internal.process.RequireWorkItemApprovalAdvisor;
import com.ibm.team.filesystem.ide.ui.process.AbstractSourceControlAdvisorDetailProvider;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.LogFactory;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.rcp.ui.WorkItemUI;

/**
 * Provides rich UI for advice failures in 
 * {@link RequireWorkItemApprovalAdvisor}.
 */
public class RequiredWorkItemApprovalDetailProvider extends AbstractSourceControlAdvisorDetailProvider {

    public RequiredWorkItemApprovalDetailProvider(Object element) {
        super(element);
    }

    private final static class FailureInput {
    	/** The set of failed requirements. */
    	public SortedSet<RequirementFailure> failures = new TreeSet<RequirementFailure>(new RequirementFailureComparator());
    	
    	/** The repo the failures come from. */
    	ITeamRepository repo;
    	
    	/** The WI handles that we're dealing with. */
    	List<IWorkItemHandle> handles = new LinkedList<IWorkItemHandle>();
    }
    
    private final static class RequirementFailure {
        public UUID workitem_uuid;
        
        public String workitem_number;
        public int workitem_number_int;
        
        public String approval_string;
        
        public String role_string;

        public int missing_count;
    }

    /** Comparator to provide ordering to the list of unsatisfied work items. */
    private final static class RequirementFailureComparator implements Comparator<RequirementFailure> {

		public int compare(RequirementFailure o1, RequirementFailure o2) {
			int r = o1.workitem_number_int - o2.workitem_number_int;
			
			if (r == 0) {
				r = String.CASE_INSENSITIVE_ORDER.compare(o1.approval_string, o2.approval_string);
			}
			
			if (r == 0) {
			    r = String.CASE_INSENSITIVE_ORDER.compare(o1.role_string, o2.role_string);
			}
			
			return r;
		}
    }
    

    @Override
    public boolean requiresLongOp() {
        return true;
    }
    
    @Override
    protected String getShortOpDetailText() {
        return Messages.RequiredWorkItemApprovalDetailProvider_FETCHING_PROGRESS;
    }

    
    @Override
    protected void appendProblem(IAdvisorInfo info, StringBuffer buf, IProgressMonitor monitor) {
        FailureInput input = getFailedRequirements(info.getData());
        
        if (input == null || input.repo == null) {
        	buf.append(info.getDescription());
        	
        	return;
        }
        
        // Find the work items
        List<IWorkItem> wis;
        try {
			wis = input.repo.itemManager().fetchCompleteItems(input.handles, IItemManager.DEFAULT, monitor);
		} catch (TeamRepositoryException e) {
			// TODO Include the stack trace
			buf.append("Error fetching work items: " + e.getLocalizedMessage()); //$NON-NLS-1$
			return;
		}

		final HashMap<UUID, IWorkItem> keyed = keyByUuid(wis);
        
		// Provide an explanation of the failure
		if (input.failures.size() == 1) {
			buf.append(Messages.RequiredWorkItemApprovalDetailProvider_MISSING_WORK_ITEM_APPROVAL_PREVENTS_DELIVERY_OF_CHANGES_SINGULAR); 
		} else {
			buf.append(Messages.RequiredWorkItemApprovalDetailProvider_MISSING_WORK_ITEM_APPROVAL_PREVENTS_DELIVERY_OF_CHANGES_PLURAL); 
		}
		
		// List the failures
        buf.append("<ul>"); //$NON-NLS-1$
        for (final RequirementFailure failure : input.failures) {
            buf.append("<li>"); //$NON-NLS-1$
            
            StringBuffer hyperlinkBuf = new StringBuffer();
            appendHyperlink(NLS.bind(Messages.RequiredWorkItemApprovalDetailProvider_WORK_ITEM_NUMBER, failure.workitem_number), "", hyperlinkBuf, new Runnable() { //$NON-NLS-1$
				public void run() {
			        IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
			        if (window != null) {
			            IWorkbenchPage page = window.getActivePage();
			            if (page != null) {
		                    WorkItemUI.open(page, keyed.get(failure.workitem_uuid));
			            }
			        }
				}
			});
            
            buf.append(NLS.bind(Messages.RequiredWorkItemApprovalDetailProvider_WORK_ITEM_IS_MISSING_NUMBER_OF_APPROVALS_OF_TYPE_FOR_GIVEN_ROLE, new Object[] { 
                    hyperlinkBuf, 
                    failure.missing_count, 
                    failure.approval_string,
                    failure.role_string}));
            
            buf.append("</li>\n"); //$NON-NLS-1$
        }
        buf.append("</ul>"); //$NON-NLS-1$
    }

    private HashMap<UUID, IWorkItem> keyByUuid(List<IWorkItem> wis) {
    	HashMap<UUID, IWorkItem> keyed = new HashMap<UUID, IWorkItem>();
    	
    	for (IWorkItem wi : wis) {
    		keyed.put(wi.getItemId(), wi);
    	}
    	
    	return keyed;
	}
    
    
    private final FailureInput getFailedRequirements(String data) {
    	FailureInput f = new FailureInput();
    	
    	if (data == null || "".equals(data)) { //$NON-NLS-1$
    		// We have no structured data.
    		return null;
    	}
    	
        Element dataElement = AbstractSourceControlAdvisorDetailProvider.getDataElement(data);
        if (dataElement == null) {
        	// The XML failed to parse
        	return null;
        }
        
        f.repo = findRepoForUuid(dataElement.getAttribute(RequireWorkItemApprovalConstants.ROOT_ATTR_REPOSITORY));
        
        HashSet<UUID> seenWiUuids = new HashSet<UUID>();
        if (dataElement != null && f.repo != null) {
            NodeList errors = dataElement.getElementsByTagName(ERROR_TAG_APPROVAL);
            for (int i = 0; i < errors.getLength(); i++) {
                Element fileElement = (Element) errors.item(i);
                
                RequirementFailure parsed = new RequirementFailure();
                f.failures.add(parsed);
                
                parsed.workitem_uuid = UUID.valueOf(fileElement.getAttribute(ERROR_ATTR_WI_UUID));
                
                parsed.workitem_number = fileElement.getAttribute(ERROR_ATTR_WI_NUMBER);     
                try {
                    parsed.workitem_number_int = Integer.parseInt(parsed.workitem_number);
                } catch (NumberFormatException e) {
                    LogFactory.getLog(getClass().getName()).error("Failed to parse \"" + parsed.workitem_number + "\"");  //$NON-NLS-1$//$NON-NLS-2$
                    parsed.workitem_number_int = -1;
                }
                
                parsed.approval_string = fileElement.getAttribute(ERROR_ATTR_REQUIRED_APPROVAL_STRING);
                
                parsed.role_string = fileElement.getAttribute(ERROR_ATTR_REQUIRED_ROLE_STRING);

                if (!seenWiUuids.contains(parsed.workitem_uuid)) {
                	f.handles.add(inflateWorkItem(parsed.workitem_uuid));
                	seenWiUuids.add(parsed.workitem_uuid);
                }

                String toParse = fileElement.getAttribute(ERROR_ATTR_MISSING_APPROVAL_COUNT);
                try {
                    parsed.missing_count = Integer.parseInt(toParse);
                } catch (NumberFormatException e) {
                    LogFactory.getLog(getClass().getName()).error("Failed to parse \"" + toParse + "\"");  //$NON-NLS-1$//$NON-NLS-2$
                    parsed.missing_count = -1;
                }
            }
        }

        return f;
    }
    
    
    /** May return null */
    private ITeamRepository findRepoForUuid(String uuidString) {
		return RepositoryUtils.getTeamRepositoryById(UUID.valueOf(uuidString));
	}

	private IWorkItemHandle inflateWorkItem(UUID uuid) {
        return (IWorkItemHandle) IWorkItem.ITEM_TYPE.createItemHandle(uuid, null);
	}

	public void openWorkItem(final IWorkItem wi) {
        IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
        if (window == null) {
            IWorkbenchWindow[] workbenchWindows = PlatformUI.getWorkbench().getWorkbenchWindows();
            if (workbenchWindows.length > 0) {
                window = workbenchWindows[0];
            }
        }
        
        if (window != null) {
            final IWorkbenchPage page = window.getActivePage();
            if (page != null) {
                window.getShell().getDisplay().asyncExec(new Runnable() {
                    public void run() {
                    	WorkItemUI.openEditor(page, wi);
                    }
                });
            }
        }
    }
    
}
