/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.metronome;

import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.core.runtime.preferences.IEclipsePreferences.IPreferenceChangeListener;
import org.eclipse.core.runtime.preferences.IEclipsePreferences.PreferenceChangeEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.internal.WorkbenchWindow;
import org.eclipse.ui.internal.layout.ITrimManager;
import org.eclipse.ui.internal.layout.IWindowTrim;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import com.ibm.team.filesystem.ui.UiPlugin;
import com.ibm.team.metronome.internal.MetronomeControl;
import com.ibm.team.metronome.internal.VisibilityController;
import com.ibm.team.metronome.traffic.TrafficControl;

/**
 * The activator class controls the plug-in life cycle
 */
public class MetronomePlugin extends AbstractUIPlugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "com.ibm.team.metronome"; //$NON-NLS-1$

	/** Id of the metronome trim. */
	public static final String METRONOME_TRIM_GROUP = "com.ibm.team.metronome.group"; //$NON-NLS-1$

    
    // The shared instance
	private static MetronomePlugin plugin;
	
	private MetronomeControl control;
	private TrafficControl trafficControl;

	public MetronomePlugin() {
	}

	/** Listener responsible for keeping metronome's visibility in sync with 
	 * the {@link UiPlugin#SHOW_METRONOME_TRIM} preference.
	 */
	IPreferenceChangeListener listener = new IPreferenceChangeListener() {

		public void preferenceChange(PreferenceChangeEvent event) {
			if (UiPlugin.SHOW_METRONOME_TRIM.equals(event.getKey())) {
			    VisibilityController.updateVisibility();
			}
		}
	};

    /**
     * Toggle the Metronome's presence in the task bar. This setting is not
     * persistent. For that, update the {@link UiPlugin#SHOW_METRONOME_TRIM}
     * preference.
     * 
     * @return Whether the attempt to update was successful.
     */
	@SuppressWarnings("restriction")
	public boolean setMetronomeVisible(boolean visible) {
		final IWorkbenchWindow[] windows= PlatformUI.getWorkbench().getWorkbenchWindows();
		if (windows.length == 0) {
			return false;
		}
		boolean updated= false;
		for (final IWorkbenchWindow window : windows) {
			if (window instanceof WorkbenchWindow) {
				final WorkbenchWindow extended= (WorkbenchWindow)window;
				final ITrimManager trimManager= extended.getTrimManager();
				if (trimManager != null) {
					final IWindowTrim trim= trimManager.getTrim(MetronomePlugin.METRONOME_TRIM_GROUP);
					if (trim != null) {
						updated= true;
						trimManager.setTrimVisible(trim, visible);
						trim.getControl().getParent().layout(true);
					}
				}
			}
		}
		/*
		 * Fallback logic for Eclipse 4.x installations. Eclipse 3.x execution
		 * path may also drop into this code during startup before trim manager
		 * is fully initialized.
		 */
		if (!updated && control != null) {
			control.setVisible(visible);
			if (trafficControl != null) {
				trafficControl.setVisible(visible);
			}

			Shell shell= control.getShell();
			Composite statusBar= control.getParent();
			// The status bar is a direct child of the shell for both Eclipse
			// 3.x and 4.x
			while (statusBar.getParent() != shell) {
				statusBar= statusBar.getParent();
			}
			statusBar.computeSize(SWT.DEFAULT, SWT.DEFAULT, true);
			if (shell != null) {
				shell.layout();
			}
		}

		return updated;
	}
	
    
	private IEclipsePreferences getUiPreferences() {
		return new InstanceScope().getNode(UiPlugin.PLUGIN_ID);
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		plugin = this;
		
		getUiPreferences().addPreferenceChangeListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
        getUiPreferences().removePreferenceChangeListener(listener);

        plugin = null;
		
		super.stop(context);
	}

	/**
	 * Returns the shared instance
	 *
	 * @return the shared instance
	 */
	public static MetronomePlugin getDefault() {
		return plugin;
	}

	public void setControl(MetronomeControl control) {
		this.control = control;
	}
	
	public void setTrafficControl(TrafficControl control) {
		this.trafficControl = control;
	}
}
