/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import com.ibm.team.scm.client.importz.spi.ChangeSetFileWriter;

/**
 * As an import process must be able to handle a vast number of import change
 * sets it is not always possible to hold all necessary information in memory.
 * Thus it is necessary to store the information temporarily on disk. For this
 * purpose we need <code>IChangeSetFileReader</code> and
 * <code>IChangeSetFileWriter</code>.
 * </p>
 * 
 * <p>
 * An </code>IChangeSetFileWriter</code> represents the general interface for
 * writing change set information into of a given file.
 * </p>
 * 
 * <p>
 * Note: A concrete log file writer must be created via a corresponding factory.
 * Thus it's necessary that one must implement the contrary reader as well.
 * </p>
 * 
 * <p>
 * This interface is not intended to be implemented by clients. Clients can
 * instantiate or subclass the {@link ChangeSetFileWriter} class instead.
 * 
 * @since 0.6
 * 
 * @see com.ibm.team.scm.client.importz.IChangeSetFileReader
 * @see ChangeSetFileWriter
 *      com.ibm.team.migration.ccvs.IMigrationFactory
 */
public interface IChangeSetFileWriter {
	
	/**
	 * Starts the write process (necessary initializations).
	 * A call to this method will open the file for writing and
	 * the file will be closed when {@link #closeWrite()} is called.
	 * 
	 * @param file the file in which the following information should be stored, must not be null
	 * @throws IOException if problems occur while accessing the given file
	 */
	void startWrite(File file) throws IOException;

	/**
	 * Starts the write process (necessary initializations) to the given stream.
	 * Clients that call this method are responsible for closing the provided
	 * stream themselves (i.e. a call to {@link #closeWrite()} will not close the
	 * stream.
	 *  
	 * 
	 * @param stream the stream to which the following information should be stored, must not be null
	 * @throws IOException if problems occur
	 */
	void startWrite(OutputStream stream) throws IOException;

	
	/**
	 * Writes the given change set with its changes inside a file.
	 * 
	 * @param changeSet the change set, must not be null
	 * @throws IOException if problems occur while accessing the file
	 */
	void writeElement(IImportChangeSet changeSet) throws IOException;

	/**
	 * Writes the given change set with the provided changes inside a file.
	 * 
	 * @param changeSet the change set, must not be null
	 * @param changes the changes to be written with the change set, must not be null
	 * @throws IOException if problems occur while accessing the file
	 */
	void writeElement(IImportChangeSet changeSet, List<IImportChange> changes) throws IOException;
	
	/**
	 * Closes the write process.
	 * 
	 * @throws IOException if problems occur while accessing the file
	 */
	void closeWrite() throws IOException;

}
