/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz;

import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.scm.client.importz.spi.ImportChange;

/**
 * An <code>IImportChange</code> is the general representation of a change. A new IFileSystemItem will
 * be imported to the Apollo repository respectively to the given repository relative path, 
 * the kind and the contents of the change.
 * <p>
 * This interface is not intended to be implemented by clients. Import service providers 
 * should instead use or subclass {@link ImportChange}.
 * 
 * @since 0.6
 */
public interface IImportChange {

	/**
	 * The resource the change is related to is unknown. Thus the change will not be
	 * treated inside the import process. 
	 */
	int UNKNOWN = 0;
	
	/**
	 * The change is related to a file.
	 */
	int FILE = 0x1;
	
	/**
	 * The change is related to a folder.
	 */
	int FOLDER = 0x2;
	
	/**
	 * The change is an addition
	 */
	int ADDITION = 0x10;
	
	/**
	 * The change is an deletion
	 */
	int DELETION = 0x20;
	
	/**
	 * The change is a modification to the file path or content
	 */
	int MODIFICATION = 0x40;
	
	/**
	 * The change is a modification to the file content
	 */
	int CONTENT = 0x100;
	
	/**
	 * The file has moved
	 */
	int MOVE = 0x200;
	
	/**
	 * Returns the type of the item (FILE, FOLDER or UNKNOWN) affected 
	 * by the change the change. 
	 * 
	 * @return the type of the item affected by the change
	 */
	int getItemType();
	
	/**
	 * Return the kind of the change (ADDITION, DELETION, MODIFICATION)
	 * @return the kind of the change (ADDITION, DELETION, MODIFICATION)
	 */
	int getKind();
	
	/**
	 * Return the bit-wise combination of flags that describe the change in more detail.
	 * If the kind is modification, the set of valid flag combinations are MOVE,
	 * MOVE | CONTENT, CONTENT, COPY, COPY | CONTENT.
	 * @return the bit-wise combination of flags that describe the change in more detail
	 */
	int getFlags();
	
	/**
	 * Returns an input stream with the contents of the related file
	 * in the after state of this change. May return <code>null</code> if 
	 * the change isn't related to a file, the contents couldn't be resolved, the
	 * change is a file deletion, or the change isn't a content change.
	 * 
	 * @param importData the change set archive being imported
	 * @param monitor TODO
	 * @return an input stream with the contents of the related file, maybe null
	 */
	InputStream getContents(IImportData importData, IProgressMonitor monitor) throws IOException;
	
	/**
	 * Returns the path of the affected file/folder before the change
	 * or <code>null</code> for an addition that does not have the 
	 * MOVE bit set.
	 * 
	 * @return the path before the change or <code>null</code>
	 */
	String getBeforePath();
	
	/**
	 * Returns path of the affected file/folder after the change
	 * or <code>null</code> for a deletion.
	 * 
	 * @return the path after the change or <code>null</code>
	 */
	String getAfterPath();
	
	/**
	 * Returns the string representation of the revision of this change,
	 * or null if it has no revision info
	 * 
	 * Note: non deletion file changes are required to have a revision in order
	 * for the content to be cached
	 * 
	 * @return the string representing the revision
	 */
	String getRevision();
	
	/**
	 * Returns the properties associated with the contents of this change or
	 * <code>null</code> if there is no content associated with the change.
	 * @return the properties associated with the contents of this change or
	 * <code>null</code>
	 */
	IFileProperties getFileProperties();

	/**
	 * Return whether this change represents a change to the ignore file 
	 * of a folder.
	 * @return whether this change represents a change to the ignore file
	 */
	boolean isIgnoreFileChange();

	/**
	 * Return a new change that is derived from this change but is converted
	 * from an addition to a modification. This method is used when importing
	 * into an existing component with reconciliation enabled.
	 * @return a new change that is derived from this change but is converted
	 * from an addition to a modification
	 */
	IImportChange asModificationForReconciliation();
}
