/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz;

import java.io.File;
import java.io.IOException;

import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.importz.spi.ChangeSetFileWriter;
import com.ibm.team.scm.client.importz.spi.LogCacheMerger;
import com.ibm.team.scm.client.importz.spi.MigrationFactory;
import com.ibm.team.scm.common.IComponent;

/**
 * Factory for the creating repository type specific classes for participating in the import
 * process.
 * 
 * <p>
 * This interface is not intended to be implemented by clients. Clients should instead subclass
 * {@link MigrationFactory}.
 * 
 * @since 0.6
 * 
 * @see MigrationFactory
 */
public interface IMigrationFactory {
	
	/**
	 * Return the human readable short name of this factory which is displayed in progress
	 * messages and error displayed during the import process.
	 * @return the name of this factory
	 */
    String getName();
    
	/**
	 * Create a change set file reader that will extract the change sets from
	 * the archive associated with the import data.
	 * @param importData the import data.
	 * @return a change set file reader
	 * @throws IOException if an error occurs extracting the log file from the archive
	 */
	IChangeSetFileReader createChangeSetFileReader(IImportData importData) throws IOException;

	/**
	 * Create a change set file reader that will extract the change sets from
	 * the given file. This method is used the {@link LogCacheMerger}.
	 * @param importData the import data.
	 * @return a change set file reader
	 * @throws IOException if an error occurs opening or reading from the file
	 */
	IChangeSetFileReader createChangeSetFileReader(File file) throws IOException;
    
	/**
	 * Create a change set file writer that writes to the given file.
	 * @param file the file to which the change set descriptions are written
	 * @return a writer that will write the change set descriptions
	 * @throws IOException if an error occurs opening the file
	 */
	ChangeSetFileWriter createChangeSetFileWriter(File file) throws IOException;
	
	/**
	 * Create a change set file writer with the same meta-data as the given reader. 
	 * This method is used the {@link LogCacheMerger}.
	 * @param importData the import data.
	 * @return a change set file reader
	 * @throws IOException if an error occurs opening or reading from the file
	 */
	IChangeSetFileWriter createChangeSetFileWriter(IChangeSetFileReader logReader);
	
	/**
	 * Return the importer used by the import process to take data from a non-Jazz
	 * repository and create a change set archive for import
	 * @return the import
	 */
    IChangeSetArchiveCreator createChangeSetArchiveCreator();
    
    /**
     * Create an import participant that will be notified at certain points during
     * the import process.
     * @param workspace the target workspace
     * @param component the target component
     * @param importData the import data
     * @return an import participant
     */
    IImportParticipant createImportParticipant(IWorkspaceConnection workspace, IComponent component,
                        IImportData importData);
    
    /** 
     * Create and return a change set archive importer that will import the archive
     * references by the given import data into the workspace and component.
     * @param importData the import data
     * @param workspace the target workspace
     * @param component the target component
     * @return an importer
     */
    IChangeSetArchiveImporter createArchiveImporter(IImportData importData, IWorkspaceConnection workspace, IComponent component);
    
}
