/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashSet;
import java.util.Set;

import com.ibm.team.repository.client.IStatistics;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.importz.IImportChangeSet;
import com.ibm.team.scm.client.importz.internal.utils.DebugUtils;

/**
 * Track information about the import
 */
public class ChangeArchiveStatistics extends AbstractChangeArchiveStatistics implements IChangeArchiveStatistics {

	private long importStartTime;
	private long commitStartTime;
	private PrintStream debugStream = System.out;
	private final IStatistics statistics;
	private final IWorkspaceConnection connection;
	private long phaseStart;
	private Set<String> phases = new HashSet<String>();

	public ChangeArchiveStatistics(IWorkspaceConnection connection) {
		this.connection = connection;
		this.statistics = connection.teamRepository().statistics();
		importStartTime = System.currentTimeMillis();
		if (DebugUtils.DEBUG_TIMINGS) {
			String filename = DebugUtils.DEBUG_TIMINGS_LOG_FILE;
			if (filename != null && filename.length() > 0) {
				try {
					File file = new File(filename);
					debugStream = new PrintStream(new FileOutputStream(file));
				} catch (Exception e) {
					// Ignore
				}
			}
		}
	}
	
	public void startCommit(IImportChangeSet importChangeSet, long time, boolean preflight) {
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_READ_TIME, time);
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_COUNT, 1);
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_COUNT, importChangeSet.getChanges().size());
	    super.startCommit(importChangeSet, time, preflight);
	    if (!preflight) {
			commitStartTime = System.currentTimeMillis();
			if (DebugUtils.DEBUG_TIMINGS) {
				debugStream.println(commitStartTime+ ": Starting to import " + importChangeSet.getChanges().size() + " changes of '" + importChangeSet.asCommentMessage(true, true) + "'"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			}
	    }
	}
	
	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#toDescriptionString()
	 */
	public String toDescriptionString() {
		StringBuffer buffer = new StringBuffer();
		buffer.append("Total change sets: " + getChangeSetCount()); //$NON-NLS-1$
		buffer.append('\n');
		buffer.append("Total changes: " + getChangesCount()); //$NON-NLS-1$
		buffer.append('\n');
		buffer.append("Total contributors: " //$NON-NLS-1$
				+ getContributorsCount());
		buffer.append('\n');
		buffer.append("Largest change set contains " //$NON-NLS-1$
				+ getLargestChangeSetSize() + " changes"); //$NON-NLS-1$
		buffer.append('\n');
		long timeSpan = timeSpan();
		long diffHours = timeSpan / (60 * 60 * 1000);
		long diffDays = timeSpan / (24 * 60 * 60 * 1000);
		long diffYears = diffDays / 365;
		long diffYearsRemain = Math.round(((double)(diffDays % 365) / 365.0) * 100);
		buffer.append("Changes happened over " + diffYears + "." //$NON-NLS-1$ //$NON-NLS-2$
				+ diffYearsRemain + " years or " + diffDays + " days or " //$NON-NLS-1$ //$NON-NLS-2$
				+ diffHours + " hours"); //$NON-NLS-1$
		buffer.append('\n');
		return buffer.toString();
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#getChangeSetCount()
	 */
	public int getChangeSetCount() {
		return (int)statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_COUNT);
	}

	Object getContext() {
		return connection.getResolvedWorkspace().getItemId();
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#getChangesCount()
	 */
	public int getChangesCount() {
		return  (int)statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_COUNT);
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#changesCollected(long)
	 */
	public void changesCollected(long time) {
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), COLLECT_CHANGES_TIME, time);
		if (DebugUtils.DEBUG_TIMINGS) {
			debugStream.println(">Preparing changes took " + time); //$NON-NLS-1$
		}
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#changesCommited(long)
	 */
	public void changesCommited(long time) {
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), TOTAL_COMMIT_CHANGES_TIME, time);
		if (DebugUtils.DEBUG_TIMINGS) {
			debugStream.println(">Commiting changes took " + time); //$NON-NLS-1$
		}
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#endCommit(com.ibm.team.scm.client.importz.IImportChangeSet)
	 */
	public void endCommit(IImportChangeSet importChangeSet) {
		long end = System.currentTimeMillis();
		long commitTime = end - commitStartTime;
		statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_COMMIT_TIME, commitTime);
		if (DebugUtils.DEBUG_TIMINGS) {
			debugStream.println(end + ": Import of change took " + (commitTime)); //$NON-NLS-1$
			debugStream.println();
			if (getChangeSetCount() % 100 == 0) {
				printAverages(System.out);
			}
		}
	}

	private void printAverages(PrintStream out) {
		long totalTime = statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_COMMIT_TIME);
		long baselineCount = statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), BASELINE_COUNT);
		out.println("Change sets " + getChangeSetCount() + " changes " + getChangesCount() + " baselines " + baselineCount); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		if (getChangeSetCount() > 0 && getChangesCount() > 0) {
			out.println("Average time for each change set: " + totalTime / getChangeSetCount()); //$NON-NLS-1$
			out.println("Average time for each change: " + totalTime / getChangesCount()); //$NON-NLS-1$
			out.println("Average time reading change set: "  //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_READ_TIME) / getChangeSetCount() + "/" //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), CHANGE_SET_READ_TIME) / getChangesCount());
			out.println("Average time collecting changes: "  //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), COLLECT_CHANGES_TIME) / getChangeSetCount() + "/" //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), COLLECT_CHANGES_TIME) / getChangesCount());
			out.println("Average time comitting changes: "  //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), TOTAL_COMMIT_CHANGES_TIME) / getChangeSetCount() + "/" //$NON-NLS-1$
					+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), TOTAL_COMMIT_CHANGES_TIME) / getChangesCount());
			for (String phase : phases) {
				out.println("Average time '" + phase + "': " //$NON-NLS-1$ //$NON-NLS-2$
						+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), phase) / getChangeSetCount() + "/" //$NON-NLS-1$
						+ statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), phase) / getChangesCount());
			}
		}
		if (baselineCount > 0) {
			out.println("Average time per baseline: " + statistics.getStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), BASELINE_TIME) / baselineCount); //$NON-NLS-1$
		}

	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#close()
	 */
	public void close() {
		if (DebugUtils.DEBUG_TIMINGS) {
			long duration = System.currentTimeMillis() - importStartTime;
			long seconds = duration / 1000;
			String totalTime =  (seconds / 60) + "min " + (seconds % 60) + "sec"; //$NON-NLS-1$ //$NON-NLS-2$
			printAverages(System.out);
			debugStream.println(toDescriptionString());
			debugStream.println("Total import time: " + totalTime); //$NON-NLS-1$
			printAverages(debugStream);
			if (debugStream != System.out) {
				System.out.println(toDescriptionString());
				System.out.println("Total import time: " + totalTime); //$NON-NLS-1$
				debugStream.close();
			}
		}
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#startPhase(java.lang.String)
	 */
	public void startPhase(String statisticsName) {
		phases.add(statisticsName);
		phaseStart = System.currentTimeMillis();
	}

	/* (non-Javadoc)
	 * @see com.ibm.team.scm.client.importz.IChangeArchiveStatistics#endPhase(java.lang.String)
	 */
	public void endPhase(String statisticsName) {
		if (phaseStart != -1) {
			statistics.incStatisticValue(STATISTICS_CONTEXT_TYPE, getContext(), statisticsName, System.currentTimeMillis() - phaseStart);
			phaseStart = -1;
		}
	}
}
