/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2015. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.ListenerList;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.core.runtime.SafeRunner;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.content.IContentDescription;
import org.eclipse.core.runtime.content.IContentTypeManager;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.filesystem.common.internal.FileContent;
import com.ibm.team.filesystem.common.util.LineDelimiterUtil;
import com.ibm.team.filesystem.common.util.LineDelimiterVerifier;
import com.ibm.team.filesystem.rcp.ui.internal.util.PathUtils;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.transport.ServiceMethodInvocationError;
import com.ibm.team.repository.common.utils.ReaderToInputStream;
import com.ibm.team.repository.rcp.core.utils.StatusUtil;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.IWorkspaceConnection.IConfigurationOp;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.client.content.AbstractVersionedContentManagerInputStreamProvider;
import com.ibm.team.scm.client.importz.IChangeSetFileReader;
import com.ibm.team.scm.client.importz.IChangeSetFileWriter;
import com.ibm.team.scm.client.importz.IFileProperties;
import com.ibm.team.scm.client.importz.IImportChange;
import com.ibm.team.scm.client.importz.IImportChangeSet;
import com.ibm.team.scm.client.importz.IImportData;
import com.ibm.team.scm.client.importz.IImportParticipant;
import com.ibm.team.scm.client.importz.IMigrationFactory;
import com.ibm.team.scm.client.importz.internal.ui.ImportConfiguration;
import com.ibm.team.scm.client.importz.internal.utils.DebugUtils;
import com.ibm.team.scm.client.importz.internal.utils.FileTree;
import com.ibm.team.scm.client.importz.internal.utils.FileTreeWithEmptyFolderTracking;
import com.ibm.team.scm.client.importz.internal.utils.IFileTree;
import com.ibm.team.scm.client.importz.spi.ImportChange;
import com.ibm.team.scm.client.importz.spi.ImportChangeSet;
import com.ibm.team.scm.client.interop.IImportChangeSetResult;
import com.ibm.team.scm.client.interop.IWorkspaceConnectionInteropAdapter;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;
import com.ibm.team.scm.common.dto.IUpdateReport;

/**
 * <p>
 * The ChangeSetArchiveImporter performs the general import process (independent of the
 * used CM system).
 * </p>
 * 
 * <p>
 * Note: Before the ChangeSetArchiveImporter is executed the necessary information must
 * be gathered inside the LogCache.
 * </p>
 */
public class ChangeSetArchiveImporter {
	
	private static final String ENCODING_PREFERENCES_FILE = "org.eclipse.core.resources.prefs"; //$NON-NLS-1$
	
	private static final String SETTINGS_FOLDER_NAME = ".settings"; //$NON-NLS-1$
	
	private static final String ENCODING_PROPERTY_PREFIX = "encoding/"; //$NON-NLS-1$
	
	private static final String PROJECT_ENCODING_PROPERTY = "<project>"; //$NON-NLS-1$

	public static final int TIMING_INTERVAL = 100;
    
    private ITeamRepository repository;
    private IWorkspaceConnection workspace;
    private IComponent component;

    private boolean preflight;
    private boolean directoriesSignificant;
    private IFileTree fileTree;

	private final IImportData importData;
	private final IMigrationFactory factory;
	private ListenerList importParticipants = new ListenerList();
	private IChangeArchiveStatistics stats;

	private Map<String, Map<String, String>> projectEncodings = new HashMap<String, Map<String,String>>();

	private Set<String> reconciledProjects = new HashSet<String>();
	
	private class CommitRecord {
		private Map<UUID, ItemsMapEntry> itemsMap = new HashMap<UUID, ItemsMapEntry>();
		private List<IImportChange> changes = new ArrayList<IImportChange>();
		private List<IVersionable> changedItems = new ArrayList<IVersionable>();
		private Set<String> pathsToReconcile;
		private Set<String> reconciledProjects;
		private Set<String> reconciledFolders;

		public ArrayList<IConfigurationOp> asConfigOps() {
			ArrayList<IConfigurationOp> configOps = new ArrayList<IConfigurationOp>(itemsMap.size());
			for (Iterator<ItemsMapEntry> i = itemsMap.values().iterator(); i.hasNext();) {
				ItemsMapEntry entry = (ItemsMapEntry)i.next();
				configOps.add(entry.configOp);
			}
			return configOps;
		}
		
		public void updateItemsMapsForSave(IVersionable newVersionable) {
			IConfigurationOp op = workspace.configurationOpFactory().save(newVersionable);
			updateItemsMap(newVersionable, op);
		}
		
		public void updateItemsMapForDelete(IVersionable itemToDelete, boolean deleteSubtreeForFolder) {
			IConfigurationOp op;
			if (itemToDelete instanceof IFolder && deleteSubtreeForFolder) {
				IFolderHandle folder = (IFolder) itemToDelete;
				op = workspace.configurationOpFactory().deleteSubtree(folder);
			} else {
				op = workspace.configurationOpFactory().delete(itemToDelete);
			}
			updateItemsMap(itemToDelete, op);
		}

		private void updateItemsMap(IVersionable item, IConfigurationOp op) {
			ItemsMapEntry entry = new ItemsMapEntry(op, item);
			itemsMap.put(item.getItemId(), entry);
		}
		
		public boolean hasItemChanges() {
			return !itemsMap.isEmpty();
		}

		public void recordLeafChange(IImportChange change, IVersionable item) {
			changes.add(change);
			changedItems.add(item);
		}
		
		public List<IVersionable> getAllChangedItems() {
			final List<IVersionable> fsItemsList = new ArrayList<IVersionable>(itemsMap.size());
			for (Iterator<ItemsMapEntry> i = itemsMap.values().iterator(); i.hasNext();) {
				ItemsMapEntry entry = (ItemsMapEntry)i.next();
				fsItemsList.add(entry.versionable);
			}
			return fsItemsList;
		}

		public List<IImportChange> getChanges() {
			return changes;
		}

		public List<IVersionable> getLeafsItemChanges() {
			return changedItems;
		}

		public void addPathsRequiringReconciliation(String parentPath, Collection<String> names) {
			if (pathsToReconcile == null) {
				pathsToReconcile = new HashSet<String>();
			}
			if (!directoriesSignificant) {
				if (reconciledFolders == null) {
					reconciledFolders = new HashSet<String>();
				}
				reconciledFolders.add(PathUtils.normalizePath(parentPath, false, false));
			}
			if (names != null) {
				for (String name : names) {
					pathsToReconcile.add(PathUtils.normalizePath(PathUtils.appendPath(parentPath, name), false, false));
				}
			}
		}

		public void pathReconciled(String path) {
			if (pathsToReconcile != null) {
				pathsToReconcile.remove(PathUtils.normalizePath(path, false, false));
			}
		}
		
		public Set<String> getPathsToReconcile() {
			return pathsToReconcile;
		}

		public void addReconciledProject(String projectName) {
			if (reconciledProjects == null) {
				reconciledProjects = new HashSet<String>();
			}
			reconciledProjects.add(projectName);
		}
		
		public Set<String> getReconciledProjects() {
			return reconciledProjects;
		}
		
		public boolean isFolderReconciled(String folderPath) {
			if (reconciledFolders == null) {
				return false;
			}
			String normalizedPath = PathUtils.normalizePath(folderPath, false, false);
			return reconciledFolders.contains(normalizedPath);
 		}

		public void folderReconciled(String parentPath) {
			pathReconciled(parentPath);
			if (!directoriesSignificant) {
				if (reconciledFolders == null) {
					reconciledFolders = new HashSet<String>();
				}
				reconciledFolders.add(PathUtils.normalizePath(parentPath, false, false));
			}
		}
	}
	
	private static class ItemsMapEntry {
		private final IConfigurationOp configOp;
		private final IVersionable versionable;
		
		private ItemsMapEntry(IConfigurationOp configOp, IVersionable versionable) {
			this.configOp = configOp;
			this.versionable = versionable;
		}
	}

	public ChangeSetArchiveImporter(IMigrationFactory factory, IImportData importData) {
		Assert.isNotNull(importData);
		Assert.isNotNull(factory);
		this.factory = factory;
		this.importData = importData;
	}

	/**
	 * <p>
	 * Executes the import with a given file containing all necessary
	 * information. The file must be created by using the same MigrationFactory.
	 * <p>
	 * <p>
	 * Thus, merely the following step will be performed:
	 * </p>
	 * <ol>
	 * <li>Create the necessary new items (IVersionable's, IChangeSet's,
	 * IChange's).</li>
	 * </ol>
	 * 
	 * @param workspace
	 *            the current workspace, must not be null
	 * @param component
	 *            the component, must not be null
	 * @param factory
	 *            the log file reader, must not be null
	 * @param preflight
	 *            set to true if this is a dry run, false if changes are to actually
	 *            be committed
	 * @param monitor
	 *            the monitor, must not be null
	 * @throws IOException
	 *            if any error occurs
	 */
	public void importChangeSetArchive(IWorkspaceConnection workspace,
			IComponent component,
			boolean preflight, IProgressMonitor monitor)
			throws TeamRepositoryException {
		
		checkAndInitArgs(workspace, component, factory, preflight);
		
		// Start gathering statistics
        if (workspace == null) {
        	stats = new PreFlightChangeArchiveStatistics();
        } else {
        	stats = new ChangeArchiveStatistics(workspace);
        }
        
        // Step 2: Import into team repository
		SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.ChangeSetArchiveImporter_0, 1000);
        
        fileTree = new FileTree(workspace, component);
        
        SubMonitor subProgress = progress.newChild(990);
        subProgress.beginTask(SCMImportMessages.ChangeSetArchiveImporter_1, importData.getArchiveSize() - 1);

        // perform import
        IChangeSetFileReader reader = null;
        try {
        	reader = factory.createChangeSetFileReader(importData);
	        this.directoriesSignificant = reader.getDirectoriesSignificant();
	        if (!directoriesSignificant) {
	        	// Wrap the file tree so we can auto-delete empty folders
	        	fileTree = new FileTreeWithEmptyFolderTracking(fileTree);
	        }
	        IImportChangeSet importChangeSet;
	        while ((importChangeSet = readChangeSet(reader, progress)) != null) {
	        	notifyPrecommit(importChangeSet, subProgress.newChild(1));
	        	subProgress.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_2, getCommentForDisplay(importChangeSet)));
	        	if (monitor.isCanceled())
	        		throw new OperationCanceledException();
	            
	            if (!preflight) {
	                try {
	                    List<IImportChange> changes = importChangeSet.getChanges();
	                    updateEncodings(changes, subProgress.newChild(1));
						commitChanges(importChangeSet, changes,
	                            subProgress.newChild(changes.size()));
			            stats.endCommit(importChangeSet);
	                } catch (ImportException e) {
	                	// We don't need to wrap import exceptions
	                	debug(importChangeSet);
	                	throw e;
	                } catch (TeamRepositoryException tre) {
	                	debug(importChangeSet);
	                	String comment = getCommentForDisplay(importChangeSet);
	                    String errorMessage = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_4, comment, tre.getMessage());
	                    throw new TeamRepositoryException(repository, errorMessage, tre);
	                } catch (ServiceMethodInvocationError smie) {
	                	debug(importChangeSet);
	                	String comment = getCommentForDisplay(importChangeSet);
	                    String errorMessage = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_6, comment);
	                    throw new RuntimeException(errorMessage, smie);
	                } catch (OperationCanceledException e) {
	                	throw e;
	                } catch (RuntimeException e) {
	                	debug(importChangeSet);
	                	String comment = getCommentForDisplay(importChangeSet);
	                    String errorMessage = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_40, comment);
	                    throw new TeamRepositoryException(repository, errorMessage, e);
	                }
	            }
	            
	            if (monitor.isCanceled()) throw new OperationCanceledException();
	        }
	        notifyImportComplete(subProgress.newChild(1));
        } catch (IOException e) {
			throw new TeamRepositoryException(SCMImportMessages.ChangeSetArchiveImporter_8, e);
		} finally {
        	if (stats != null) {
        		stats.close();
        	}
        	if (reader != null)
        		reader.close();
        	monitor.done();
        }
	}

	private void updateEncodings(List<IImportChange> changes, SubMonitor monitor) {
		monitor.setWorkRemaining(changes.size());
		for (IImportChange change : changes) {
			if (isEncodingPropertiesFileChange(change)) {
				updateEncodings(change, monitor.newChild(1));
			}
		}
	}

	private void updateEncodings(IImportChange change, IProgressMonitor monitor) {
		String path = change.getAfterPath();
		if (path == null)
			path = change.getBeforePath();
		String name = PathUtils.getBaseName(path);
		if (name.equals(ENCODING_PREFERENCES_FILE)) {
			// This is a change to the encoding file itself
			String projectPath = PathUtils.getParentFolderPath(PathUtils.getParentFolderPath(path));
			if (change.getKind() == IImportChange.DELETION) {
				// Handle the deletion
				clearEncodings(projectPath);
			} else if (change.getKind() == IImportChange.ADDITION 
					|| isContentChange(change)) {
				// Handle an addition or content change
				try {
					Properties properties = new Properties();
					InputStream contents = change.getContents(importData, monitor);
					if (contents != null) {
						try {
							properties.load(contents);
						} finally {
							try {
								contents.close();
							} catch (IOException e) {
								// Ignore
							}
						}
					}
					updateEncodings(projectPath, properties);
				} catch (IOException e) {
					StatusUtil.log(this, NLS.bind("An error occurred reading properties file {0}", path), e); //$NON-NLS-1$
				}
			}
		} else if (name.equals(SETTINGS_FOLDER_NAME)) {
			if (change.getKind() == IImportChange.DELETION) {
				// The .settings folder has been deleted so clear the encodings
				String projectPath = PathUtils.getParentFolderPath(path);
				clearEncodings(projectPath);
			}
		} else {
			String projectPath = PathUtils.normalizePath(path, false, false);
			if (change.getKind() == IImportChange.DELETION) {
				// The project has been deleted so clear the encodings
				clearEncodings(projectPath);
			} else if (change.getKind() == IImportChange.MODIFICATION
					&& (change.getFlags() & IImportChange.MOVE) > 0) {
				// The project was moved so move the encoding properties
				Map<String, String> properties = projectEncodings.get(projectPath);
				if (properties != null) {
					projectEncodings.remove(projectPath);
					projectEncodings.put(PathUtils.normalizePath(change.getAfterPath(), false, false), properties);
				}
			}
			
		}
	}

	private void updateEncodings(String projectPath, Properties properties) {
		clearEncodings(projectPath);
		Set<Entry<Object, Object>> entries = properties.entrySet();
		for (Entry<Object, Object> entry : entries) {
			String key = (String)entry.getKey();
			if (key.startsWith(ENCODING_PROPERTY_PREFIX)) {
				String path = key.substring(ENCODING_PROPERTY_PREFIX.length());
				if (path.equals(PROJECT_ENCODING_PROPERTY)) {
					setEncoding(projectPath, "", (String)entry.getValue()); //$NON-NLS-1$
				} else {
					setEncoding(projectPath, path, (String)entry.getValue());
				}
			}
		}
	}

	private String getEncoding(String path) {
		String normalizePath = PathUtils.normalizePath(path, false, false);
		for (String projectPath : projectEncodings.keySet()) {
			if (PathUtils.isParentFolder(projectPath, normalizePath)) {
				String relativePath = PathUtils.getRelativePath(projectPath, normalizePath);
				Map<String, String> encodings = projectEncodings.get(projectPath);
				String match = null;
				for (String projectRelativePath : encodings.keySet()) {
					if (projectRelativePath.equals(relativePath)) {
						match = projectRelativePath;
						break;
					} else if (PathUtils.isParentFolder(projectRelativePath, relativePath)) {
						if (match == null || projectRelativePath.length() > match.length()) {
							match = projectRelativePath;
						}
					}
				}
				if (match != null) {
					String encoding = encodings.get(match);
					if (encoding != null && Charset.isSupported(encoding))
						return encoding;
					// A match was found but the encoding is no good so stop searching
					return null;
				}
			}
		}
		return null;
	}
	
	private void setEncoding(String projectPath, String subpath, String encoding) {
		String normalizePath = PathUtils.normalizePath(projectPath, false, false);
		Map<String, String> encodings = projectEncodings.get(normalizePath);
		if (encodings == null) {
			encodings = new HashMap<String, String>();
			projectEncodings.put(normalizePath, encodings);
		}
		encodings.put(PathUtils.normalizePath(subpath, false, false), encoding);
	}

	private void clearEncodings(String projectPath) {
		String normalizePath = PathUtils.normalizePath(projectPath, false, false);
		projectEncodings.remove(normalizePath);
	}

	private boolean isEncodingPropertiesFileChange(IImportChange change) {
		String path = change.getAfterPath();
		if (path == null)
			path = change.getBeforePath();
		String name = PathUtils.getBaseName(path);
		if (name.equals(ENCODING_PREFERENCES_FILE)) {
			// Return true if we are changing the resource settings file
			String parentPath = PathUtils.getParentFolderPath(path);
			if (!parentPath.equals(PathUtils.PATH_SEPARATOR)) {
				String settingsFolder = PathUtils.getBaseName(parentPath);
				if (settingsFolder.equals(SETTINGS_FOLDER_NAME)) {
					return true;
				}
			}
		} else if (name.equals(SETTINGS_FOLDER_NAME)) {
			return true;
		} else {
			// Also return true if the operation is being performed on the project 
			// containing the settings
			String normalizePath = PathUtils.normalizePath(path, false, false);
			return projectEncodings.get(normalizePath) != null;
		}
		return false;
	}

	private String getCommentForDisplay(IImportChangeSet importChangeSet) {
		String comment = importChangeSet.asCommentMessage(true, true);
		if (comment.length() > 128) {
			comment = comment.substring(0, 125) + "..."; //$NON-NLS-1$
		}
		return comment;
	}

	private IImportChangeSet readChangeSet(IChangeSetFileReader reader, IProgressMonitor monitor)
			throws TeamRepositoryException {
		try {
			long start = System.currentTimeMillis();
			IImportChangeSet changeSet = reader.next(monitor);
			if (changeSet != null) {
				stats.startCommit(changeSet, System.currentTimeMillis() - start, preflight);
			}
			return changeSet;
		} catch (IOException e) {
			throw new TeamRepositoryException(SCMImportMessages.ChangeSetArchiveImporter_9, e);
		}
	}
	
	private void debug(IImportChangeSet changeSet) {
		if (!DebugUtils.DEBUG_IMPORT)
			return;
		IChangeSetFileWriter writer = factory.createChangeSetFileWriter((IChangeSetFileReader)null);
		try {
			System.out.println("Dumping Change set for import failure"); //$NON-NLS-1$
			writer.startWrite(System.out);
			writer.writeElement(changeSet);
			writer.closeWrite();
			System.out.println("Change set dump finished"); //$NON-NLS-1$
		} catch (IOException e) {
			// Ignore
		}
	}

	private void checkAndInitArgs(IWorkspaceConnection workspace, IComponent component,
			IMigrationFactory factory, boolean preflight)
			throws TeamRepositoryException {
		Assert.isNotNull(factory);
		this.preflight = preflight;
		if (!preflight) {
            Assert.isNotNull(workspace);
            Assert.isNotNull(component);
            this.component = component;
            this.workspace = workspace;
			this.repository = workspace.teamRepository();
		}
	}

	/* ----------------- Import resources ----------------- */

	private void commitChanges(final IImportChangeSet importChangeSet,
			List<IImportChange> changes, IProgressMonitor monitor) throws TeamRepositoryException {
		try {
			SubMonitor progress = SubMonitor.convert(monitor, 100);
			progress.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_11, getCommentForDisplay(importChangeSet)));
			fileTree.prepareForChanges(changes, importData.isImportOverExisting(), directoriesSignificant, progress.newChild(10));
	        CommitRecord record = collectChanges(importChangeSet, changes, progress.newChild(45));
	        commitChangesFromMap(importChangeSet, changes, record, progress.newChild(45));
		} finally {
			monitor.done();
		}
    }
    
    private CommitRecord collectChanges(IImportChangeSet importChangeSet, List<IImportChange> changes, IProgressMonitor monitor)
            throws TeamRepositoryException {
    	long startTime = System.currentTimeMillis();
    	try {
	    	SubMonitor progress = SubMonitor.convert(monitor, changes.size());
	    	CommitRecord record = new CommitRecord();
	        Date date = importChangeSet.getDate();
	        for (IImportChange change : changes) {
	        	collectChange(change, date, record, progress.newChild(1));
	        }
	        if (importData.isImportOverExisting()) {
	        	deleteUnreconciledItems(record, ""); //$NON-NLS-1$
	        	Set<String> projects = record.getReconciledProjects();
	        	if (projects != null) {
	        		reconciledProjects.addAll(projects);
	        	}
	        }
	        deleteEmptyFoldersIfRequired(record);
	        return record;
    	} finally {
    		stats.changesCollected(System.currentTimeMillis() - startTime);
    		monitor.done();
    	}
    }

	private void deleteUnreconciledItems(CommitRecord record, String revision) throws TeamRepositoryException {
		Set<String> paths = record.getPathsToReconcile();
		if (paths != null) {
			for (String path : paths) {
				IVersionable item = fileTree.getItem(path);
				if (item != null) {
	       			deleteVersionable(path, record);
	    			record.recordLeafChange(new ImportChange(IImportChange.DELETION, revision, path, null, null, null), item);
				}	
			}
		}
	}

	private void collectChange(IImportChange change, Date date, CommitRecord record,
			IProgressMonitor monitor) throws TeamRepositoryException {
		IVersionable item;
		if (change.getKind() == IImportChange.DELETION) {
			String path = change.getBeforePath();
			monitor.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_33, path));
			item = deleteVersionable(path, record);
		} else if (change.getKind() == IImportChange.ADDITION) {
			SubMonitor progress = SubMonitor.convert(monitor, 100);
	        if (importData.isImportOverExisting() && (!directoriesSignificant || !isProjectForPathReconciled(change.getAfterPath()))) {
	        	// We are importing into an existing component so we need to check to see if the 
	        	// target already exists and deal with it appropriately
				if (reconcileAdditionWithExisting(change, date, record, progress.newChild(50))) {
					record.addReconciledProject(PathUtils.getFirstSegment(change.getAfterPath()));
					return;
				}
	        } else {
	        	progress.setWorkRemaining(50);
	        }
	        item = performCreate(change, date, record, progress.newChild(50));
		} else if (change.getKind() == IImportChange.MODIFICATION) {
			String afterPath = change.getAfterPath();
        	if (importData.isImportOverExisting() && (change.getFlags() & IImportChange.MOVE) != 0 && !isProjectForPathReconciled(afterPath)) {
        		// For a move, delete the target before moving
        		IVersionableHandle targetItemHandle = fileTree.getItemHandle(afterPath);
        		if (targetItemHandle != null) {
        			// Delete the item and record the change
        			IVersionable targetItem = fileTree.getItem(afterPath);
        			deleteVersionable(afterPath, record);
        			record.recordLeafChange(new ImportChange(IImportChange.DELETION, change.getRevision(), afterPath, null, null, null), targetItem);
        			record.addReconciledProject(PathUtils.getFirstSegment(change.getAfterPath()));
        		}
        	}
			item = performModification(change, date, record, monitor);
		} else {
			throw new IllegalArgumentException(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_16, change.getItemType()));
		}
		record.recordLeafChange(change, item);
	}

	private boolean isProjectForPathReconciled(String path) {
		if (reconciledProjects == null || reconciledProjects.isEmpty()) {
			return false;
		}
		return reconciledProjects.contains(PathUtils.getFirstSegment(path));
	}

	private boolean reconcileAdditionWithExisting(IImportChange change, Date date, CommitRecord record, IProgressMonitor monitor) throws TeamRepositoryException {
		ensureParentsReconciled(change.getAfterPath(), record);
		String afterPath = change.getAfterPath();
		IVersionableHandle targetItemHandle = fileTree.getItemHandle(afterPath);
		if (targetItemHandle != null) {
			// There is a target item there already so we need to reconcile that with the current change
			// Update the record to indicate that this path was reconciled
			record.pathReconciled(afterPath);
			IVersionable targetItem = fileTree.getItem(afterPath);
			if (change.getItemType() != getItemType(targetItem)) {
				// Delete the target item since it's type differs from the item being added
				deleteVersionable(afterPath, record);
    			record.recordLeafChange(new ImportChange(IImportChange.DELETION, change.getRevision(), afterPath, null, null, null), targetItem);
    			return false; // Indicate that the item should be added
			}
			if (change.getItemType() == IImportChange.FILE) {
				// compare the new file state with the existing state and only perform a change
				// if they differ
				SubMonitor progress = SubMonitor.convert(monitor, 100);
				IFileItem fileItem = (IFileItem)targetItem;
				if (contentDiffers(fileItem, change, progress.newChild(50))) {
					IImportChange modChange = change.asModificationForReconciliation();
					performModification(modChange, date, record, progress.newChild(50));
					record.recordLeafChange(modChange, targetItem);
				}
			} else if (change.getItemType() == IImportChange.FOLDER) {
				// Record the children of the folder in order to ensure any that are not
				// included in the change set get deleted
				Collection<String> names = fileTree.getChildNames(afterPath);
				record.addPathsRequiringReconciliation(afterPath, names);
			}
			return true;
		}
		return false;
	}

	private void ensureParentsReconciled(String childPath, CommitRecord record) {
		if (directoriesSignificant) {
			// If directories are significant, we would have an add for the folder
			// before any children are processed
			return;
		}
		String parentPath = PathUtils.getParentFolderPath(childPath);
		if (PathUtils.isEmptyPath(parentPath)) {
			// Don't reconcile the root since we don't want to reconcile a top-level
			// folder unless the change set contains changes for that folder
			return;
		}
		if (!record.isFolderReconciled(parentPath)) {
			ensureParentsReconciled(parentPath, record);
			IVersionableHandle targetItemHandle = fileTree.getItemHandle(parentPath);
			if (targetItemHandle != null) {
				// There is a target item there already so we need to reconcile that with the current change
				// Update the record to indicate that this path was reconciled
				record.addPathsRequiringReconciliation(parentPath, null);
			}
			record.folderReconciled(parentPath);
			record.addReconciledProject(PathUtils.getFirstSegment(parentPath));
		}
	}

	private boolean contentDiffers(IFileItem fileItem, IImportChange change,
			SubMonitor progress) throws TeamRepositoryException {
		IFileContent content = fileItem.getContent();
		IFileProperties properties = change.getFileProperties();
		if (properties.isExecutable() == fileItem.isExecutable()
				&& properties.getContentType().equals(fileItem.getContentType())
				&& properties.getLineDelimiter().equals(content.getLineDelimiter())
				&& !properties.isText() || isCharacterEncodingEqual(change, content, progress)) {
			// All the attributes of the content match so let's test the content itself
			InputStream newContents = null;
			InputStream oldContents = null;
			try {
				progress.setWorkRemaining(10);
				newContents = change.getContents(importData, progress.newChild(1));
				IFileItemHandle contentState;
				if (!fileItem.isWorkingCopy()) {
					contentState = fileItem;
				} else {
					//we uploaded the content, committed and kept the working copy
					//original containing state will have our state id
					contentState = (IFileItemHandle)content.getOriginalContainingState();
				}
				if (contentState != null) {
					oldContents = retrieveContent(progress.newChild(9), contentState, content);
					return !streamsEqual(oldContents, newContents);
				}
			} catch (IOException e) {
				throw new TeamRepositoryException(e);
			} finally {		
				try {
					if (newContents != null) {
						newContents.close();
					}
				} catch (IOException e) {
					// Ignore
				} finally {
					try {
						if (oldContents != null) {
							oldContents.close();
						}
					} catch (IOException e) {
						// Ignore
					}
				}
			}
		}
		// If we get here, the content differs in some way.
		return true;
	}

	private boolean isCharacterEncodingEqual(IImportChange change,
			IFileContent content, SubMonitor progress) {
		String encoding = content.getCharacterEncoding();
		String newEncoding = getEncoding(change, null, progress);
		if (encoding == null) {
			return newEncoding == null;
		}
		if (newEncoding == null) {
			return false;
		}
		return encoding.equals(newEncoding);
	}

	private InputStream retrieveContent(SubMonitor progress, IFileItemHandle item,
			IFileContent content) throws TeamRepositoryException {
		if (content.getLineDelimiter().equals(FileLineDelimiter.LINE_DELIMITER_PLATFORM)) {
			// For PLATFORM line delimiters, we expect the contents from the archive to have
			// a LF delimiter (i.e. the archive needs to be platform independent)
			return FileSystemCore.getContentManager(repository).retrieveContentStream(item,
					content, content.getCharacterEncoding(), FileLineDelimiter.LINE_DELIMITER_LF, progress);
		}
		return FileSystemCore.getContentManager(repository).retrieveContentStream(item, content, progress);
	}
	
	private boolean streamsEqual(InputStream expected, InputStream actual) throws IOException {
		BufferedInputStream b1 = new BufferedInputStream(expected);
		BufferedInputStream b2 = new BufferedInputStream(actual);
		int r1 = b1.read();
		int r2 = b2.read();
		while (r1 == r2 && r1 != -1 && r2 != -1) {
			r1 = b1.read();
			r2 = b2.read();
		}
		if (r1 != -1 || r2 != -1) {
			return false;
		}
		return true;
	}

	private int getItemType(IVersionable targetItem) {
		if (targetItem instanceof IFolder) {
			return IImportChange.FOLDER;
		}
		if (targetItem instanceof IFileItem) {
			return IImportChange.FILE;
		}
		return IImportChange.UNKNOWN;
	}

	private IVersionable performCreate(IImportChange change, Date date, CommitRecord record,
			IProgressMonitor monitor) throws TeamRepositoryException {
		if (change.getItemType() == IImportChange.FILE) {
			return createFile(change, date, record, monitor);
		} else if (change.getItemType() == IImportChange.FOLDER) {
			return createFolder(change.getAfterPath(), record);
		}
		throw new IllegalArgumentException(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_3, change.getBeforePath()));
	}
    
    private void commitChangesFromMap(IImportChangeSet importChangeSet, List<IImportChange> changes, CommitRecord record,
            IProgressMonitor monitor) throws TeamRepositoryException {
    	SubMonitor progress = SubMonitor.convert(monitor, 5);
		progress.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_38, getCommentForDisplay(importChangeSet)));
		long startTime = System.currentTimeMillis();
		IChangeSetHandle changeSet = null;
		IUpdateReport report = null;
		if (record.hasItemChanges()) {
	        boolean committed = false;
	        try {
	        	IContributorHandle contributor = getContributor(workspace.teamRepository(), importChangeSet.getAuthor(), progress.newChild(1));
	        	String comment = ImportChangeSet.truncateComment(importChangeSet.asCommentMessage(contributor == null, false));
	        	IWorkspaceConnectionInteropAdapter op = (IWorkspaceConnectionInteropAdapter)workspace.getAdapter(IWorkspaceConnectionInteropAdapter.class);
	        	if (op == null) {
					changeSet = createActiveChangeSet(comment, workspace, component,
	                        progress.newChild(1));
	                report = commit(record, changeSet, progress.newChild(1));
	                closeChangeSet(workspace, changeSet, contributor, importChangeSet.getDate(), progress.newChild(1));       
	                committed = true;
	        	} else {
	        		if (contributor == null) {
	            		contributor = workspace.teamRepository().loggedInContributor();
	            	}
	        		IImportChangeSetResult result = op.importChangeSet(component, comment, record.asConfigOps(), importChangeSet.getDate().getTime(), contributor, progress.newChild(3));
	        		changeSet = result.getChangeSet();
	        		report = result.getUpdateReport();
	        		committed = true;
	        	}
	        } finally {
	            if (!committed && changeSet != null ) {
	                workspace.discardChangeSets(false, Collections.singletonList(changeSet),
	                        progress.newChild(0));
	            }
	        }
		}
		stats.changesCommited(System.currentTimeMillis() - startTime);
		
		notifyPostCommit(importChangeSet, report, changeSet, changes, record,
				progress.newChild(1));
        
        monitor.done();
	}
    
    private IContributorHandle getContributor(ITeamRepository teamRepository, String author, SubMonitor monitor) throws TeamRepositoryException {
    	if (importData == null) {
    		return null;
    	}
    	return importData.getContributor(teamRepository, author, monitor);
	}

	private void closeChangeSet(IWorkspaceConnection connection, IChangeSetHandle changeSet, IContributorHandle contributor, Date creationDate, SubMonitor monitor) throws TeamRepositoryException {
		stats.startPhase(IChangeArchiveStatistics.CLOSE_CHANGE_SET);
		if (contributor == null) {
    		contributor = connection.teamRepository().loggedInContributor();
    	}
    	connection.closeChangeSets(Collections.singletonList(changeSet), monitor);
		stats.endPhase(IChangeArchiveStatistics.CLOSE_CHANGE_SET);
    }
	
	private IUpdateReport commit(CommitRecord record, IChangeSetHandle changeset, IProgressMonitor monitor)
            throws TeamRepositoryException {
		try {
			stats.startPhase(IChangeArchiveStatistics.COMMIT_CHANGE_SET);
			ArrayList<IConfigurationOp> configOps = record.asConfigOps();
			IUpdateReport report = workspace.commit(changeset, configOps, monitor);
			return report;
		} finally {
			stats.endPhase(IChangeArchiveStatistics.COMMIT_CHANGE_SET);
		}
	}

	private void notifyPrecommit(final IImportChangeSet importChangeSet, SubMonitor monitor) {
		if (importParticipants != null && !importParticipants.isEmpty()) {
			Object[] participants = importParticipants.getListeners();
			final SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.ChangeSetArchiveImporter_17, participants.length);
			for (int i = 0; i < participants.length; i++) {
				final IImportParticipant participant = (IImportParticipant)participants[i];
				SafeRunner.run(new ISafeRunnable() {
					public void run() {
						participant.changeSetRead(workspace, component, importChangeSet, progress.newChild(1));
					}
					public void handleException(Throwable exception) {
						// Logged by safe runner
					}
				});
			}
		}
	}
	
	private void notifyImportComplete(SubMonitor monitor) {
		if (importParticipants != null && !importParticipants.isEmpty()) {
			Object[] participants = importParticipants.getListeners();
			final SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.ChangeSetArchiveImporter_18, participants.length);
			for (int i = 0; i < participants.length; i++) {
				final IImportParticipant participant = (IImportParticipant)participants[i];
				SafeRunner.run(new ISafeRunnable() {
					public void run() {
						participant.postImport(workspace, component, progress.newChild(1));
					}
					public void handleException(Throwable exception) {
						// Logged by safe runner
					}
				});
			}
		}
	}
	
	private void notifyPostCommit(final IImportChangeSet importChangeSet,
			final IUpdateReport report, final IChangeSetHandle committedChangeSet,
			List<IImportChange> changesCollection, final CommitRecord record, final IProgressMonitor monitor) {
		if (importParticipants != null && !importParticipants.isEmpty()) {
			final List fsItemsList = record.getAllChangedItems();
            
            //collections don't have a guaranteed iteration order, convert to a list
            final List changesList = record.getChanges();
            final List changeLeafsList = record.getLeafsItemChanges();
			
			Object[] participants = importParticipants
					.getListeners();
			final SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.ChangeSetArchiveImporter_20, participants.length);
			for (int i = 0; i < participants.length; i++) {
				final IImportParticipant participant = (IImportParticipant)participants[i];
				SafeRunner.run(new ISafeRunnable() {
					public void run() {
						IImportChange[] changes = new IImportChange[changesList.size()];
                        changesList.toArray(changes);
						IVersionable [] fsItems = new IVersionable[fsItemsList.size()];
                        fsItemsList.toArray(fsItems);
                        IVersionableHandle [] changeLeafs = new IVersionableHandle[changeLeafsList.size()];
                        changeLeafsList.toArray(changeLeafs);
						participant.postCommit(workspace, component, importChangeSet, committedChangeSet,
                                report, fsItems, changes, changeLeafs, progress.newChild(1));
					}
					public void handleException(Throwable exception) {
						exception.printStackTrace();
					}
				});
			}
		}
	}
	
	private IFolder createFolder(String path, CommitRecord record) throws TeamRepositoryException {
        ensureParentExists(path, record);
        IFolder folder = fileTree.createFolder(path);
        record.updateItemsMapsForSave(folder);
        return folder;
	}
	
	private IFileItem createFile(IImportChange change, Date timestamp, CommitRecord record, IProgressMonitor monitor)
            throws TeamRepositoryException {
		SubMonitor progress = SubMonitor.convert(monitor, 100);
        String path = change.getAfterPath();
		progress.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_26, path));
        
        ensureParentExists(path, record);
        IFileItem fileItem = fileTree.createFile(path);
		setContents(fileItem, change, timestamp, -1, null, progress.newChild(100));
		record.updateItemsMapsForSave(fileItem);
		return fileItem;
	}

	private void ensureParentExists(String path, CommitRecord record) throws TeamRepositoryException {
		String parentPath = PathUtils.getParentFolderPath(path);
		IVersionableHandle versionableHandle = fileTree.getItemHandle(parentPath);
		if (versionableHandle == null) {
			if (directoriesSignificant) {
				throw new IllegalStateException(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_5, parentPath));
			} else {
				ensureParentExists(parentPath, record);
				createFolder(parentPath, record);
			}
		}
	}

	private void setContents(final IFileItem fileItem, final IImportChange change,
			Date timestamp, int retry, String encodingOverride, IProgressMonitor monitor) throws TeamRepositoryException {
		if (retry == -1)
			retry = 1; // By default, retry once.
		AbstractVersionedContentManagerInputStreamProvider contents = null;
		final SubMonitor progress = SubMonitor.convert(monitor, 100);
		try {
			final String encoding = getEncoding(change, encodingOverride, progress.newChild(1));
			final IFileProperties properties = change.getFileProperties();
			final int tryNum = retry;
			contents = new AbstractVersionedContentManagerInputStreamProvider() {
				@Override
				public void dispose() throws IOException,
						TeamRepositoryException {
					// versioned content manager will close the stream
				}
				@Override
				public InputStream getInputStream(int flags)
						throws IOException, TeamRepositoryException {
					InputStream in = change.getContents(importData, progress.newChild(1));
					if (isLineDelimiterChange(fileItem, properties) && in == null) {
						// the content didn't change, only the properties changed
						final FileContent seed = (FileContent) fileItem.getContent();
						final FileLineDelimiter lineDelimiter = properties.getLineDelimiter();
						final InputStream seedContentStream = FileSystemCore.getContentManager(repository)
								.retrieveContentStream((IFileItemHandle) seed.getOriginalContainingState(), seed, progress.newChild(19));
						String seedEncoding = seed.getCharacterEncoding();
						if (seedEncoding == null) {
							// The encoding of the previous content was not known so don't try and guess it
							in = FileSystemCore.getContentManager(repository)
									.retrieveContentStream((IFileItemHandle) seed.getOriginalContainingState(), seed, progress.newChild(19));
						} else {
							LineDelimiterVerifier lineDelimiterConverter = LineDelimiterUtil.getNonVerifyingLineDelimiterConverter(
								seedContentStream, seedEncoding, lineDelimiter);
							in = new ReaderToInputStream(lineDelimiterConverter);
						}
					} else if (in == null) {
						// Not sure how we ended up here but let's use the content from the seed
						final FileContent seed = (FileContent) fileItem.getContent();
						in = FileSystemCore.getContentManager(repository)
								.retrieveContentStream((IFileItemHandle) seed.getOriginalContainingState(), seed, progress.newChild(19));
					}
					if (tryNum == 0 && properties.getLineDelimiter() != FileLineDelimiter.LINE_DELIMITER_NONE) {
						// A failure may be due to a line delimiter problem so lets makes sure on the last try
						in = new ReaderToInputStream(LineDelimiterUtil.getNonVerifyingLineDelimiterConverter(in, encoding, properties.getLineDelimiter()));
					}
					return in;
				}
				
				@Override
				public InputStream wrapInputStream(InputStream in)
						throws IOException, TeamRepositoryException {
					return in;
				}
			};
			
			setContents(fileItem, 
					contents, 
					properties.getLineDelimiter(), 
					properties.getContentType(), 
					encoding, 
					properties.isExecutable(), 
					timestamp, 
					progress.newChild(50));
		} catch (TeamRepositoryException e) {
			// The exception could be an intermittent failure or could be an encoding problem.
			// Try again with an encoding we know will work
			if (retry > 0 || encodingOverride == null) {
				String path = change.getAfterPath();
				String revision = change.getRevision();
				String message;
				String encoding = ImportConfiguration.DEFAULT_ENCODING;
				if (encodingOverride == null) {
					message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_39, new Object[] { revision, path, encoding });
				} else {
					message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_27, revision, path);
				}
				StatusUtil.log(this, message, e);
				setContents(fileItem, change, timestamp, encodingOverride == null ? retry : retry - 1, encoding, progress.newChild(40));
			} else {
				String path = change.getAfterPath();
				String revision = change.getRevision();
				String message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_28, revision, path);
				StatusUtil.log(this, message, e);
				throw new ImportException(message, e);
			}
		} catch (UnsupportedCharsetException e) {
			// It looks like the default encoding for the file is not supported so use the default
			if (retry > 0 || encodingOverride == null) {
				String path = change.getAfterPath();
				String revision = change.getRevision();
				String message;
				String encoding = ImportConfiguration.DEFAULT_ENCODING;
				if (encodingOverride == null) {
					message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_39, new Object[] { revision, path, encoding });
				} else {
					message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_27, revision, path);
				}
				StatusUtil.log(this, message, e);
				setContents(fileItem, change, timestamp, encodingOverride == null ? retry : retry - 1, encoding, progress.newChild(40));
			} else {
				String path = change.getAfterPath();
				String revision = change.getRevision();
				String message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_28, revision, path);
				StatusUtil.log(this, message, e);
				throw new ImportException(message, e);
			}
		}
	}

	private String getEncoding(IImportChange change, String encodingOverride, IProgressMonitor monitor) {
		IFileProperties properties = change.getFileProperties();
		String encoding = properties.getEncoding();
		if (encodingOverride != null) {
			encoding = encodingOverride;
		} else if (encoding == null && properties.isText()) {
			encoding = getCharset(change, monitor);
		}
		return encoding;
	}

	private String getCharset(IImportChange change, IProgressMonitor monitor) {
		// Try to determine the encoding using the content type manager
		InputStream contents = null;
		try {
			IContentTypeManager contentTypeManager = Platform.getContentTypeManager();
			if (contentTypeManager != null) {
				contents = change.getContents(importData, monitor);
				IContentDescription desc = contentTypeManager.getDescriptionFor(
						contents, 
						PathUtils.getBaseName(change.getAfterPath()), 
						new QualifiedName[] { IContentDescription.CHARSET });
				if (desc != null) {
					String charset = desc.getCharset();
					if (charset != null && Charset.isSupported(charset))
						return charset;
				}
			}
		} catch (IOException e) {
			// Ignore
		} finally {
			if (contents != null)
				try {
					contents.close();
				} catch (IOException e) {
					// Ignore
				}
		}
		// Try to get the encoding from the eclipse settings
		String eclipseEncoding = getEncoding(change.getAfterPath());
		if (eclipseEncoding != null) {
			return eclipseEncoding;
		}
		
		// If the content type manager doesn't know the encoding, return the default
		String workspaceEncoding = importData.getWorkspaceEncoding();
		if (workspaceEncoding != null) {
			return workspaceEncoding;
		}
		
		return ImportConfiguration.DEFAULT_ENCODING;
	}

	private void setContents(IFileItem fileItem, AbstractVersionedContentManagerInputStreamProvider contents, 
			FileLineDelimiter lineDelimeter, String contentType, String encoding, boolean executable,
			Date timestamp, IProgressMonitor monitor) throws TeamRepositoryException {
		IFileContent content = FileSystemCore.getContentManager(repository).storeContent(
                encoding, 
                lineDelimeter, contents,
                fileItem.getContent() == null ? null : fileItem.getContent().getHash(), monitor);
		fileItem.setContentType(contentType);
        fileItem.setContent(content);
        fileItem.setFileTimestamp(timestamp);
        fileItem.setExecutable(executable);
	}
    
    private IVersionable performModification(IImportChange change, Date timestamp, CommitRecord record, IProgressMonitor monitor)
                throws TeamRepositoryException {
    	try {
	    	SubMonitor progress = SubMonitor.convert(monitor, 20);
	    	String message;
	    	if ((change.getFlags() & IImportChange.MOVE) == 0) {
	    		message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_30, change.getAfterPath());
	    	} else {
	    		message = NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_31, change.getBeforePath());
	    	}
			progress.subTask(message);
	
			IVersionable item;
			String afterPath = change.getAfterPath();
			if ((change.getFlags() & IImportChange.MOVE) >0) {
				String beforePath = change.getBeforePath();
				ensureParentExists(afterPath, record);
				item = fileTree.move(beforePath, afterPath);
			} else {
	            item = fileTree.getItem(afterPath);
	        }
	        
	        if (item instanceof IFileItem) {
	        	IFileItem fileItem = (IFileItem) item;
				IFileProperties fileProperties = change.getFileProperties();
				if (isContentChange(change)) {
					progress.subTask(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_32, change.getBeforePath()));
					setContents(fileItem, change, timestamp, -1, null, progress.newChild(19));
				} else if (isExecuteBitChange((IFileItem)item, fileProperties)) {
					fileItem.setExecutable(fileProperties.isExecutable());
					fileItem.setFileTimestamp(timestamp);
				}
			}
	
	        // update map and item list
	        record.updateItemsMapsForSave(item);
	        return item;
    	} finally {
    		monitor.done();
    	}
    }

	private boolean isContentChange(IImportChange change) {
		return (change.getFlags() & IImportChange.CONTENT) > 0;
	}

	private boolean isLineDelimiterChange(IFileItem item, IFileProperties fileProperties) {
		IFileContent content = item.getContent();
		if (content == null) {
			return false;
		}
		return content.getLineDelimiter() != fileProperties.getLineDelimiter();
	}
    
    private boolean isExecuteBitChange(IFileItem item, IFileProperties fileProperties) {
		return item.isExecutable() != fileProperties.isExecutable();
	}

	private IVersionable deleteVersionable(String path, CommitRecord record)
            throws TeamRepositoryException {
		boolean childrenKnown = fileTree.areChilrenKnown(path);
		if (childrenKnown) {
			// The children are known so we should delete each child individually
			deleteChildren(path, record);
		}
        IVersionable item = fileTree.remove(path);
        if (item == null) {
        	throw new TeamRepositoryException(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_34, path));
        }
        // If the children are not know, we can perform a deleteSubtree safely
        record.updateItemsMapForDelete(item, !childrenKnown);
        return item;
    }
    
    private void deleteChildren(String path, CommitRecord record) throws TeamRepositoryException {
		String[] childNames = fileTree.getChildNames(path).toArray(new String[0]);
		for (String childName : childNames) {
			deleteVersionable(PathUtils.appendPath(path, childName), record);
		}
	}

	private void deleteEmptyFoldersIfRequired(CommitRecord record) throws TeamRepositoryException {
        if (!directoriesSignificant && fileTree instanceof FileTreeWithEmptyFolderTracking) {
			FileTreeWithEmptyFolderTracking vt = (FileTreeWithEmptyFolderTracking) fileTree;	
			Set<String> emptyFolders;
			// We need to loop copy the list and loop because deleting a child
			// folder can cause it's parent to become empty
			while (!(emptyFolders = vt.getEmptyFolders()).isEmpty()) {
				for (String path : emptyFolders) {
			        IVersionable item = fileTree.remove(path);
			        record.updateItemsMapForDelete(item, false);
				}
			}
		}
    }
    
	private IChangeSetHandle createActiveChangeSet(String comment, IWorkspaceConnection workspace,
            IComponent component, IProgressMonitor monitor) throws TeamRepositoryException {
		SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.ChangeSetArchiveImporter_37, 2);
		IChangeSetHandle changeSet = null;
		try {
			stats.startPhase(IChangeArchiveStatistics.CREATE_CHANGE_SET);
			changeSet = workspace.createChangeSet(component, progress.newChild(1));
			addCommentsTaskToChangeSet(comment, changeSet, workspace, progress.newChild(1));
			return changeSet;
		} catch (TeamRepositoryException e) {
			try {
				if (changeSet != null) {
					workspace.discardChangeSets(false, Collections.singletonList(changeSet), progress.newChild(0));
				}
			} catch (TeamRepositoryException ex) {}
            throw e;
		} finally {
			stats.endPhase(IChangeArchiveStatistics.CREATE_CHANGE_SET);
			monitor.done();
		}
	}

	private void addCommentsTaskToChangeSet(String comment, IChangeSetHandle changeSet,
            IWorkspaceConnection workspace, IProgressMonitor monitor) throws TeamRepositoryException {
		SubMonitor progress = SubMonitor.convert(monitor, 100);		
		SCMPlatform.getWorkspaceManager(workspace.teamRepository()).setComment(changeSet, 
				comment, progress.newChild(100));
		monitor.done();
	}

	public void addImportParticipant(IImportParticipant participant) {
		synchronized (importParticipants) {
			importParticipants.add(participant);
		}
	}
	
	public IChangeArchiveStatistics getStats() {
		return stats;
	}
}
