/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.ibm.team.scm.client.importz.IChangeSetFileReader;
import com.ibm.team.scm.client.importz.IImportChangeSet;
import com.ibm.team.scm.client.importz.spi.DerivedImportChangeSet;

public class ChangeSetBuffer {
    private int offset;
    private List<DerivedImportChangeSet> changeSets = new LinkedList<DerivedImportChangeSet>();
	private final IChangeSetFileReader reader;
	private boolean doneReading;
    
	public ChangeSetBuffer(IChangeSetFileReader reader) {
		this.reader = reader;
	}
    
    public void resetOffset() {
    	offset = 0;
	}

	public DerivedImportChangeSet removeFirst() {
		return changeSets.remove(0);
	}

	public DerivedImportChangeSet getFirst() {
		return changeSets.get(0);
	}

	public void add(DerivedImportChangeSet changeSet) {
		changeSets.add(changeSet);
	}

	public boolean isEmpty() {
		return changeSets.isEmpty();
	}

	public void primeBuffer() throws IOException {
        if (changeSets.isEmpty()) {
        	IImportChangeSet next = readNext();
        	if (next != null) {
            	changeSets.add((DerivedImportChangeSet)next);
        	}
        }
	}

	private IImportChangeSet readNext() throws IOException {
		if (doneReading)
			return null;
		IImportChangeSet next = reader.next(null);
		doneReading = next == null;
		return next;
	}
	
	public boolean walkBuffer(final DerivedImportChangeSet changeSet) throws IOException {
        
        boolean globalModified = false;
        boolean modified;

        // The outer loop is here to allow us to read more change sets if the change set that
        // was passed in expands (i.e. as more changes are added, the latest date may increase)
    done:
        do {
            modified = false;
            fillBuffer(changeSet);

            for (Iterator<DerivedImportChangeSet> i = changeSets.listIterator(offset); i.hasNext();) {
            	DerivedImportChangeSet nextChangeSet = i.next();
                if (changeSet == nextChangeSet) {
                	// We started at the change set that was passed in.
                	// Increment the start index so we start one later
                	offset++;
                    continue;
                }

                if (nextChangeSet.getEarliestDate().getTime() - changeSet.getLatestDate().getTime()
                         > DerivedImportChangeSet.MAX_TIME_DIFF) {
                    break done;
                }
                
                if (changeSet.sameAs(nextChangeSet)) {
                	// The provide change set and the one being tested are the same.
                	// Combine them and continue processing.
                	changeSet.combine(nextChangeSet);
                    i.remove();
                    modified = true;
                    globalModified = true;
                } else {
                	// The change set being tested is not the same as the one that was passed in.
                	// Adjust the start index so the we will skip it if we loop again.
                	offset++;
                }
            }
        } while (modified);
        
        return globalModified;
    }
    
    public void fillBuffer(DerivedImportChangeSet changeSet) throws IOException {
    	IImportChangeSet next;
		while ((next = readNext()) != null) {
			changeSets.add((DerivedImportChangeSet) next);
			if (((DerivedImportChangeSet) next).getEarliestDate().getTime()
					- changeSet.getLatestDate().getTime() < DerivedImportChangeSet.MAX_TIME_DIFF) {
				// The change the just read is out of the date range for the change
				// set of interest so we'll stop buffering
				return;
			}
		}
	}
}