/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal;

import java.util.List;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.ide.ui.internal.logical.IChangeSetFilter;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IBaselineConnection;
import com.ibm.team.scm.client.IChangeHistory;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.client.importz.IChangeSetArchiveImporter;
import com.ibm.team.scm.client.importz.IImportData;
import com.ibm.team.scm.client.importz.IImportParticipant;
import com.ibm.team.scm.client.importz.IMigrationFactory;
import com.ibm.team.scm.common.IBaselineHandle;
import com.ibm.team.scm.common.IChangeHistoryEntryChange;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponent;

public class ImportArchiveOperation extends AbstractImportOperation implements IChangeSetArchiveImporter {

    private IWorkspaceConnection userWorkspace;
    private IComponent component;
    private IMigrationFactory factory;
	private IChangeArchiveStatistics stats;
    
    public ImportArchiveOperation(IImportData importData, IWorkspaceConnection workspace, IComponent component, IMigrationFactory factory) {
        super(factory, importData);
        Assert.isNotNull(workspace);
        Assert.isNotNull(component);
        this.userWorkspace = workspace;
        this.component = component;
        this.factory = factory;

    }
    
    /**
     * Import the contents of the archive file into the component that is the 
     * target of the import.
     * @param archiveFile the archive file
     * @param importData the source SCM specific data that is passed to any
     * 		facilities obtained from the migration factory
     * @param monitor a progress monitor
     * @throws FileSystemException
     * @throws TeamRepositoryException
     */
	public void run(IProgressMonitor monitor) throws TeamRepositoryException, FileSystemException {
		boolean closeArchive = false;
		try {
			closeArchive = getImportData().openArchive(monitor);
			doImport(getImportData(), monitor);
		} finally {
			if (closeArchive)
				getImportData().closeArchive();
		}
	}
	
    private void doImport(IImportData importData, IProgressMonitor monitor)
            throws FileSystemException, TeamRepositoryException {
    	String message = NLS.bind(SCMImportMessages.ImportArchiveOperation_0, factory.getName());
    	SubMonitor progress = SubMonitor.convert(monitor, message, 100);
    	progress.setTaskName(message);
    	if (progress.isCanceled())
    		throw new OperationCanceledException();
    	
        ChangeSetArchiveImporter importer;
        IImportParticipant importParticipant;
        
        IChangeSetHandle initial = null;
    	ImportChangeSetFilter filter = null;
    	// For now, only use a temporary workspace when syncing
        IWorkspaceConnection importWorkspace;
        importWorkspace = userWorkspace;
        importer = new ChangeSetArchiveImporter(factory, importData);
        
        importParticipant = factory.createImportParticipant(userWorkspace, component, importData);;
        if (importParticipant != null)
        	importer.addImportParticipant(importParticipant);
        
        importer.importChangeSetArchive(importWorkspace, component, false, progress.newChild( 80));
        this.stats = importer.getStats();
        
        monitor.done();
    }
    
    /* package */ static IChangeSetHandle getInitialCS(IComponent component, ITeamRepository repo,
    		IProgressMonitor monitor) throws TeamRepositoryException {
    	SubMonitor sm = SubMonitor.convert(monitor, 100);
        IBaselineHandle initial = component.getInitialBaseline();
        
        IBaselineConnection bc = SCMPlatform.getWorkspaceManager(repo)
        					.getBaselineConnection(initial, sm.newChild(50));
        IChangeHistory hist = bc.changeHistory();
        List<IChangeHistoryEntryChange> recent = hist.recent(sm.newChild(50));
        if (recent.size() != 1) {
        	throw new IllegalStateException(SCMImportMessages.ImportArchiveOperation_2);
        }
        if (monitor.isCanceled())
        	throw new OperationCanceledException();
    	monitor.done();
        return recent.get(0).changeSet();
    }
    
    /*package*/ static class ImportChangeSetFilter implements IChangeSetFilter {
        private IChangeSetFilter userFilter;
        private IChangeSetHandle firstChangeSet;
        
        public ImportChangeSetFilter(IChangeSetFilter userFilter, IChangeSetHandle firstChangeSet) {
            this.userFilter = userFilter;
            this.firstChangeSet = firstChangeSet;
        }
        
        public boolean excludeChangeSet(IChangeSet changeSet) {
        	if (firstChangeSet.sameItemId(changeSet)) {
        		return false;
        	}
        	
            return userFilter == null || userFilter.excludeChangeSet(changeSet);
        }
        
        public boolean allDiscarded() {
            return userFilter != null && userFilter.allDiscarded();
        }
        
        public IChangeSetFilter getUserFilter() {
            return userFilter;
        }
    }

	public IChangeArchiveStatistics getStats() {
		return stats;
	}

	public void performImport(IProgressMonitor monitor)
			throws TeamRepositoryException, FileSystemException {
		run(monitor);
	}
}
