/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.File;
import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IImportWizard;
import org.eclipse.ui.IWorkbench;

import com.ibm.team.filesystem.ui.wrapper.WorkspaceComponentWrapper;
import com.ibm.team.internal.filesystem.ui.wizards.TeamPlaceSelectionPage;
import com.ibm.team.internal.filesystem.ui.wizards.sharing.TeamAreaSelection;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.ui.utils.UIContext;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;

/**
 * Abstract import wizard that provides the pages for specifying the
 * target repository/workspace/component of an import.
 */
public abstract class AbstractImportWizard extends Wizard implements IImportWizard {

    private static final String TEAM_PLACE_SELECTION_PAGE_TITLE = SCMImportMessages.AbstractImportWizard_0;
    private static final String TEAM_PLACE_SELECTION_PAGE_DESCRIPTION = SCMImportMessages.AbstractImportWizard_1;
    
    /*package*/ static final String DEFAULT_FILENAME = new File(System.getProperty("user.home"), "archive.tar").getAbsolutePath(); //$NON-NLS-1$ //$NON-NLS-2$
    
    private TeamPlaceSelectionPage teamPlaceSelectionPage;
	private ConfigureImportProcessingPage configureImportProcessingPage; 
    private Author2ContributorMappingPage contributorMappingPage;
	private ImportConfiguration configuration;
	private UIContext context;

	@Override
	public void addPages() {
		configuration = createConfiguration();
		initializeConfiguration();
        addPagesAtBeginning(); 
        teamPlaceSelectionPage = createTargetComponentSelectionPage();
        addPage(teamPlaceSelectionPage);
        
		configureImportProcessingPage = createConfigureImportProcessingPage();
		addPage(configureImportProcessingPage);
		
        contributorMappingPage = createContributorMappingPage();
        addPage(contributorMappingPage);
        
        addPagesAtEnd();
	}

	protected ConfigureImportProcessingPage createConfigureImportProcessingPage() {
		ConfigureImportProcessingPage page = new ConfigureImportProcessingPage(
				"CustomizeImportProcessingPage", SCMImportMessages.AbstractImportFromOtherWizard_3, //$NON-NLS-1$
				getBanner(), getConfiguration());
        page.setDescription(SCMImportMessages.AbstractImportFromOtherWizard_7);
		return page;
	}
	
	private Author2ContributorMappingPage createContributorMappingPage() {
		Author2ContributorMappingPage page = new Author2ContributorMappingPage(
				"ContributorMappingPage", //$NON-NLS-1$
				getBanner(),
				configuration,
				false,
				false);
		return page;
	}

	private TeamPlaceSelectionPage createTargetComponentSelectionPage() {
		TeamPlaceSelectionPage page = new TeamPlaceSelectionPage("TeamPlaceSelectionPage", TEAM_PLACE_SELECTION_PAGE_TITLE, getBanner(), false, IHelpContextIds.HELP_CONTEXT_TEAM_PLACE_SELECTION_FOR_IMPORT_PAGE); //$NON-NLS-1$
		page.setDescription(TEAM_PLACE_SELECTION_PAGE_DESCRIPTION);
		return page;
	}

	protected abstract void addPagesAtBeginning();

	protected void addPagesAtEnd() {
		// Do nothing by default
	}
	
    protected ImportConfiguration createConfiguration() {
    	ImportConfiguration importConfiguration = new ImportConfiguration(getSourceRepositoryKind(), false);
    	importConfiguration.setProperty(ReconcileOptionsArea.RECONCILE_WITH_EXISTING, Boolean.TRUE);
		return importConfiguration;
    }
    
    protected abstract String getSourceRepositoryKind();

	protected void initializeConfiguration() {
    	configuration.setArchiveFilename(getDefaultFilename());
	}
    
    protected String getDefaultFilename() {
    	return AbstractImportWizard.DEFAULT_FILENAME;
    }
    
	public IWizardPage getNextPage(IWizardPage page) {
        return getNextPage(page, true);
    }
    
    public IWizardPage getNextPage(IWizardPage page, boolean aboutToShow) {
    	ITeamRepository repo = teamPlaceSelectionPage.getRepository();
    	if (repo != null && aboutToShow) {
    		configuration.setRepository(repo);
    	} 
        return super.getNextPage(page);
    }
    
    public boolean canFinish() {
        return teamPlaceSelectionPage.isPageComplete() && configureImportProcessingPage.isPageComplete();
    }
    
	protected boolean aboutToFinish() {
		configureImportProcessingPage.storeValues();
		// Validate the state of the configuration
    	IStatus status = configuration.validateState();
    	if (status.getSeverity() == IStatus.ERROR) {
    		ImportLoggingHelper.displayError(SCMImportMessages.AbstractImportFromOtherWizard_4, status.getMessage());
            return false;
    	}
    	if (status.getSeverity() == IStatus.WARNING) {
    		if (!MessageDialog.openQuestion(getShell(), SCMImportMessages.AbstractImportFromOtherWizard_5, NLS.bind(SCMImportMessages.AbstractImportFromOtherWizard_6, status.getMessage())))
    			return false;
    	}
		if (configuration.isImportIntoWorkspace()) {
			ITeamRepository repository = teamPlaceSelectionPage.getRepository();
			if (!confirmImportForNonAdminUser(getShell(), repository)) {
				return false;
			}
			if (!Author2ContributorMappingArea.promptForSuspectMappings(getShell(), repository.loggedInContributor(), contributorMappingPage.getAuthorMappings(), configuration.getSourceRepositoryKind())) {
				return false;
			}
			configuration.setRepository(repository);
			if (teamPlaceSelectionPage.creatingNewWorkspace()) {
		        configuration.setWorkspaceName(teamPlaceSelectionPage.getWorkspaceName());	
			} else {
		        final TeamAreaSelection teamSelection = teamPlaceSelectionPage.getTeamSelection();
		        try {
					getContainer().run(true, true, new IRunnableWithProgress() {
						public void run(IProgressMonitor monitor) throws InvocationTargetException,
								InterruptedException {
					    	try {
								WorkspaceComponentWrapper wrapper = teamSelection.getComponent(configuration.getRepository(), monitor);
								if (wrapper != null) {
								    configuration.setTargetComponent(wrapper.getWorkspaceConnection(), wrapper.getComponent());
								}
							} catch (TeamRepositoryException e) {
								throw new InvocationTargetException(e);
							}
						}
					
					});
				} catch (InvocationTargetException e) {
					ImportLoggingHelper.displayError(SCMImportMessages.AbstractImportWizard_2, e.getTargetException());
					return false;
				} catch (InterruptedException e) {
					return false;
				}
			}
			configuration.setAuthorMappings(contributorMappingPage.getAuthorMappings());
		}
        return true;
	}
	
	private boolean confirmImportForNonAdminUser(Shell shell, ITeamRepository repository) {
		Object property = getConfiguration().getProperty(Author2ContributorMappingPage.AUTHOR_MAPPING_PROPERTY);
		if (property == null || property == Author2ContributorMappingPage.NO_MAPPING) {
			boolean isAdmin = hasAdminPrivileges(repository);
			if (!isAdmin) {
				return MessageDialog.openConfirm(shell, SCMImportMessages.AbstractImportWizard_3, NLS.bind(SCMImportMessages.AbstractImportWizard_4, repository.loggedInContributor().getUserId()));
			}
		}
		// We are admin so confirm that the import can continue
		return true;
	}

	private boolean hasAdminPrivileges(final ITeamRepository repository) {
        try {
        	final boolean[] isAdmin = new boolean[] { false };
			getContainer().run(true, true, new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor) throws InvocationTargetException,
						InterruptedException {
					isAdmin[0] = Author2ContributorMappingArea.hasJazzAdminPermissions(repository, monitor);
				}
			
			});
			return isAdmin[0];
		} catch (InvocationTargetException e) {
			ImportLoggingHelper.displayError(SCMImportMessages.AbstractImportWizard_2, e.getTargetException());
			return false;
		} catch (InterruptedException e) {
			return false;
		}
	}

	protected abstract ImageDescriptor getBanner();

	public ImportConfiguration getConfiguration() {
		return configuration;
	}
	
    public void init(IWorkbench workbench, IStructuredSelection selection) {
        context = UIContext.createShellContext(workbench.getActiveWorkbenchWindow().getShell());
    }

	public UIContext getContext() {
		return context;
	}

	public void setContext(UIContext context) {
		this.context = context;
	}
	
	@Override
	public final boolean performFinish() {
		if (!aboutToFinish())
			return false;
		ImportOperation op = createImportOperation();
		getContext().getUserOperationRunner().enqueue(op.getName(), op);
		return true;
	}

	protected abstract ImportOperation createImportOperation();

}
