/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.PrintStream;

import org.eclipse.core.runtime.Assert;

import com.ibm.team.repository.common.IContributor;

public class Author2ContributorMapping {

	private final String authorId;
	private String contributorId;
	private IContributor contributor;

	/**
	 * Create a mapping from the given author ID to a contributor. If the contributor
	 * is <code>null</code>, the author ID is used as the contributor ID.
	 * @param author the author ID (or user ID) from the external repository.
	 * @param contributor the contributor that represents the author or <code>null</code>
	 * 		if a contributor could not be found or is not assigned for the author.
	 */
	public Author2ContributorMapping(String author, IContributor contributor) {
		this(author, contributor == null ? author : contributor.getUserId());
		this.contributor = contributor;
	}
	
	public Author2ContributorMapping(String authorId, String contributorId) {
		Assert.isNotNull(authorId);
		Assert.isNotNull(contributorId);
		this.authorId = authorId;
		this.contributorId = contributorId;
	}

	/**
	 * Return the author ID (i.e. the user ID from the external repository).
	 * @return the author ID
	 */
	public String getAuthorId() {
		return authorId;
	}
	
	/**
	 * Return the contributor ID that is used to fetch the contributor.
	 * @return the contributor ID that is used to fetch the contributor
	 */
	public String getContributorId() {
		return contributorId;
	}

	public void setContributor(IContributor contributor) {
		if (contributor == null) {
			this.contributor = null;
			this.contributorId = authorId;
		} else {
			this.contributor = contributor;
			this.contributorId = contributor.getUserId();
		}
	}

	/**
	 * Get the contributor for the author. A <code>null</code> is returned
	 * if the contributor is not known or one has not been assigned to the 
	 * author ID.
	 * @return the contributor for the author or <code>null</code>
	 */
	public IContributor getContributor() {
		return contributor;
	}

	/**
	 * Write the mapping to the given stream as a single line.
	 * @param out the output stream
	 */
	public void write(PrintStream out) {
		out.print(authorId);
		if (!contributorId.equals(authorId)) {
			out.print(","); //$NON-NLS-1$
			out.print(contributorId);
		}
		out.println();
	}

	public static Author2ContributorMapping createFrom(String line) {
		line = line.trim();
		if (line.length() > 0) {
			int index = line.indexOf(',');
			String authorId;
			String contributorId;
			if (index == -1) {
				authorId = line;
				contributorId = line;
			} else {
				authorId = line.substring(0, index).trim();
				contributorId = line.substring(index + 1).trim();
			}
			return new Author2ContributorMapping(authorId, contributorId);
		}
		return null;
	}

	public void setContributorId(String contributorId) {
		if (!this.contributorId.equals(contributorId)) {
			this.contributorId = contributorId;
			this.contributor = null;
		}
		
	}

}
