/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.preferences.ConfigurationScope;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.ViewerLabel;
import org.eclipse.jface.window.Window;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Shell;

import com.ibm.team.filesystem.ui.wrapper.ContributorWrapper;
import com.ibm.team.process.rcp.ui.TeamContributorSelectionDialog;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.util.IClientLibraryContext;
import com.ibm.team.repository.client.util.IClientLibraryContext.IServiceRunnable;
import com.ibm.team.repository.common.IContributor;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.service.IPermissionService;
import com.ibm.team.repository.rcp.core.utils.StatusUtil;
import com.ibm.team.repository.rcp.ui.internal.utils.BaseLabelProvider;
import com.ibm.team.repository.rcp.ui.internal.utils.JFaceUtils;
import com.ibm.team.repository.rcp.ui.internal.viewers.AbstractSetWithListeners;
import com.ibm.team.repository.rcp.ui.internal.viewers.ConvertColumn;
import com.ibm.team.repository.rcp.ui.internal.viewers.IConversion;
import com.ibm.team.repository.rcp.ui.internal.viewers.ISetWithListeners;
import com.ibm.team.repository.rcp.ui.internal.viewers.LabelColumn;
import com.ibm.team.repository.rcp.ui.internal.viewers.SetTransformer;
import com.ibm.team.repository.rcp.ui.internal.viewers.SimpleTableViewer;
import com.ibm.team.repository.rcp.ui.operations.IOperationRunner;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;

public class Author2ContributorMappingArea extends AbstractOptionsArea {

	private static boolean hasMappingToLoggedInUser(IContributor loggedInContributor, Collection<Author2ContributorMapping> mappings) {
		for (Author2ContributorMapping mapping : mappings) {
			IContributor contributor = mapping.getContributor();
			if (contributor != null && contributor.sameItemId(loggedInContributor) && !loggedInContributor.getUserId().equals(mapping.getAuthorId())) {
				// Return true if the contributor for the mapping matches the logged in contributor
				// but the user id of the logged in contributor does not match the user id of the mapping
				return true;
			}
		}
		return false;
	}
	
	public static boolean promptForSuspectMappings(Shell shell, IContributor loggedInContributor, Collection<Author2ContributorMapping> mappings, String repoType) {
		if (hasMappingToLoggedInUser(loggedInContributor, mappings)) {
			return MessageDialog.openConfirm(shell, SCMImportMessages.Author2ContributorMappingArea_0, NLS.bind(SCMImportMessages.Author2ContributorMappingArea_1, repoType));
		}
		return true;
	}
	
	private final class ContributorConversion extends
			SetTransformer<Author2ContributorMapping, ContributorWrapper> {
		public ContributorConversion(ISetWithListeners<Author2ContributorMapping> domain) {
			super(domain, true);
		}

		public final void fireEvent(Collection elementsThatChanged) {
			Map<Author2ContributorMapping, ContributorWrapper> newResults = new HashMap<Author2ContributorMapping, ContributorWrapper>();
			for (Object next : elementsThatChanged) {
				if (next instanceof Author2ContributorMapping) {
					Author2ContributorMapping mapping = (Author2ContributorMapping) next;
					newResults.put(mapping, doComputeResult(mapping));
				}
			}
			reportResults(newResults);
		}
		
		@Override
		protected ContributorWrapper doComputeResult(
				Author2ContributorMapping element) {
			IContributor contributor = element.getContributor();
			if (contributor == null)
				return null;
			return new ContributorWrapper(contributor);
		}
	}

	private final class ContributorIdLabelProvider extends BaseLabelProvider {
		@Override
		public void updateLabel(ViewerLabel label, Object element) {
			if (element instanceof Author2ContributorMapping) {
				Author2ContributorMapping mapping = (Author2ContributorMapping) element;
				IContributor contributor = mapping.getContributor();
				if (contributor != null || !mapping.getContributorId().equals(mapping.getAuthorId()))
					label.setText(mapping.getContributorId());
			}
		}
		
		public void fireLabelChange(Collection changed) {
			fireChangeEvent(changed);
		}
		
	}


	private final class Author2ContributorSet extends AbstractSetWithListeners<Author2ContributorMapping> {
		private List<Author2ContributorMapping> elements = new ArrayList<Author2ContributorMapping>();
		public Author2ContributorSet(
				Collection<Author2ContributorMapping> authorMappings) {
			if (authorMappings != null) {
				for (Author2ContributorMapping author2ContributorMapping : authorMappings) {
					elements.add(author2ContributorMapping);
				}
			}
		}
		@Override
		protected Collection<Author2ContributorMapping> computeElements() {
			return elements;
		}
		public void remove(List list) {
			List<Author2ContributorMapping> removed = new ArrayList<Author2ContributorMapping>(list.size());
			for (Object object : list) {
				if (object instanceof Author2ContributorMapping) {
					Author2ContributorMapping mapping = (Author2ContributorMapping) object;
					elements.remove(mapping);
					removed.add(mapping);
				}
			}
			fireCollectionRemoved(removed);
		}
		public void addAuthor(String author, IContributor contributor) {
			Author2ContributorMapping mapping = new Author2ContributorMapping(author, contributor);
			elements.add(mapping);
			fireAdded(mapping);
		}
		public Author2ContributorMapping getMapping(String author) {
			if (author.length() == 0)
				return null;
			for (Author2ContributorMapping mapping : elements) {
				if (mapping.getAuthorId().equals(author)) {
					return mapping;
				}
			}
			return null;
		}
		public void addAll(List<Author2ContributorMapping> toAdd) {
			if (toAdd.isEmpty())
				return;
			elements.addAll(toAdd);
			fireCollectionAdded(toAdd);
		}
		public boolean containsAuthor(String authorId) {
			return getAuthor(authorId) != null;
		}
		public Author2ContributorMapping getAuthor(String authorId) {
			for (Author2ContributorMapping mapping : elements) {
				if (mapping.getAuthorId().equals(authorId))
					return mapping;
			}
			return null;
		}
		public void write(PrintStream out) {
			for (Author2ContributorMapping mapping : elements) {
				mapping.write(out);
			}
		}
		public void load(BufferedReader in) throws IOException {
			// Load into a hash table to avoid duplicates
			String line;
			final Map<String, Author2ContributorMapping> loaded = new HashMap<String, Author2ContributorMapping>();
			while ((line = in.readLine()) != null) {
				Author2ContributorMapping mapping = Author2ContributorMapping.createFrom(line);
				loaded.put(mapping.getAuthorId(), mapping);
			}
			try {
				runnableContext.run(true, true, new IRunnableWithProgress() {
					public void run(IProgressMonitor monitor) throws InvocationTargetException,
							InterruptedException {
						FetchContributorsFunction.fetchContributors(configuration.getRepository(), loaded.values(), monitor);
					}
				});
			} catch (InvocationTargetException e) {
				StatusUtil.log(this, e);
			} catch (InterruptedException e) {
				// Ignore
			}
			final List<Author2ContributorMapping> added = new ArrayList<Author2ContributorMapping>();
			final List<Author2ContributorMapping> changed = new ArrayList<Author2ContributorMapping>();
			for (Map.Entry<String, Author2ContributorMapping> entry : loaded.entrySet()) {
				Author2ContributorMapping mapping = entry.getValue();
				Author2ContributorMapping existing = getAuthor(mapping.getAuthorId());
				if (existing == null) {
					elements.add(mapping);
					added.add(mapping);
				} else if (!existing.getContributorId().equals(mapping.getContributorId())) {
					existing.setContributorId(mapping.getContributorId());
					existing.setContributor(mapping.getContributor());
					changed.add(existing);
				} else if (existing.getContributor() == null) {
					existing.setContributor(mapping.getContributor());
					changed.add(existing);
				}			
			}
			if (!added.isEmpty()) {
				fireCollectionAdded(added);
			}
			if (!changed.isEmpty()){
				updateElements(changed);
			}
		}
		public List<Author2ContributorMapping> getElements() {
			return elements;
		}
	}
	
	private SimpleTableViewer<Author2ContributorMapping> viewer;
	private Author2ContributorSet input;
	private FetchContributorsFunction fetchFunction;
	private boolean canAdd;
	private Button addButton;
	private Button editButton;
	private Button removeButton;
	private final ImportConfiguration configuration;
	private Composite composite;
	private ContributorIdLabelProvider labelUpdater;
	private ContributorConversion contributorUpdater;
	private IRunnableContext runnableContext;
	private Button loadButton;
	private Button saveButton;
	private Button createButton;

	public Author2ContributorMappingArea(Composite parent, ImportConfiguration configuration, boolean canAdd, IRunnableContext runnableContext, IOperationRunner operationRunner) {
		this.configuration = configuration;
		this.canAdd = canAdd;
		this.runnableContext = runnableContext;
		composite = new Composite(parent, SWT.NONE);
		{
			viewer = new SimpleTableViewer<Author2ContributorMapping>(composite, SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL | SWT.FULL_SELECTION | SWT.MULTI, getConfigPrefs(), null);
			input = new Author2ContributorSet(configuration.getAuthorMappings());
			// We delay setting the viewer input until the page is visible
			viewer.getSelectionProvider().addSelectionChangedListener(new ISelectionChangedListener() {
				public void selectionChanged(SelectionChangedEvent event) {
					updateEnablements();
				}
			});
			new ConvertColumn<Author2ContributorMapping, String>(viewer, NLS.bind(SCMImportMessages.Author2ContributorMappingArea_2, configuration.getSourceRepositoryKind()), 100, new IConversion<Author2ContributorMapping, String>() {
				public String createAdapter(Author2ContributorMapping source) {
					return source.getAuthorId();
				}
			});
			contributorUpdater = new ContributorConversion(input);
			new ConvertColumn<Author2ContributorMapping, ContributorWrapper>(viewer, SCMImportMessages.Author2ContributorMappingArea_3, 150, contributorUpdater);
			labelUpdater = new ContributorIdLabelProvider();
			new LabelColumn<Author2ContributorMapping>(viewer, SCMImportMessages.Author2ContributorMappingArea_4, 100, labelUpdater);
			GridDataFactory.defaultsFor(viewer.getTable()).grab(true, true).applyTo(viewer.getTable());
			
			Composite buttons = new Composite(composite, SWT.NONE);
			{
				createButtons(buttons);
			}
			GridLayoutFactory.fillDefaults().applyTo(buttons);
		}
		GridLayoutFactory.fillDefaults()
			.numColumns(2)
			.generateLayout(composite);
		updateEnablements();
	}

	public void linkInputToViewer(ITeamRepository repository,
			IOperationRunner operationRunner) {
		if (fetchFunction == null || !fetchFunction.getRepository().getId().equals(repository.getId())) {
			fetchFunction = new FetchContributorsFunction(repository, input, operationRunner);
			viewer.setInput(fetchFunction.getRange());
			updateEnablements();
		}
	}

	protected void updateEnablements() {
		boolean enabled = isEnabled();
		if (addButton != null) {
			addButton.setEnabled(enabled);
		}
		if (removeButton != null) {
			removeButton.setEnabled(enabled && !viewer.getSelection().isEmpty());
		}
		editButton.setEnabled(enabled && !viewer.getSelection().isEmpty());
		createButton.setEnabled(enabled && !input.getElements().isEmpty());
		loadButton.setEnabled(enabled);
		saveButton.setEnabled(enabled && !input.getElements().isEmpty());
		viewer.getTable().setEnabled(enabled);
		getControl().setEnabled(enabled);
	}

	private void createButtons(Composite buttons) {
		GridDataFactory gdFactory = GridDataFactory.swtDefaults().align(SWT.FILL, SWT.TOP);
		if (canAdd) {
			addButton = new Button(buttons, SWT.PUSH);
			addButton.setText(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_5, configuration.getSourceRepositoryKind()));
			addButton.addSelectionListener(new SelectionListener() {
				public void widgetSelected(SelectionEvent e) {
					addAuthor();
				}
				public void widgetDefaultSelected(SelectionEvent e) {
					// Do nothing
				}
			});
			gdFactory.applyTo(addButton);
		}
		editButton = new Button(buttons, SWT.PUSH);
		editButton.setText(SCMImportMessages.Author2ContributorMappingArea_6);
		editButton.setToolTipText(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_7, configuration.getSourceRepositoryKind()));
		editButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				editAuthors();
			}
			public void widgetDefaultSelected(SelectionEvent e) {
				// Do nothing
			}
		});
		gdFactory.applyTo(editButton);
		createButton = new Button(buttons, SWT.PUSH);
		createButton.setText(SCMImportMessages.Author2ContributorMappingArea_25);
		createButton.setToolTipText(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_26, configuration.getSourceRepositoryKind()));
		createButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				createContributors();
			}
			public void widgetDefaultSelected(SelectionEvent e) {
				// Do nothing
			}
		});
		gdFactory.applyTo(createButton);
		removeButton = new Button(buttons, SWT.PUSH);
		if (canAdd) {
			removeButton.setText(SCMImportMessages.Author2ContributorMappingArea_8);
			removeButton.setToolTipText(SCMImportMessages.Author2ContributorMappingArea_9);
		} else {
			removeButton.setText(SCMImportMessages.Author2ContributorMappingArea_27);
			removeButton.setToolTipText(SCMImportMessages.Author2ContributorMappingArea_28);
		}
		removeButton.addSelectionListener(new SelectionListener() {
			public void widgetSelected(SelectionEvent e) {
				removeAuthors();
			}
			public void widgetDefaultSelected(SelectionEvent e) {
				// Do nothing
			}
		});
		gdFactory.applyTo(removeButton);
		
		loadButton = new Button(buttons, SWT.PUSH);
		loadButton.setText(SCMImportMessages.Author2ContributorMappingArea_10);
		loadButton.setToolTipText(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_11, configuration.getSourceRepositoryKind()));
		gdFactory.applyTo(loadButton);
		loadButton.addSelectionListener(new SelectionListener() {
				public void widgetSelected(SelectionEvent e) {
					loadFromFile();
				}
				public void widgetDefaultSelected(SelectionEvent e) {
					// Do nothing
				}
			});
		
		saveButton = new Button(buttons, SWT.PUSH);
		saveButton.setText(SCMImportMessages.Author2ContributorMappingArea_12);
		saveButton.setToolTipText(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_13, configuration.getSourceRepositoryKind()));
		gdFactory.applyTo(saveButton);
		saveButton.addSelectionListener(new SelectionListener() {
				public void widgetSelected(SelectionEvent e) {
					saveToFile();
				}
				public void widgetDefaultSelected(SelectionEvent e) {
					// Do nothing
				}
			});
	}

	protected void saveToFile() {
		FileDialog dialog = new FileDialog(composite.getShell(), SWT.SAVE);
		String fileName = dialog.open();
		if (fileName != null) {
			PrintStream out = null;
			try {
				out = new PrintStream(new FileOutputStream(fileName));
				input.write(out);
			} catch (IOException e) {
				JFaceUtils.showError(SCMImportMessages.Author2ContributorMappingArea_14, NLS.bind(SCMImportMessages.Author2ContributorMappingArea_15, fileName), e);
			} finally {
				if (out != null) {
					out.close();
				}
			}
		}
	}

	protected void loadFromFile() {
		FileDialog dialog = new FileDialog(composite.getShell(), SWT.OPEN);
		final String fileName = dialog.open();
		if (fileName != null) {
			try {
				BufferedReader in = null;
				try {
					in = new BufferedReader(new FileReader(fileName));
					input.load(in);
				} finally {
					if (in != null) {
						try {
							in.close();
						} catch (IOException e) {
							// Ignore
						}
					}
				}
			} catch (IOException e) {
				JFaceUtils.showError(SCMImportMessages.Author2ContributorMappingArea_16, NLS.bind(SCMImportMessages.Author2ContributorMappingArea_17, fileName), e);
			}
			updateEnablements();
		}
	}

	protected void editAuthors() {
		TeamContributorSelectionDialog dialog = new TeamContributorSelectionDialog(editButton.getShell(), getTeamRepository(), Collections.EMPTY_LIST, false);
		int result = dialog.open();
		if (result == Window.OK) {
			try {
				Object element = dialog.getFirstResult();
				if (element instanceof IContributor) {
					IContributor contributor = (IContributor) element;
					updateContributor(contributor);
				} else if (element instanceof IContributorHandle) {
					IContributorHandle handle = (IContributorHandle) element;
					IContributor c = fetchContributor(handle);
					if (c != null)
						updateContributor(c);
				}
				updateEnablements();
			} catch (InvocationTargetException e) {
				ImportLoggingHelper.displayError(SCMImportMessages.Author2ContributorMappingArea_18, e.getTargetException());
			} catch (InterruptedException e) {
				// Nothing to do
			}
		}
	}

	private IContributor fetchContributor(final IContributorHandle handle) throws InvocationTargetException, InterruptedException {
		final IContributor[] contributor = new IContributor[] { null };
		runnableContext.run(true, true, new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor) throws InvocationTargetException,
					InterruptedException {
				try {
					contributor[0] = fetchContributor(handle, monitor);
				} catch (TeamRepositoryException e) {
					throw new InvocationTargetException(e);
				}
			}
		
		});
		return contributor[0];
	}
	
	private IContributor fetchContributor(IContributorHandle handle, IProgressMonitor monitor) throws TeamRepositoryException {
		return (IContributor)getTeamRepository().itemManager().fetchCompleteItem(handle, IItemManager.DEFAULT, monitor);
	}
	
	private IContributor fetchContributorFor(final String userId) throws InvocationTargetException, InterruptedException {
		final IContributor[] contributor = new IContributor[] { null };
		runnableContext.run(true, true, new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor) throws InvocationTargetException,
					InterruptedException {
				try {
					contributor[0] = getTeamRepository().contributorManager().fetchContributorByUserId(userId, null);
				} catch (TeamRepositoryException e) {
					throw new InvocationTargetException(e);
				}
			}
		
		});
		return contributor[0];
	}

	private void updateContributor(IContributor contributor) {
		IStructuredSelection selection = viewer.getSelection();
		List list = selection.toList();
		for (Object object : list) {
			if (object instanceof Author2ContributorMapping) {
				Author2ContributorMapping mapping = (Author2ContributorMapping) object;
				mapping.setContributor(contributor);
			}
		}
		updateElements(list);
	}
	
	protected void createContributors() {
		IStructuredSelection selection = viewer.getSelection();
		List list;
		if (selection.isEmpty()) {
			list = input.getElements();
		} else {
			list = selection.toList();
		}
		List<Author2ContributorMapping> result = new ArrayList<Author2ContributorMapping>(list.size());
		for (Object object : list) {
			if (object instanceof Author2ContributorMapping) {
				Author2ContributorMapping mapping = (Author2ContributorMapping) object;
				if (mapping.getContributor() == null) {
					result.add(mapping);
				}
			}
		}
		if (!result.isEmpty()) {
			createContributors(result);
		}
	}

	private void updateElements(List list) {
		labelUpdater.fireLabelChange(list);
		contributorUpdater.fireEvent(list);
	}

	private ITeamRepository getTeamRepository() {
		return configuration.getRepository();
	}

	protected void removeAuthors() {
		List list = viewer.getSelection().toList();
		if (canAdd) {
			input.remove(list);
		} else {
			// Remove the contributors for any selected users
			for (Object object : list) {
				if (object instanceof Author2ContributorMapping) {
					Author2ContributorMapping mapping = (Author2ContributorMapping) object;
					mapping.setContributor(null);
				}
			}
			updateElements(list);
		}
		updateEnablements();
	}

	protected void addAuthor() {
		InputDialog dialog = new InputDialog(editButton.getShell(), NLS.bind(SCMImportMessages.Author2ContributorMappingArea_19, getSourceRepositoryKind()), NLS.bind(SCMImportMessages.Author2ContributorMappingArea_20, getSourceRepositoryKind()), "", new IInputValidator() { //$NON-NLS-1$
			public String isValid(String newText) {
				for (int i = 0; i < newText.length(); i++) {
					char c = newText.charAt(i);
					if (Character.isWhitespace(c)) {
						return SCMImportMessages.Author2ContributorMappingArea_22;
					}
				}
				if (input.getMapping(newText) != null) {
					return NLS.bind(SCMImportMessages.Author2ContributorMappingArea_23, newText);
				}
				return null;
			}
		});
		int result = dialog.open();
		if (result == Window.OK) {
			String author = dialog.getValue();
			IContributor contributor;
			try {
				contributor = fetchContributorFor(author);
			} catch (InvocationTargetException e) {
				if (!(e.getTargetException() instanceof ItemNotFoundException)) {
					ImportLoggingHelper.displayError(NLS.bind(SCMImportMessages.Author2ContributorMappingArea_24, author), e);
				}
				contributor = null;
			} catch (InterruptedException e) {
				contributor = null;
			}
			input.addAuthor(author, contributor);
			Author2ContributorMapping mapping = input.getMapping(author);
			viewer.getSelectionProvider().setSelection(new StructuredSelection(mapping));
			updateEnablements();
		}
	}

	private String getSourceRepositoryKind() {
		return configuration.getSourceRepositoryKind();
	}

	public Control getControl() {
		return composite;
	}

	public ISetWithListeners<Author2ContributorMapping> getAuthorMappings() {
		return input;
	}

    protected IEclipsePreferences getConfigPrefs() {
        IEclipsePreferences result = new ConfigurationScope().getNode("com.ibm.team.scm.client.importz" + '/' + getClass().getName()); //$NON-NLS-1$
        return result;
    }

	public void ensureAuthorIdsAreAvailable(Collection<String> authorsForSelectedFolders) {
		List<Author2ContributorMapping> toAdd = new ArrayList<Author2ContributorMapping>();
		for (String authorId : authorsForSelectedFolders) {
			if (!input.containsAuthor(authorId)) {
				toAdd.add(new Author2ContributorMapping(authorId, authorId));
			}
		}
		input.addAll(toAdd);
	}
	
	protected void createContributors(final List<Author2ContributorMapping> mappings) {
		try {
			runnableContext.run(true, true, new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor) throws InvocationTargetException,
						InterruptedException {
					createContributors(getTeamRepository(), mappings, monitor);
				}
			});
		} catch (OperationCanceledException e) {
			// Ignore
		} catch (InvocationTargetException e) {
			ImportLoggingHelper.displayError(SCMImportMessages.Author2ContributorMappingArea_31, e.getTargetException());
		} catch (InterruptedException e) {
			// Ignore
		} finally {
			updateElements(mappings);
		}
	}
	
	public static final void createContributors(
			ITeamRepository reposiory,
			final Collection<Author2ContributorMapping> mappings,
			IProgressMonitor monitor) throws InvocationTargetException {
		SubMonitor progress = SubMonitor.convert(monitor, SCMImportMessages.Author2ContributorMappingArea_29, mappings.size() * 100);
		try {
			for (Author2ContributorMapping mapping : mappings) {
				if (mapping.getContributor() == null) {
					IContributor contributor = (IContributor) IContributor.ITEM_TYPE.createItem(reposiory);
					contributor.setName(mapping.getContributorId());
					contributor.setEmailAddress(SCMImportMessages.Author2ContributorMappingArea_30);
					contributor.setUserId(mapping.getContributorId());
					mapping.setContributor(reposiory.contributorManager().saveContributor(contributor, progress.newChild(100)));
				}
			}
		} catch (TeamRepositoryException e) {
			throw new InvocationTargetException(e);
		}
	}

	public static boolean hasJazzAdminPermissions(ITeamRepository repository, IProgressMonitor monitor) {
		IClientLibraryContext client = (IClientLibraryContext) repository;
		final IPermissionService permissionService = (IPermissionService) client.getServiceInterface(IPermissionService.class);
		try {
			client.callCancelableService(new IServiceRunnable<Boolean>() {
				public Boolean run(IProgressMonitor monitor)
						throws TeamRepositoryException {
					permissionService.assertPermission(IPermissionService.JAZZ_ADMINS);
					return null;
				}
			}, monitor);
			// If the call succeeds, we have admin permissions
			return true;
		} catch (TeamRepositoryException e) {
			
		}
		return false;

	}
    
}
