/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IImportWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.plugin.AbstractUIPlugin;

import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.importz.IChangeSetArchiveCreator;
import com.ibm.team.scm.client.importz.IMigrationFactory;
import com.ibm.team.scm.client.importz.internal.ChangeSetArchiveImporter;
import com.ibm.team.scm.client.importz.internal.IChangeArchiveStatistics;
import com.ibm.team.scm.client.importz.internal.IConstants;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;
import com.ibm.team.scm.client.importz.spi.ImportData;
import com.ibm.team.scm.client.importz.spi.MigrationFactory;

public class ChangeArchiveImportWizard extends AbstractImportWizard implements IImportWizard {

    private final class EmbeddedStatisticsArea extends StatisticsArea {
		private EmbeddedStatisticsArea(Composite parent, ImportConfiguration configuration) {
			super(parent, configuration);
		}

		@Override
		protected void showStatistics() {
			try {
				getContainer().run(true, true, new IRunnableWithProgress() {
					public void run(IProgressMonitor monitor) throws InvocationTargetException,
							InterruptedException {
				    	try {
							final IChangeArchiveStatistics stats = ChangeArchiveImportWizard.this.getStatistics(monitor);
							if (stats != null) {
								Display.getDefault().asyncExec(new Runnable() {
									public void run() {
										EmbeddedStatisticsArea.this.setStatistics(stats.toDescriptionString());
									}
								});
							}
						} catch (IOException e) {
							throw new InvocationTargetException(e);
						} catch (TeamRepositoryException e) {
							throw new InvocationTargetException(e);
						}
					}
				});
			} catch (InvocationTargetException e) {
				ImportLoggingHelper.displayError(SCMImportMessages.ChangeArchiveImportWizard_0, e.getTargetException());
			} catch (InterruptedException e) {
				// Ignore
			}
		}

		@Override
		protected void updateEnablements() {
			// Nothing to do
		}
	}

	private ConfigureArchiveFileImportPage configureArchiveImportPage;
    private String initialFilename;

    public ChangeArchiveImportWizard() {
        setWindowTitle(SCMImportMessages.ChangeArchiveImportWizard_1);
        setNeedsProgressMonitor(true);
    }


    protected ImportConfiguration createConfiguration() {
    	return new ImportConfiguration(getSourceRepositoryKind(), true);
    }
    
    protected String getDefaultFilename() {
    	if (initialFilename != null)
    		return initialFilename;
    	return AbstractImportWizard.DEFAULT_FILENAME;
    }
    
    @Override
    protected void addPagesAtBeginning() {
        configureArchiveImportPage = new ConfigureArchiveFileImportPage("ConfigureArchiveFileImportPage", SCMImportMessages.ChangeArchiveImportWizard_3, getBanner(), //$NON-NLS-1$
        		getConfiguration()) {
        	@Override
        	protected StatisticsArea createStatisticsGroup(
        			Composite composite, ImportConfiguration configuration) {
        		return new EmbeddedStatisticsArea(composite, configuration);
        	}
        };
        configureArchiveImportPage.setDescription(SCMImportMessages.ChangeArchiveImportWizard_4);
        addPage(configureArchiveImportPage);
    }
    
	public boolean canFinish() {
        if (!configureArchiveImportPage.isPageComplete())
            return false;
        return super.canFinish();
    }
    
    public void init(IWorkbench workbench, IStructuredSelection selection) {
    	super.init(workbench, selection);
        for (Iterator i = selection.iterator(); i.hasNext();) {
            Object o = i.next();
            
            if (o instanceof File) {
                initialFilename = ((File)o).getAbsolutePath();
                break;
            }
            
            if (o instanceof IAdaptable) {
                IAdaptable a = (IAdaptable)o;
                IFile file = (IFile)a.getAdapter(IFile.class);
                if (file != null) {
                    IPath path = file.getLocation();
                    if (path != null) {
                        initialFilename = path.toOSString();
                        break;
                    }
                }
                
                File f = (File)a.getAdapter(File.class);
                if (f != null) {
                    initialFilename = f.getAbsolutePath();
                    break;
                }
            }
        }
    }

	private IChangeArchiveStatistics getStatistics(IProgressMonitor monitor) throws IOException, TeamRepositoryException {
		String filename = getConfiguration().getArchiveFilename();
		final File file = new File(filename);
		if (!file.exists()) {
			ImportLoggingHelper.displayError(SCMImportMessages.ChangeArchiveImportWizard_Error, NLS.bind(SCMImportMessages.ChangeArchiveImportWizard_6, filename));
		    return null;
		}
		final IMigrationFactory factory = getMigrationFactory();
		ImportData data = new ImportData();
		data.setArchiveFile(file);
		boolean closeArchive = false;
		try {
			closeArchive = data.openArchive(null);
			ChangeSetArchiveImporter importer = new ChangeSetArchiveImporter(factory, data);
			importer.importChangeSetArchive(null, null, true, monitor);
			return importer.getStats();
		} finally {
			if (closeArchive)
				data.closeArchive();
		}
	}

	protected IMigrationFactory getMigrationFactory() {
		return new MigrationFactory() {
			public String getName() {
				return SCMImportMessages.MigrationFactory_0;
			}
			public IChangeSetArchiveCreator createChangeSetArchiveCreator() {
				// We don't need this since we already have an archive
				return null;
			}
		};
	}
	

	@Override
	protected ImportOperation createImportOperation() {
		return new ImportOperation(SCMImportMessages.ChangeArchiveImportWizard_2, getMigrationFactory(), getConfiguration());
	}
	
	@Override
	protected String getSourceRepositoryKind() {
		return SCMImportMessages.ChangeArchiveImportWizard_5;
	}

	@Override
	protected ImageDescriptor getBanner() {
		return AbstractUIPlugin.imageDescriptorFromPlugin(IConstants.PLUGIN_ID, IConstants.CHANGE_SET_ARCHIVE_IMPORT_WIZARD_BANNER);
	}
}
