/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.ListenerList;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.IWorkspaceManager;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;
import com.ibm.team.scm.client.importz.internal.utils.DebugUtils;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IComponentHandle;

/**
 * An <code>ImportConfiguration</code> manages the information that controls
 * an import of data from another repository type into SCM. 
 * <p>
 * This class may be instantiated or subclasses by clients.
 */
public class ImportConfiguration {
	
	public static final String DEFAULT_ENCODING = "ISO-8859-1"; //$NON-NLS-1$

	private static final String PLUGIN_ID = "com.ibm.team.scm.client.importz"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate changes to the "import into
	 * workspace" property of this configuration.
	 * 
	 * @see #isImportIntoWorkspace()
	 * @see #setImportIntoWorkspace(boolean)
	 */
	public static final String IMPORT_INTO_WORKSPACE = "importIntoWorkspace"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate changes to the "import full history" 
	 * property of this configuration.
	 * 
	 * @see #isImportFullHistory()
	 * @see #setImportFullHistory(boolean)
	 */
	public static final String IMPORT_FULL_HISTORY = "importFullHistory"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate changes to the "delete archive" 
	 * property of this configuration.
	 * 
	 * @see #isDeleteArchiveFile()
	 * @see #setDeleteArchiveFile(boolean)
	 */
	public static final String DELETE_ARCHIVE_FILE = "deleteArchiveFile"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate changes to the "archive
	 * file name" property of this configuration.
	 * 
	 * @see #getArchiveFilename()
	 * @see #setArchiveFilename(String)
	 */
	public static final String ARCHIVE_FILENAME = "archiveFilename"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate that a single folder is being
	 * imported.
	 * 
	 * @see #isSingleFolderSelected()
	 * @see #setSingleFolderSelected(boolean)
	 */
	public static final String SINGLE_FOLDER_SELECTED = "singleFolderSelected"; //$NON-NLS-1$
	
	/**
	 * Property constant that is used to indicate that, when a single folder is being
	 * imported, it should be mapped to the component root. This property is ignored
	 * if {@link #isSingleFolderSelected()} returns false.
	 * 
	 * @see #isMapFolderToComponent()
	 * @see #setMapFolderToComponent(boolean)
	 */
	public static final String MAP_FOLDER_TO_COMPONENT = "mapFolderToComponent"; //$NON-NLS-1$

	public static final String POST_IMPORT_BASELINE_NAME = "postImportBaselineName"; //$NON-NLS-1$

	public static final String TEAM_REPOSITORY_PROPERTY = "teamReposiotry"; //$NON-NLS-1$
	
	public static final String DEFAULT_ENCODING_PROPERTY = "defaultEncoding"; //$NON-NLS-1$
	
	private boolean importIntoWorkspace = true;
	private boolean fullHistory = true;
	private String filename;
	private boolean deleteArchiveFile = true;
	private boolean importFromArchive;
	private ListenerList listeners = new ListenerList(ListenerList.IDENTITY);
	private ITeamRepository repository;
	private String workspaceName;
	private IComponent component;
	private IWorkspaceConnection workspaceConnection;
	private Map<String, Object> properties = new HashMap<String, Object>();

	private File file;

	private final String sourceRepositoryKind;

	private Collection<Author2ContributorMapping> authorMappings;

	private boolean incrementalBaselinesSupported;

	private boolean createIncrementalBaselines;
	
	/**
	 * Create an import configuration
	 * 
	 * @param importFromArchive
	 *            indicates whether the import is from an archive or from a
	 *            repository
	 */
	public ImportConfiguration(String sourceRepositoryKind, boolean importFromArchive) {
		this.sourceRepositoryKind = sourceRepositoryKind;
		this.importFromArchive = importFromArchive;
		if (importFromArchive)
			deleteArchiveFile = false;
		setDefaultEncoding(DEFAULT_ENCODING);
	}
	
	/**
	 * Add a property change listener to this configuration. Adding a 
	 * listener that is already added has no effect.
	 * @param listener a property change listener
	 */
	public void addPropertyChangeListener(IPropertyChangeListener listener) {
		listeners.add(listener);
	}
	
	/**
	 * Remove the provided property change listener from this configuration.
	 * Removing a listener that is not registered has no effect.
	 * @param listener a property change listener
	 */
	public void removePropertyChangeListener(IPropertyChangeListener listener) {
		listeners.remove(listener);
	}
	
	/**
	 * Fire a property change event for the given property.
	 * @param propertyName the name of the property that changed
	 * @param oldValue the previous value of the property
	 * @param newValue the new value of the property
	 */
	protected void firePropertyChange(String propertyName,
			Object oldValue, Object newValue) {
		Object[] list = listeners.getListeners();
		for (int i = 0; i < list.length; i++) {
			final IPropertyChangeListener listener = (IPropertyChangeListener)list[i];
			listener.propertyChange(new PropertyChangeEvent(this, propertyName, oldValue, newValue));
		}
	}
	
	/**
	 * Return whether the import should be performed directly into an SCM
	 * workspace. If <code>false</code>, the import information is written to
	 * an intermediate archive file whose path is obtained from 
	 * {@link #getArchiveFilename()}.
	 * 
	 * @return whether the import should be performed directly into an SCM
	 * workspace
	 */
	public boolean isImportIntoWorkspace() {
		return importIntoWorkspace;
	}

	/**
	 * Set whether the import should be performed directly into an SCM
	 * workspace.
	 * @param importIntoWorkspace whether the import should be performed directly into an SCM
	 * workspace
	 * @see #isImportIntoWorkspace()
	 */
	public void setImportIntoWorkspace(boolean importIntoWorkspace) {
		if (this.importIntoWorkspace != importIntoWorkspace) {
			this.importIntoWorkspace = importIntoWorkspace;
			firePropertyChange(IMPORT_INTO_WORKSPACE, Boolean.valueOf(!importIntoWorkspace), Boolean.valueOf(importIntoWorkspace));
		}
		if (!importIntoWorkspace)
			setDeleteArchiveFile(false);
	}

	/**
	 * Return whether the full history of the data should be imported. If <code>false</code>,
	 * only the current state of the source repository is imported.
	 * @return whether the full history of the data should be imported
	 */
	public boolean isImportFullHistory() {
		return fullHistory;
	}

	/**
	 * Set whether the full history of the data should be imported.
	 * @param fullHistory whether the full history of the data should be imported
	 * @see #isImportFullHistory()
	 */
	public void setImportFullHistory(boolean fullHistory) {
		if (this.fullHistory != fullHistory) {
			this.fullHistory = fullHistory;
			firePropertyChange(IMPORT_FULL_HISTORY, Boolean.valueOf(!fullHistory), Boolean.valueOf(fullHistory));
		}
	}

	/**
	 * Return the filename path that is to be used for the archive file of the import.
	 * A filename is required is in the following cases:
	 * <ul>
	 * <li>{@link #isImportIntoWorkspace()} is <code>false</code> in which case the 
	 *   results of the import are written to the archive file.</li>
	 * <li>{@link #isDeleteArchiveFile()} is <code>false</code>, in which case the archive
	 * is to be kept around after the import is finished.</li>
	 * <li>{@link #isImportFromArchive()} is <code>true</code>, in which case the archive
	 * is the input to the import process.
	 * </ul>
	 * @return the filename path that is to be used for the archive file of the import
	 */
	public String getArchiveFilename() {
		return filename;
	}
	
	/**
	 * Return the file handle for the archive file to be used during the import.
	 * If {@link #isDeleteArchiveFile()} is <code>true</code> then a call to this
	 * method creates a temporary file. The caller is responsible for deleting this file
	 * once the import is complete.
	 * @return the file handle for the archive file to be used during the import
	 * @throws IOException
	 */
	public File getArchiveFile() throws IOException {
        if (filename == null || (isDeleteArchiveFile() && !isImportFromArchive())) {
        	if (file == null) {
        		file = DebugUtils.createTempFile("importArchive"); //$NON-NLS-1$
        	}
			return file;
        } else {
            return new File(getArchiveFilename());
        }
	}
	
	/**
	 * Set the filename path that is to be used for the archive file of the import.
	 * The provided filename should be an absolute OS file path.
	 * @param filename the filename path that is to be used for the archive file of the import
	 * @see #getArchiveFilename()
	 */
	public void setArchiveFilename(String filename) {
		if (this.filename == null || !this.filename.equals(filename)) {
			String oldName = this.filename;
			this.filename = filename;
			firePropertyChange(ARCHIVE_FILENAME, oldName, filename);
		}
	}

	/**
	 * Indicate whether the archive file should be deleted after the import
	 * completes. This method should only be consulted if {@link #canDeleteArchive()}
	 * is <code>true</code>.
	 * @return whether the archive file should be deleted after the import
	 * completes
	 * @see #canDeleteArchive()
	 */
	public boolean isDeleteArchiveFile() {
		return deleteArchiveFile;
	}

	/**
	 * Set whether the archive file should be deleted after the import
	 * completes
	 * @param deleteArchiveFile whether the archive file should be deleted after the import
	 * completes
	 */
	public void setDeleteArchiveFile(boolean deleteArchiveFile) {
		if (deleteArchiveFile)
			Assert.isTrue(!isImportFromArchive(), SCMImportMessages.ImportConfiguration_6);
		if (this.deleteArchiveFile != deleteArchiveFile) {
			this.deleteArchiveFile = deleteArchiveFile;
			firePropertyChange(DELETE_ARCHIVE_FILE, Boolean.valueOf(!deleteArchiveFile), Boolean.valueOf(deleteArchiveFile));
		}
	}

	/**
	 * Return whether it is safe to delete the archive file.
	 * @return whether it is safe to delete the archive file
	 */
	public boolean canDeleteArchive() {
		return isImportFullHistory() && isImportIntoWorkspace() && !isImportFromArchive();
	}

	/**
	 * Return whether this import is from an archive file or from a source repository.
	 * @return whether this import is from an archive file
	 * @see #getArchiveFilename()
	 */
	public boolean isImportFromArchive() {
		return importFromArchive;
	}

	/**
	 * Set whether this import is from an archive file or from a source repository.
	 * @param importFromArchive whether this import is from an archive file
	 * @see #isImportFromArchive()
	 */
	public void setImportFromArchive(boolean importFromArchive) {
		this.importFromArchive = importFromArchive;
	}
	
	/**
	 * Validate the state of this configuration and return a status indicating any
	 * problems that prevent the import from being performed.
	 * @return the state of this configuration
	 */
	public IStatus validateState() {
		String filename = getArchiveFilename();
		if (isImportFromArchive()) {
	        if (filename == null) {
	            return new Status(IStatus.ERROR, PLUGIN_ID, 0, SCMImportMessages.ImportConfiguration_7, null);
	        }
	        File file = new File(filename);
	        if (!file.exists()) {
	            return new Status(IStatus.ERROR, PLUGIN_ID, 0, NLS.bind(SCMImportMessages.ImportConfiguration_8, filename), null);
	        }
		} else if (filename != null && !isDeleteArchiveFile()) {
			if (filename.length() == 0) {
				return new Status(IStatus.ERROR, PLUGIN_ID, 0, SCMImportMessages.ImportConfiguration_9, null);
			}
			final File file = new File(filename);
			File parent = file.getParentFile();
			if (parent == null) {
				return new Status(IStatus.ERROR, PLUGIN_ID, 0, SCMImportMessages.ImportConfiguration_10, null);
			}
			if (!parent.exists()) {
				return new Status(IStatus.ERROR, PLUGIN_ID, 0, NLS.bind(SCMImportMessages.ImportConfiguration_11, filename), null);
			}
			if (!parent.isDirectory()) {
				return new Status(IStatus.ERROR, PLUGIN_ID, 0, NLS.bind(SCMImportMessages.ImportConfiguration_12, filename), null);
			}
			if (file.exists()) {
				if (file.isDirectory())
					return new Status(IStatus.ERROR, PLUGIN_ID, 0, NLS.bind(SCMImportMessages.ImportConfiguration_13, filename), null);
				return new Status(IStatus.WARNING, PLUGIN_ID, 0, NLS.bind(SCMImportMessages.ImportConfiguration_14, filename), null);
			}
		}
		return Status.OK_STATUS;
	}

	public void setRepository(ITeamRepository repository) {
		if (this.repository != repository) {
			ITeamRepository oldRepository = this.repository;
			this.repository = repository;
			firePropertyChange(TEAM_REPOSITORY_PROPERTY, oldRepository, repository);
		}
	}

	public ITeamRepository getRepository() {
		return repository;
	}

	public void setWorkspaceName(String workspaceName) {
		this.workspaceName = workspaceName;
	}

	public String getWorkspaceName() {
		return workspaceName;
	}

	public void setTargetComponent(IWorkspaceConnection workspaceConnection, IComponent component) {
		Assert.isNotNull(workspaceConnection);
		Assert.isNotNull(component);
		this.component = component;
		this.workspaceConnection = workspaceConnection;
	}

	public IComponent getComponent() {
		return component;
	}

	public IWorkspaceConnection getWorkspaceConnection() {
		return workspaceConnection;
	}

	public boolean isAllowSync() {
		// TODO Currently not supported through the UI
		return false;
	}
	
	/**
	 * Return the component that is the target of the import or <code>null</code>
	 * if {@link #isImportIntoWorkspace()} is <code>false</code>. If no 
	 * workspace connection has been set using 
	 * {@link #setTargetComponent(IWorkspaceConnection, IComponentHandle)}then a workspace and component is
	 * created using {@link #getWorkspaceName()} as the name.
	 * Otherwise, the component is fetched.
	 * @param monitor a progress monitor.
	 * @throws TeamRepositoryException
	 */
	public IComponent getComponent(IProgressMonitor monitor) throws TeamRepositoryException {
		if (!isImportIntoWorkspace())
			return null;
		SubMonitor progress = SubMonitor.convert(monitor, 30);
    	if (workspaceConnection == null) {
            IWorkspaceManager wm = SCMPlatform.getWorkspaceManager(getRepository());

            progress.subTask(NLS.bind(SCMImportMessages.ImportConfiguration_15, getWorkspaceName()));
            workspaceConnection = wm.createWorkspace(
            		getRepository().loggedInContributor(), 
            		getWorkspaceName(), 
            		"",  //$NON-NLS-1$
            		progress.newChild(15));
            IComponent component = wm.createComponent(
            		getWorkspaceName(), 
            		getRepository().loggedInContributor(), 
            		progress.newChild(15));
			this.component = component;
            workspaceConnection.addComponent(component, false, progress.newChild(10));
    	}
    	return component;
	}

	public void setProperty(String key, Object value) {
		Object oldValue = getProperty(key);
		if (oldValue != value) {
			if (oldValue != null && value != null && oldValue.equals(value)) {
				return;
			}
			if (value == null)
				properties.remove(key);
			else
				properties .put(key, value);
			firePropertyChange(key, oldValue, value);
		}
	}

	public Object getProperty(String key) {
		return properties.get(key);
	}
	
	public Object getProperty(String key, Object defaultValue) {
		Object object = getProperty(key);
		if (object == null) {
			return defaultValue;
		}
		return object;
	}

	public String getSourceRepositoryKind() {
		return sourceRepositoryKind;
	}

	public void setAuthorMappings(
			Collection<Author2ContributorMapping> authorMappings) {
		this.authorMappings = authorMappings;
	}
	
	public Collection<Author2ContributorMapping> getAuthorMappings() {
		return authorMappings;
	}

	public String getPostImportBaselineName() {
		return (String)getProperty(POST_IMPORT_BASELINE_NAME);
	}
	
	public void setPostImportBaselineName(String name) {
		setProperty(POST_IMPORT_BASELINE_NAME, name);
	}

	public boolean isIncrementalBaselinesSupported() {
		return incrementalBaselinesSupported;
	}
	
	public void setIncrementalBaselinesSupported(boolean incrementalBaselinesSupported) {
		this.incrementalBaselinesSupported = incrementalBaselinesSupported;
	}

	public boolean isCreateIncrementalBaselines() {
		return createIncrementalBaselines;
	}
	
	public void setCreateIncrementalBaselines(boolean createIncrementalBaselines) {
		this.createIncrementalBaselines = createIncrementalBaselines;
	}

	public String getDefaultEncoding() {
		return (String)getProperty(DEFAULT_ENCODING_PROPERTY);
	}

	public void setDefaultEncoding(String encoding) {
		setProperty(DEFAULT_ENCODING_PROPERTY, encoding);
	}

	public boolean isMapFolderToComponent() {
		return ((Boolean)getProperty(MAP_FOLDER_TO_COMPONENT, Boolean.FALSE)).booleanValue();
	}

	public void setMapFolderToComponent(boolean mapFolderToComponent) {
		setProperty(MAP_FOLDER_TO_COMPONENT, mapFolderToComponent);
	}

	public boolean isSingleFolderSelected() {
		return ((Boolean)getProperty(SINGLE_FOLDER_SELECTED, Boolean.FALSE)).booleanValue();
	}
	
	public void setSingleFolderSelected(boolean singleFolderSelected) {
		setProperty(SINGLE_FOLDER_SELECTED, singleFolderSelected);
	}
}
