/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.ui;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.jface.window.Window;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Display;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.foundation.common.internal.util.TeamFoundationException;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.rcp.core.utils.StatusUtil;
import com.ibm.team.repository.rcp.ui.internal.utils.JFaceUtils;
import com.ibm.team.repository.rcp.ui.operations.IStatusCollector;
import com.ibm.team.repository.rcp.ui.operations.RepositoryOperation;
import com.ibm.team.scm.client.importz.IChangeSetArchiveImporter;
import com.ibm.team.scm.client.importz.IChangeSetFileReader;
import com.ibm.team.scm.client.importz.IImportChangeSet;
import com.ibm.team.scm.client.importz.IImportData;
import com.ibm.team.scm.client.importz.IMigrationFactory;
import com.ibm.team.scm.client.importz.internal.CreateArchiveOperation;
import com.ibm.team.scm.client.importz.internal.IChangeArchiveStatistics;
import com.ibm.team.scm.client.importz.internal.ImportArchiveOperation;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;
import com.ibm.team.scm.client.importz.internal.utils.DebugUtils;
import com.ibm.team.scm.client.importz.spi.ImportData;
import com.ibm.team.scm.common.IComponent;

public class ImportOperation extends RepositoryOperation {

	private IChangeArchiveStatistics stats;
	private final ImportConfiguration configuration;
	private final IMigrationFactory factory;
	private final IImportData importData;
	private final String name;
	private boolean promptForContributors = false;

	public static void recordAuthorMappings(
			Collection<Author2ContributorMapping> authorMappings,
			ImportData data) {
		for (Author2ContributorMapping mapping : authorMappings) {
			data.addAuthorMapping(mapping.getAuthorId(), mapping.getContributorId(), mapping.getContributor());
		}
	}
	
	public ImportOperation(String name, IMigrationFactory factory, IImportData importData,
			ImportConfiguration configuration) {
		this.name = name;
		this.factory = factory;
		this.importData = importData;
		this.configuration = configuration;
		initializeData();
	}

	public ImportOperation(String name, IMigrationFactory migrationFactory, ImportConfiguration configuration) {
		this(name, migrationFactory, new ImportData(), configuration);
	}

	protected void initializeData() {
		/*
		 * Determine whether we need to prompt for user mappings or not. If the auto options was chosen,
		 * the generation of data will be performed later.
		 */
		String property = (String)configuration.getProperty(Author2ContributorMappingPage.AUTHOR_MAPPING_PROPERTY);
		boolean noMapping = property != null && property.equals(Author2ContributorMappingPage.NO_MAPPING);
		boolean mappingsProvided = property != null && property.equals(Author2ContributorMappingPage.USE_PROVIDED_MAPPINGS);
		boolean prompt = property == null || property.equals(Author2ContributorMappingPage.PROMPT_FOR_MAPPINGS);
		ImportData data = (ImportData)getImportData();
		if (noMapping) {
			data.setContributorMappingEnablement(false);
		} else if (mappingsProvided) {
			recordAuthorMappings(getConfiguration().getAuthorMappings(), data);
		} else if (prompt) {
			promptForContributors = true;
		}
		data.setPostImportBaselineName(configuration.getPostImportBaselineName());
		Object importOverExisting = configuration.getProperty(ReconcileOptionsArea.RECONCILE_WITH_EXISTING);
		if (importOverExisting instanceof Boolean) {
			Boolean b = (Boolean) importOverExisting;
			// Only set to true if the an existing component is that target of the import
			data.setImportOverExisting(configuration.getComponent() != null && b.booleanValue());
		}
		data.setWorkspaceEncoding(configuration.getDefaultEncoding());
	}

	@Override
	public void repositoryRun(IProgressMonitor monitor,
			IStatusCollector problems) throws TeamRepositoryException,
			FileSystemException {
		SubMonitor progress = SubMonitor.convert(monitor, NLS.bind(SCMImportMessages.ImportOperation_0, getFactory().getName()), configuration.isImportIntoWorkspace() ? 100 : 50);
        try {
			((ImportData)getImportData()).setArchiveFile(configuration.getArchiveFile());
			if (configuration.isImportFromArchive()) {
				importIntoWorkspace(getFactory(), getImportData(), progress.newChild(100));
			} else {
				try {
					createArchive(getFactory(), getImportData(), progress.newChild(50));
					if (progress.isCanceled())
						return;
					if (configuration.isImportIntoWorkspace()) {
			        	importIntoWorkspace(getFactory(), getImportData(), progress.newChild(50));
					}
				} finally {
					if (configuration.isDeleteArchiveFile())
						DebugUtils.deleteTempFile(((ImportData)getImportData()).getArchiveFile());
				}
			}
		} catch (IOException e) {
			throw new TeamRepositoryException(e);
		} catch (TeamRepositoryException e) {
			if (TeamFoundationException.isExpected(e)) {
				// If the exception is recognized, we want to log it since the framework won't.
				// The main problem is that the isExpected method goes deep looking for the cause
				// but the dialog displaying the error does not
				StatusUtil.log(this, e);
			}
			throw e;
		}
	}
	
	/**
	 * Method called from {@link #doImport(IMigrationFactory, IImportData, IProgressMonitor)}
	 * when creating a change set archive during import.
	 * @param factory the migration factory that provides SCM specific behavior
	 * @param importData SCM specific import data which includes the source of the import
	 * @param file the output archive file
	 * @param progress a progress monitor
	 * @throws TeamRepositoryException
	 */
	protected void createArchive(IMigrationFactory factory,
			IImportData importData, SubMonitor progress)
			throws TeamRepositoryException {
		CreateArchiveOperation createArchive = new CreateArchiveOperation(factory, importData);
		createArchive.run(progress);
	}

	/**
	 * Method called from {@link #doImport(IMigrationFactory, IImportData, IProgressMonitor)}
	 * when importing a change set archive file.
	 * @param factory the migration factory that provides SCM specific behavior
	 * @param importData SCM specific import data
	 * @param file the change set archive file
	 * @param monitor a progress monitor
	 * @throws TeamRepositoryException
	 * @throws FileSystemException
	 */
	protected void importIntoWorkspace(IMigrationFactory factory,
			IImportData importData,
			SubMonitor monitor) throws TeamRepositoryException,
			FileSystemException {
		boolean closeArchive = false;
		try {
			monitor.beginTask(null, 100);
			closeArchive = importData.openArchive(monitor.newChild(1));
			if (isOkToPrompt()) {
				promptToConfigureArchiveImport(importData, Author2ContributorMappingArea.hasJazzAdminPermissions(configuration.getRepository(), monitor.newChild(1)), monitor.newChild(1));
			}
			if (isAutoMapUsers()) {
				performAutoMapOfUsers(monitor.newChild(1));
			}
			IComponent component = configuration.getComponent(monitor.newChild(3));
			monitor.setTaskName(NLS.bind(SCMImportMessages.ImportOperation_3, configuration.getSourceRepositoryKind(), component.getName()));
			IChangeSetArchiveImporter importer = factory.createArchiveImporter(importData, configuration.getWorkspaceConnection(), component);
			importer.performImport(monitor.newChild(95));
			if (importer instanceof ImportArchiveOperation) {
				ImportArchiveOperation importArchiveOp = (ImportArchiveOperation) importer;			
				if (importArchiveOp.getStats() != null) {
					setStats(importArchiveOp.getStats());
					if (importArchiveOp.getStats().getChangeSetCount() == 0) {
						handleNoChanges();
					}
				}
			}
		} finally {
			if (closeArchive)
				importData.closeArchive();
			monitor.done();
		}
	}

	protected void performAutoMapOfUsers(IProgressMonitor monitor) throws TeamRepositoryException {
		try {
			SubMonitor progress = SubMonitor.convert(monitor, 100);
			final Set<String> authors = extractAuthorsFromArchiveFile(importData, progress.newChild(25));
			if (!authors.isEmpty()) {
				Collection<Author2ContributorMapping> mappings = createMappings(authors);
				Map<Author2ContributorMapping, Author2ContributorMapping> result = FetchContributorsFunction.fetchContributors(configuration.getRepository(), mappings, progress.newChild(25));
				configuration.setAuthorMappings(result.values());
				Author2ContributorMappingArea.createContributors(configuration.getRepository(), configuration.getAuthorMappings(), progress.newChild(50));
				recordAuthorMappings(configuration.getAuthorMappings(), (ImportData)importData);
			}
		} catch (InvocationTargetException e) {
			Throwable targetException = e.getTargetException();
			if (targetException instanceof TeamRepositoryException) {
				throw (TeamRepositoryException) targetException;
			}
			throw new TeamRepositoryException(e);
		} catch (IOException e) {
			throw new TeamRepositoryException(e);
		} finally {
			monitor.done();
		}
	}

	protected boolean isAutoMapUsers() {
		String property = (String)configuration.getProperty(Author2ContributorMappingPage.AUTHOR_MAPPING_PROPERTY);
		return property != null && property.equals(Author2ContributorMappingPage.MAP_MATCHING_IDS);
	}

	protected void handleNoChanges() {
		if (isOkToPrompt())
			JFaceUtils.showMessage(SCMImportMessages.ImportOperation_1, SCMImportMessages.ImportOperation_2, IStatus.WARNING);
	}
	
	private void setStats(IChangeArchiveStatistics stats) {
		this.stats = stats;
	}

	public IChangeArchiveStatistics getStats() {
		return stats;
	}

	public IMigrationFactory getFactory() {
		return factory;
	}

	public IImportData getImportData() {
		return importData;
	}

	public ImportConfiguration getConfiguration() {
		return configuration;
	}
	
	/**
	 * Method used by tests to disable any prompting.
	 * @return
	 */
	protected boolean isOkToPrompt() {
		return true;
	}

	public String getName() {
		return name;
	}

	public boolean isPromptForContributors() {
		return promptForContributors;
	}
	
	/**
	 * Method that is called after the archive file is created but before
	 * it is imported into the workspace in order to allow subclasses to
	 * obtaining additional information from the user if required. The 
	 * default implementation prompts for author to contributor mappings
	 * if {@link #extractAuthorsFromArchiveFile(File, IProgressMonitor)} is
	 * overridden by subclasses to return a set of authors for the data being
	 * imported.
	 * @param importData the import data
	 * @param isAdmin whether the logged in user has admin privileges
	 * @param monitor a progress monitor
	 * @param file the change set archive file in TAR format
	 * @throws TeamRepositoryException
	 */
    protected void promptToConfigureArchiveImport(final IImportData importData,
    		final boolean isAdmin, SubMonitor monitor) throws TeamRepositoryException {
    	if (shouldPromptForAuthorMappings(monitor)) {
    		try {
				final Set<String> authors = extractAuthorsFromArchiveFile(importData, monitor);
				if (!authors.isEmpty()) {
					Collection<Author2ContributorMapping> mappings = createMappings(authors);
					Map<Author2ContributorMapping, Author2ContributorMapping> result = FetchContributorsFunction.fetchContributors(configuration.getRepository(), mappings, monitor);
					recordAuthorMappings(result.values(), (ImportData)importData);
					configuration.setAuthorMappings(result.values());
					final boolean[] ok = new boolean [] { true };
					Display.getDefault().syncExec(new Runnable() {
						public void run() {
							ok[0] = promptForAuthorMappings(importData, isAdmin);
						}
					});
					if (!ok[0]) {
						throw new OperationCanceledException();
					}
				}
			} catch (IOException e) {
				StatusUtil.log(this, e);
			}
    	}
    }
	
    private Collection<Author2ContributorMapping> createMappings(
			Set<String> authors) {
    	List<Author2ContributorMapping> result = new ArrayList<Author2ContributorMapping>();
    	for (String string : authors) {
    		result.add(new Author2ContributorMapping(string, string));
		}
		return result;
	}

	protected boolean promptForAuthorMappings(IImportData importData, boolean isAdmin) {
		AuthorMappingWizard wizard = new AuthorMappingWizard(getConfiguration(), importData, isAdmin);
		WizardDialog dialog = new WizardDialog(null /* no shell since the original wizard is closed */, wizard);
		int result = dialog.open();
		return result == Window.OK;
	}

	/**
     * Method invoked from {@link #promptToConfigureArchiveImport(IImportData, boolean, SubMonitor)}
     * to extract the authors from the import data or the change set archive file.
     * By default, the authors are extract from the archived log file but subclasses may override.
	 * @param importData the import data
	 * @param file the change set archive file in TAR format
	 * @param monitor a progress monitor
     * @return the authors (user IDs) of the changes being imported
	 * @throws IOException 
	 * @throws TeamRepositoryException 
     */
	protected Set<String> extractAuthorsFromArchiveFile(IImportData importData,
			IProgressMonitor monitor) throws IOException, TeamRepositoryException {
    	boolean closeArchive = false;
		try {
			closeArchive = importData.openArchive(monitor);
			IChangeSetFileReader reader = factory.createChangeSetFileReader(importData);
			try {
				IImportChangeSet changeSet;
				Set<String> authors = new HashSet<String>();
				while ((changeSet = reader.next(monitor)) != null) {
					authors.add(changeSet.getAuthor());
				}
				return authors;
			} finally {
				reader.close();
			}
		} finally {
			if (closeArchive)
				importData.closeArchive();
		}
	}

	/**
	 * Return whether prompting for author ID to contributor mapping
	 * should occur.
	 * @return whether prompting for author ID to contributor mapping
	 * should occur
	 */
	protected boolean shouldPromptForAuthorMappings(IProgressMonitor monitor) {
		Object property = getConfiguration().getProperty(Author2ContributorMappingPage.AUTHOR_MAPPING_PROPERTY);
		if (property == null) {
			ITeamRepository repository = configuration.getRepository();
			if (repository != null) {
				boolean isAdmin = Author2ContributorMappingArea.hasJazzAdminPermissions(repository, monitor);
				if (!isAdmin) {
					// No point prompting if the current user doesn't have admin privileges
					return false;
				}
			}
		}
		return property == null || property.equals(Author2ContributorMappingPage.PROMPT_FOR_MAPPINGS);
	}

}
