/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.rcp.ui.internal.util.PathUtils;
import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.scm.client.IConfiguration;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.importz.IImportChange;
import com.ibm.team.scm.client.importz.internal.SCMImportMessages;
import com.ibm.team.scm.client.importz.spi.ImportChange;
import com.ibm.team.scm.common.ComponentNotInWorkspaceException;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Default implementation of {@link IFileTree}
 */
public class FileTree implements IFileTree {
	
	private FolderEntry root;
    private final IWorkspaceConnection workspace;
    private final IComponent component;
    private boolean alwaysInitialize = DebugUtils.ALWAYS_INITIALIZE_IMPORT_TREE;

	private abstract static class AbstractEntry {
    	private String name;
    	public String getName() {
			return name;
		}

		private FolderEntry parent;
    	private IVersionableHandle itemHandle;
        
		public AbstractEntry(String name, FolderEntry parent, IVersionableHandle itemHandle) {
            this.name = name;
            this.parent = parent;
            this.itemHandle = itemHandle;
        }
		
        public IVersionableHandle getItemHandle() {
			return itemHandle;
		}

		public IVersionable getItem() {
        	if (itemHandle instanceof IVersionable) {
				return (IVersionable) itemHandle;
			}
			return null;
		}

		public void setItem(IVersionable item) {
			this.itemHandle = item;
		}
        
        public String getPath() {
            if (parent == null) {
                return name;
            } else {
                return parent.getPath() + "/" + name; //$NON-NLS-1$
            }
        }
        
        public abstract boolean isContainer();

		public FolderEntry getParent() {
			return parent;
		}

		public void setParent(FolderEntry parent) {
			this.parent = parent;
		}

		public void prepareForChange() {
			IVersionable item = getItem();
			if (item != null) {
				setItem((IVersionable)item.getWorkingCopy());
			}
		}

		public void setName(String name) {
			this.name = name;
		}
    }

    private static class FileEntry extends AbstractEntry {
        public FileEntry(String name, FolderEntry parent, IVersionableHandle item) {
            super(name, parent, item);
        }
        
        public boolean isContainer() {
            return false;
        }
    }

    private static class FolderEntry extends AbstractEntry {
        private Map<String, AbstractEntry> children;
        
        public FolderEntry(String name, FolderEntry parent, IVersionableHandle item) {
            super(name, parent, item);
        }
        
        public boolean isContainer() {
            return true;
        }

    	protected Collection<AbstractEntry> setChildren(Map<String, IVersionableHandle> handlesMap) {
    		this.children = new LinkedHashMap<String, AbstractEntry>();
    		if (handlesMap != null) {
	    		for (Entry<String, IVersionableHandle> entry : handlesMap.entrySet()) {
	    			AbstractEntry childEntry = createEntry(entry.getKey(), entry.getValue());
	    			children.put(childEntry.getName(), childEntry);
	    		}
    		}
    		return children.values();
    	}
    	
		public boolean hasChildren() {
			ensureChildrenKnown();
			return !children.isEmpty();
		}

		private void ensureChildrenKnown() {
			if (children == null) {
				throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_0, getPath()));
			}
		}

		public void remove(String name) {
			ensureChildrenKnown();
			AbstractEntry entry = children.get(name);
			if (entry != null) {
				children.remove(name);
				entry.setParent(null);
			}
		}

		public AbstractEntry getChild(String name) {
			ensureChildrenKnown();
			return children.get(name);
		}

		public void add(AbstractEntry entry) {
			ensureChildrenKnown();
			children.put(entry.getName(), entry);
			entry.setParent(this);
		}

		public AbstractEntry createChild(String name, boolean isFolder) {
			IVersionable child;
			if (isFolder) {
				child = (IFolder)IFolder.ITEM_TYPE.createItem();
			} else {
				child = (IFileItem)IFileItem.ITEM_TYPE.createItem();
			}
			child.setName(name);
			child.setParent((IFolderHandle)getItemHandle());
			AbstractEntry entry = createEntry(name, child);
			// If we created a folder, we should initialize it's children to be empty
			if (entry instanceof FolderEntry) {
				FolderEntry folderEntry = (FolderEntry) entry;
				folderEntry.setChildren(null);
				
			}
			add(entry);
			return entry;
		}

		public AbstractEntry createEntry(String name, IVersionableHandle child) {
			if (child instanceof IFolderHandle) {
				FolderEntry folderEntry = new FolderEntry(name, this, child);
				return folderEntry;
			}
			return new FileEntry(name, this, child);
		}

		public Collection<AbstractEntry> getChildren() {
			ensureChildrenKnown();
			return new ArrayList<AbstractEntry>(children.values());
		}

		public boolean areChildrenKnown() {
			return children != null;
		}

		public Collection<String> getChildNames() {
			ensureChildrenKnown();
			return children.keySet();
		}
    }
    
	/**
	 * Tracks the moves made in a set of changes so that the
	 * pre-change before path of the affected files can be determined.
	 * This is required to properly resolve paths when preparing changes
	 * (i.e. to populate the tree based on a set of changes before the
	 * changes are made).
	 */
	private static class PrechangePathMap {
		Map<String, String> pathBackMap = new HashMap<String, String>();
		public String getBeforePath(IImportChange change) {
			String beforePath = findPrechangePath(change.getBeforePath());
			if (beforePath != null && (change.getFlags() & IImportChange.MOVE) > 0) {
				recordPathChange(change, beforePath);
			}
			return beforePath;
		}
		private void recordPathChange(IImportChange change, String beforePath) {
			pathBackMap.put(change.getAfterPath(), beforePath);
		}
		public String findPrechangePath(String beforePath) {
			if (beforePath == null)
				return null;
			// We want to find a match for the longest path possible
			// (i.e. find the parent path with the shortest relative segment)
			String rest = null;
			String path = beforePath;
			for (Entry<String, String> entry : pathBackMap.entrySet()) {
				if (PathUtils.isParentFolder(entry.getKey(), beforePath)) {
					String nextRest = PathUtils.getRelativePath(entry.getKey(), beforePath);
					if (rest == null || nextRest.length() < rest.length()) {
						rest = nextRest;
						path = PathUtils.appendPath(entry.getValue(), rest);
					}
				}
			}
			return path;
		}
	}
    
    public FileTree(IWorkspaceConnection workspace, IComponent component) {
		this.workspace = workspace;
		this.component = component;
	}

	private AbstractEntry getEntry(String path) {
        AbstractEntry lastEntry = root;
        int index = 0;
        String segmentName;
        if (index <= path.length() - 1 && path.charAt(index) == '/')
        	index++;
        String nextSegment = PathUtils.getNextSegment(index, path);
        if (nextSegment != null)
        	index += nextSegment.length();
        while ((segmentName = nextSegment) != null) {
            if (index <= path.length() - 1 && path.charAt(index) == '/')
            	index++;
            nextSegment = PathUtils.getNextSegment(index, path);
            if (nextSegment != null)
            	index += nextSegment.length();
            FolderEntry parent = (FolderEntry)lastEntry;
            if (parent.areChildrenKnown()) {
	            AbstractEntry entry = (AbstractEntry)parent.children.get(segmentName);
	            if (entry == null) {
	                return null;
	            } else if (!entry.isContainer() && nextSegment != null) {
	                return null;
	            }
	            lastEntry = entry;
            } else {
            	return null;
            }
        }
        return lastEntry;
	}
	
	public IVersionableHandle getItemHandle(String path) {
        AbstractEntry entry = getEntry(path);
        if (entry != null) {
        	return entry.getItemHandle();
        }
        return null;
	}
	
	public IVersionable getItem(String path) {
        AbstractEntry entry = getEntry(path);
        if (entry != null) {
        	IVersionable item = entry.getItem();
        	if (item != null) {
        		return item;
        	}
        }
		throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_1, path));

	}
	
	public boolean hasChildren(String parentPath) {
		AbstractEntry entry = getEntry(parentPath);
		if (entry.isContainer()) {
			return ((FolderEntry)entry).hasChildren();
		}
		return false;
	}

	private void internalRemoveEntry(AbstractEntry entry) {
		FolderEntry parent = entry.getParent();
		parent.remove(entry.getName());
	}

	public IVersionable move(String beforePath, String afterPath) {
		AbstractEntry entry = getEntry(beforePath);
		if (entry == null) {
			throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_4, beforePath, afterPath));
		}
		FolderEntry parent = entry.getParent();
		IVersionable item = entry.getItem();
		if (item == null) {
			throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_5, beforePath, afterPath));
		}
		parent.remove(item.getName());
		AbstractEntry newParent = getEntry(PathUtils.getParentFolderPath(afterPath));
		if (newParent == null) {
			throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_6, beforePath, afterPath));
		}
		if (newParent instanceof FolderEntry) {
			FolderEntry np = (FolderEntry) newParent;
			entry.setName(PathUtils.getBaseName(afterPath));
			item.setName(entry.getName());
			item.setParent((IFolderHandle)np.getItemHandle());
			np.add(entry);
			return item;
		}
		throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_7, beforePath, afterPath));
	}
	
	private IVersionable createItem(String path, boolean isFolder) {
		String parentPath = PathUtils.getParentFolderPath(path);
		AbstractEntry entry = getEntry(parentPath);
		if (entry == null || entry.getItemHandle() == null) {
			throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_2, path));
		}
		if (!entry.isContainer()) {
			throw new IllegalStateException(NLS.bind(SCMImportMessages.FileTree_3, path));
		}
		FolderEntry parentEntry = (FolderEntry)entry;
		AbstractEntry child = parentEntry.createChild(PathUtils.getBaseName(path), isFolder);
		return child.getItem();
	}

	public IFileItem createFile(String path) {
		IVersionable item = createItem(path, false);
		return (IFileItem)item;
	}

	public IFolder createFolder(String path) {
		IVersionable item = createItem(path, true);
		return (IFolder)item;
	}

	public IVersionable remove(String path) {
		AbstractEntry entry = getEntry(path);
		if (entry != null) {
			removeEntry(entry);
			return entry.getItem();
		}
		return null;
	}

	private void removeEntry(AbstractEntry entry) {
		if (entry instanceof FolderEntry) {
			FolderEntry folderEntry = (FolderEntry) entry;
			if (folderEntry.areChildrenKnown() && folderEntry.hasChildren()) {
				Collection<AbstractEntry> children = folderEntry.getChildren();
				for (AbstractEntry child : children) {
					removeEntry(child);
				}
			}
		}
		internalRemoveEntry(entry);
	}
	
    public void prepareForChanges(List<IImportChange> changes,
			boolean fetchDeepForAdds, boolean directoriesSignificant, IProgressMonitor monitor) throws TeamRepositoryException {
		try {
			SubMonitor progress = SubMonitor.convert(monitor, changes.size() * 2);
			if (root == null || isAlwaysInitialize()) {
				progress.setWorkRemaining(40 + changes.size() * 2);
				init(progress.newChild(20));
			}
			Map<UUID, AbstractEntry> entriesToUpdate = new HashMap<UUID, AbstractEntry>();
			if (fetchDeepForAdds && !directoriesSignificant) {
				// We need to fetch the entire tree in case there are deletions in folders
				// that have no additions in them
				fetchDescendants((FolderEntry)root, true, entriesToUpdate, progress);
			} else {
				// Either directories are significant or we are not reconciling so we can
				// update the tree for the changes
				PrechangePathMap prechangeMap = new PrechangePathMap();
				for (IImportChange change : changes) {
					updateTree(change, fetchDeepForAdds, prechangeMap, entriesToUpdate, progress.newChild(2));
				}
			}
			fetchItems(entriesToUpdate, progress);
		} finally {
			monitor.done();
		}
	}

	private void fetchItems(Map<UUID, AbstractEntry> entriesToUpdate,
			SubMonitor progress) throws TeamRepositoryException,
			ItemNotFoundException, ComponentNotInWorkspaceException {
		if (entriesToUpdate.size() > 0) {
			List<IVersionableHandle> versionableItems = new ArrayList<IVersionableHandle>();
			for (AbstractEntry entry : entriesToUpdate.values()) {
				// Only update entries that don't already have an item
				if (entry.getItem() == null) {
					versionableItems.add(entry.getItemHandle());
				}
			}
			if (versionableItems.size() > 0) {
				List result = getConfiguration().fetchCompleteItems(versionableItems , progress);
				for (Object object : result) {
					IVersionable v = (IVersionable) object;
					AbstractEntry entry = entriesToUpdate.get(v.getItemId());
					entry.setItem(v);
					entry.prepareForChange();
				}
			}
		}
	}

	private void init(SubMonitor progress) throws TeamRepositoryException {
		progress.setWorkRemaining(100);
		IFolder root = (IFolder)getConfiguration().completeRootFolder(progress.newChild(50));
        if (root == null) {
            throw new TeamRepositoryException(NLS.bind(SCMImportMessages.ChangeSetArchiveImporter_10, new Object[] { workspace.getName(), component.getName()}));
        }
        this.root = new FolderEntry("", null, (IFolder)root.getWorkingCopy()); //$NON-NLS-1$
        Map<UUID, AbstractEntry> entriesToUpdate = new HashMap<UUID, AbstractEntry>();
        fetchChildren(this.root, entriesToUpdate, progress.newChild(30));
        fetchItems(entriesToUpdate, progress.newChild(20));
	}

	private IConfiguration getConfiguration() throws ItemNotFoundException,
			ComponentNotInWorkspaceException {
		return workspace.configuration(component);
	}
	
	@SuppressWarnings("unchecked")
	private void fetchChildren(FolderEntry folderEntry, Map<UUID, AbstractEntry> entriesToUpdate,
			SubMonitor progress) throws TeamRepositoryException {
		Map<String, IVersionableHandle> handlesMap = getConfiguration().childEntries(
				(IFolderHandle)folderEntry.getItemHandle(), 
				progress);
		folderEntry.setChildren(handlesMap);
		Collection<AbstractEntry> children = folderEntry.getChildren();
		for (AbstractEntry entry : children) {
    		entriesToUpdate.put(entry.getItemHandle().getItemId(), entry);
    	}
	}
	
	private void updateTree(IImportChange change, boolean fetchDeepForAdds, PrechangePathMap prechangeMap, Map<UUID, AbstractEntry> entriesToUpdate, SubMonitor progress) throws TeamRepositoryException {
		String beforePath = prechangeMap.getBeforePath(change);
		String afterPath = change.getAfterPath();
		if (change.getItemType() == IImportChange.FOLDER) {
	        if (beforePath != null) {
	        	resolveFolder(beforePath, true, entriesToUpdate, progress.newChild(1));
	        }
	        if (afterPath != null) {
	        	AbstractEntry entry = resolveFolder(afterPath, false, entriesToUpdate, progress.newChild(1));
	        	if (entry != null) {
	        		if (fetchDeepForAdds
	        				&& entry instanceof FolderEntry
	        				&& (change.getKind() == IImportChange.ADDITION
	        						|| change.getKind() == ImportChange.MODIFICATION 
	        							&& (change.getFlags() & IImportChange.MOVE) != 0)) {
	        			// We are reconciling additions with the pre-existing state of the component
	        			fetchDescendants((FolderEntry)entry, false, entriesToUpdate, progress);
	        		}
	        	}
	        }
		} else if (change.getItemType() == IImportChange.FILE) {
		    if (beforePath != null) {
		    	resolveFile(beforePath, true, entriesToUpdate, progress.newChild(1));
		    }
		    if (afterPath != null) {
		    	resolveFile(afterPath, false, entriesToUpdate, progress.newChild(1));
		    }
		} else if (change.getItemType() == IImportChange.UNKNOWN) {
	        if (beforePath != null) {
	        	resolvePath(beforePath, true, entriesToUpdate, progress.newChild(1));
	        }
	        if (afterPath != null) {
	        	resolvePath(afterPath, false, entriesToUpdate, progress.newChild(1));
	        }
		}
		// For additions, we need to make sure the parent's children are fetched
		// using the parent's before state
		if (beforePath == null && change.getKind() == IImportChange.ADDITION) {
			beforePath = prechangeMap.findPrechangePath(afterPath);
        	resolvePath(beforePath, false, entriesToUpdate, progress.newChild(1));
		}
	}
	
    private void fetchDescendants(FolderEntry entry, boolean alwaysRecurse, Map<UUID, AbstractEntry> entriesToUpdate, SubMonitor progress) throws TeamRepositoryException {
    	boolean recurse = alwaysRecurse;
		if (!entry.areChildrenKnown()) {
			fetchChildren(entry, entriesToUpdate, progress.newChild(1));
			recurse = true;
		}
		if (recurse) {
			Collection<AbstractEntry> children = entry.getChildren();
			for (AbstractEntry childEntry : children) {
				if (childEntry instanceof FolderEntry) {
					FolderEntry childFolder = (FolderEntry) childEntry;
					fetchDescendants(childFolder, alwaysRecurse, entriesToUpdate, progress.newChild(1));
				}
			}
		}
	}

	private AbstractEntry resolveFile(String path, boolean errorOnFailure, Map<UUID, AbstractEntry> entriesToUpdate,
			IProgressMonitor monitor) throws TeamRepositoryException {
    	AbstractEntry entry = resolvePath(path, errorOnFailure, entriesToUpdate, monitor);
		if (entry instanceof FileEntry)
			return (FileEntry) entry;
		if (errorOnFailure) {
			throw new IllegalStateException(NLS.bind(
					SCMImportMessages.ChangeSetArchiveImporter_21, path));
		}
		return null;
	}

	private AbstractEntry resolveFolder(String path, boolean errorOnFailure, Map<UUID, AbstractEntry> entriesToUpdate,
			IProgressMonitor monitor) throws TeamRepositoryException {
		AbstractEntry entry = resolvePath(path, errorOnFailure, entriesToUpdate, monitor);
		if (!(entry instanceof FolderEntry) && errorOnFailure) {
			throw new IllegalStateException(NLS.bind(
					SCMImportMessages.ChangeSetArchiveImporter_22, path));
		}
		return entry;
	}

	private AbstractEntry resolvePath(String path, boolean errorOnFailure, Map<UUID, AbstractEntry> entriesToUpdate,
			IProgressMonitor monitor) throws TeamRepositoryException {
		SubMonitor progress = SubMonitor.convert(monitor, 100);
		AbstractEntry entry = getEntry(path);
		if (entry == null || entry.getItemHandle() == null) {
			String name = PathUtils.getBaseName(path);
			String parentPath = PathUtils.getParentFolderPath(path);
			AbstractEntry parent = resolveFolder(parentPath, errorOnFailure, entriesToUpdate, progress.newChild(50));

			if (parent == null || !(parent instanceof FolderEntry)) {
				// can happen if errorOnFailure is false and we failed to
				// resolve
				return null;
			}
			FolderEntry parentFolder = (FolderEntry)parent;
			if (!parentFolder.areChildrenKnown()) {
				fetchChildren(parentFolder, entriesToUpdate, progress.newChild(20));
			}
			
			AbstractEntry childEntry = parentFolder.getChild(name);
			if (childEntry == null) {
				if (errorOnFailure) {
					throw new IllegalStateException(NLS.bind(
							SCMImportMessages.ChangeSetArchiveImporter_23,
							path));
				} else {
					return null;
				}
			}
			entry = childEntry;
		}

		monitor.done();

		return entry;
	}
	
    public boolean isAlwaysInitialize() {
		return alwaysInitialize;
	}

	public void setAlwaysInitialize(boolean alwaysInitialize) {
		this.alwaysInitialize = alwaysInitialize;
	}

	public Collection<String> getChildNames(String path) {
		AbstractEntry entry = getEntry(path);
		if (entry.isContainer()) {
			return ((FolderEntry)entry).getChildNames();
		}
		return Collections.emptyList();
	}

	public boolean areChilrenKnown(String path) {
		AbstractEntry entry = getEntry(path);
		if (entry instanceof FolderEntry) {
			FolderEntry folder = (FolderEntry) entry;
			return folder.areChildrenKnown();
		}
		return false;
	}
	
}
