/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.utils;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.rcp.ui.internal.util.PathUtils;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.importz.IImportChange;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Tree that wraps a file tree and tracks empty folders so that repositories
 * that don;t track folders (e.g. CVS) can auto add and remove folders
 * as required.
 */
public class FileTreeWithEmptyFolderTracking implements IFileTree {
	
	private IFileTree wrappedTree;
    private final Set<String> emptyFolders;
    
	public FileTreeWithEmptyFolderTracking(IFileTree tree) {
        emptyFolders = new HashSet<String>();
        wrappedTree = tree;
	}
	
    public void prepareForChanges(List<IImportChange> changes,
			boolean fetchDeep, boolean directoriesSignificant, IProgressMonitor monitor) throws TeamRepositoryException {
		wrappedTree.prepareForChanges(changes, fetchDeep, directoriesSignificant, monitor);
	}

	public IVersionableHandle getItemHandle(String path) {
		return wrappedTree.getItemHandle(path);
	}
	
	public IVersionable getItem(String path) {
		return wrappedTree.getItem(path);
	}
	
	public IFolder createFolder(String path) {
		IFolder folder = wrappedTree.createFolder(path);
		emptyFolders.remove(PathUtils.getParentFolderPath(path));
		emptyFolders.add(path);
		return folder;
	}
	
	public IFileItem createFile(String path) {
		IFileItem file = wrappedTree.createFile(path);
		emptyFolders.remove(PathUtils.getParentFolderPath(path));
		return file;
	}
	
	public IVersionable move(String beforePath, String afterPath) {
        IVersionable item = wrappedTree.move(beforePath, afterPath);
        emptyFolders.remove(PathUtils.getParentFolderPath(afterPath));
        testForEmptyFolder(beforePath);
		return item;
	}

	private void testForEmptyFolder(String beforePath) {
		String parent = PathUtils.getParentFolderPath(beforePath);
		// Only record empty folders if they are not the root folder
		if (!wrappedTree.hasChildren(parent)&& PathUtils.getBaseName(parent).length() > 0) {
        	emptyFolders.add(parent);
        }
	}
	
	/**
	 * Remove the versionable at the given path. The remove will only succeed
	 * if the versionable at the path is a file or a folder with no children.
	 * When removing folders that contain children, the children must be recursively
	 * removed before the folder can be removed.
	 * @param path the path of the versionable to be removed
	 */
	public IVersionable remove(String path) {
		IVersionable item = wrappedTree.remove(path);
		emptyFolders.remove(path);
		testForEmptyFolder(path);
		return item;
	}
	
	public boolean hasChildren(String path) {
		return wrappedTree.hasChildren(path);
	}
	
	public Set<String> getEmptyFolders() {
		// Return a copy so clients cannot modify the list directly
		return new HashSet<String>(emptyFolders);
	}

	public Collection<String> getChildNames(String path) {
		return wrappedTree.getChildNames(path);
	}

	public boolean areChilrenKnown(String path) {
		return wrappedTree.areChilrenKnown(path);
	}

}