/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.scm.client.importz.internal.utils;

import java.util.Collection;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.importz.IImportChange;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Interface that represents the structure of a repository workspace.
 * This tree only provides the structure of the files and folders
 * but does not manage file contents. The tree is populated by calling
 * {@link #prepareForChanges(List, boolean, boolean, IProgressMonitor)} and only items and
 * ancestor items for the before and after paths of changes will be available
 * in the tree.
 */
public interface IFileTree {
	
	/**
	 * Return the item handle at the given path or <code>null</code>
	 * if there isn't an item handle at the given location in the tree.
	 * Note that it is only guaranteed that handles will be available
	 * for paths that appear in changes passed to {@link #prepareForChanges(List, boolean, boolean, IProgressMonitor)}.
	 * If a <code>null</code> is returned, it is possible that the item doesn't exist
	 * or that the item was not a item or ancestor item of a change path.
	 * @param path the path
	 * @return the handle at the given path or <code>null</code>
	 */
	IVersionableHandle getItemHandle(String path);
	
	/**
	 * Return the item at the given path. This method will throw
	 * an {@link IllegalStateException} if there an item doesn't exist
	 * at the given path or the item was not involved in a change
	 * that was passed to {@link #prepareForChanges(List, boolean, boolean, IProgressMonitor)}
	 * before the call to this method.
	 * @param path the path
	 * @return the item at the given path
	 */
	IVersionable getItem(String path);

	/**
	 * Create a folder at the given path. The parent of the
	 * path must exist.
	 * @param path the path
	 * @return the created folder
	 */
	IFolder createFolder(String path);

	/**
	 * Create a file at the given path. The parent of the
	 * path must exist.
	 * @param path the path
	 * @return the created file
	 */
	IFileItem createFile(String path);

	/**
	 * Move the item at the <code>beforePath</code> to the <code>afterPath</code>.
	 * This method will throw
	 * an {@link IllegalStateException} if there an item doesn't exist
	 * at the given path or the item was not involved in a change
	 * that was passed to {@link #prepareForChanges(List, boolean, boolean, IProgressMonitor)}
	 * before the call to this method.
	 * 
	 * @param beforePath the path of the item to be moved
	 * @param afterPath the path of the item after the move
	 * @return the item that was moved
	 */
	IVersionable move(String beforePath, String afterPath);
	
	/**
	 * Remove the item at the given path. This method will throw
	 * an {@link IllegalStateException} if the item was not involved in a change
	 * that was passed to {@link #prepareForChanges(List, boolean, boolean, IProgressMonitor)}
	 * before the call to this method.
	 * @param path the path of the item to be removed
	 * @return the item that was removed or <code>null</code> if there is no item
	 * 		at the given path
	 */
	IVersionable remove(String path);

	/**
	 * Return whether the item at the given path has children
	 * @param path the path
	 * @return whether the item at the given path has children
	 */
	boolean hasChildren(String path);

	/**
	 * Prepare the tree for the given changes to be performed on it.
	 * @param changes the changes
	 * @param fetchDeepForAdditions indicates that we are reconciling with pre-existing content so we 
	 * 		we need to fetch deeply any subtree that overlaps with an add.
	 * @param directoriesSignificant TODO
	 * @param monitor a progress monitor
	 * @throws TeamRepositoryException
	 */
	void prepareForChanges(List<IImportChange> changes, boolean fetchDeepForAdditions, boolean directoriesSignificant, IProgressMonitor monitor) throws TeamRepositoryException;

	/**
	 * Return the list of names of the children for the given path.
	 * @param path the path
	 * @return the list of names of the children for the given path
	 */
	Collection<String> getChildNames(String path);

	boolean areChilrenKnown(String path);

}
