/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.spi;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Date;

import com.ibm.team.internal.repository.rcp.tar.TarEntry;
import com.ibm.team.internal.repository.rcp.tar.TarOutputStream;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.importz.IChangeSetArchiveCreator;
import com.ibm.team.scm.client.importz.IImportChangeSet;
import com.ibm.team.scm.client.importz.IMigrationFactory;
import com.ibm.team.scm.client.importz.internal.utils.DebugUtils;

/**
 * Abstract implementation of {@link IChangeSetArchiveCreator} that can be subclassed
 * by clients.
 * 
 * @since 0.6.1
 */
public abstract class ChangeSetArchiveCreator implements IChangeSetArchiveCreator {

	private TarOutputStream tarStream;
	private File tempLogFile;
	private ChangeSetFileWriter writer;
	private final IMigrationFactory migrationFactory;

	public ChangeSetArchiveCreator(IMigrationFactory migrationFactory) {
		this.migrationFactory = migrationFactory;
	}
	
	/**
	 * Return the migration factory associated with this object.
	 * @return the migration factory
	 */
	public IMigrationFactory getMigrationFactory() {
		return migrationFactory;
	}
	
	/**
	 * Opens the archive file for writing. This method must be called before
	 * either of the archive content methods or the {@link #writeChangeSet(IImportChangeSet)}
	 * method. Callers of this method must call {@link #closeArchive()} when finished.
	 * @param file the archive file
	 * @throws IOException
	 * @throws TeamRepositoryException 
	 */
	protected void openArchive(File file) throws IOException {
		try {
			file.createNewFile();
			tarStream= new TarOutputStream(file);
		} finally {
            if (tarStream == null) {
                file.delete();
            }
		}
	}
	
	/**
	 * Closes the archive after it has been populated. This should always be called
	 * if {@link #openArchive(File)} was called and returned successfully.
	 * @throws TeamRepositoryException 
	 */
	protected void closeArchive() throws IOException {
		try {
			archiveLogFile();
		} finally {
			try {
				if (tarStream != null)
					tarStream.close();
			} catch (IOException e) {
				// Ignore
			}
		}
	}
	
	/**
	 * This method must be called to initialize the change set file writer before 
	 * @throws TeamRepositoryException
	 */
	private void initializeChangeSetFileWriter() throws IOException {
		tempLogFile = DebugUtils.createTempFile(ChangeSetFileWriter.LOG_FILENAME);
		writer = migrationFactory.createChangeSetFileWriter(tempLogFile);
	}
	
	/** 
	 * Write the contents of the given temp file into the archive at the given path.
	 * @param archivePath the archive path for the file
	 * @param file the file whose contents are to be written to the archive
	 * @param date the date to be associated with the file in the archive
	 * @throws IOException
	 */
	protected void archiveContents(String archivePath, File file, Date date) throws IOException {
		FileInputStream stream = new FileInputStream(file);
		try {
			archiveContents(archivePath, stream, file.length(), date);
		} finally {
			try {
				stream.close();
			} catch (IOException e) {
				// Ignore
			}
		}
	}
	
	/** 
	 * Write the contents of the stream into the archive at the given path.
	 * This method will close the given stream
	 * @param archivePath the archive path for the file
	 * @param stream the stream with the contents to be added to the archive
	 * @param length the number of bytes in the stream
	 * @param date the date to be associated with the file in the archive
	 * @throws IOException
	 */
	protected void archiveContents(String archivePath, InputStream stream, long length, Date date) throws IOException {
		InputStream contents = null;
		try {
			if (archivePath != null) {
				// We care about this content so we'll put it in the archive
				TarEntry entry = getTarEntry(archivePath, date.getTime(), length);
				contents = new BufferedInputStream(stream);
				addTarEntry(entry, contents);
			}
		} finally {
			if (contents != null) {
				try {
					contents.close();
				} catch (IOException e) {
					// Ignore
				}
			} else {
				try {
					stream.close();
				} catch (IOException e) {
					// Ignore
				}
			}
		}
	}
	
	private TarEntry getTarEntry(String path, long timestamp, long length) {
		long seconds = timestamp / 1000;
		return new TarEntry(path, 0644, 0, 0, length, seconds, TarEntry.FILE_TYPE,
            "", "", "", 0, 0, seconds, seconds); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}
	
	private void addTarEntry(TarEntry entry, InputStream contents) throws IOException {
		tarStream.putNextEntry(entry, contents, TarOutputStream.gzipFilter);
	}
	
	private void archiveLogFile() throws IOException {
		if (tempLogFile == null) {
			initializeChangeSetFileWriter();
		}
		if (tempLogFile != null) {
			writer.closeWrite();
			archiveContents(ChangeSetFileWriter.LOG_FILENAME, tempLogFile, new Date());
			DebugUtils.deleteTempFile(tempLogFile);
		}
	}
	
	/**
	 * Write the given change set to the changes file that will be written to the
	 * archive when it is closed.
	 * @param changeSet
	 * @throws IOException
	 */
	protected void writeChangeSet(IImportChangeSet changeSet) throws IOException {
		if (tempLogFile == null) {
			initializeChangeSetFileWriter();
		}
		writer.writeElement(changeSet);
	}
	
}
