/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.scm.client.importz.svn.internal;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class CopyFromPreviousSubtrees {

	private Map<Long, SubtreeDescriptor> subtreesOfInterest = new HashMap<Long, SubtreeDescriptor>();

	private static class SubtreeDescriptor {

		private final long revision;
		private final Set<String> paths;
		private long lastReferenceRevision;

		public SubtreeDescriptor(long copyFromRevision, long lastAccessRevision) {
			this.revision = copyFromRevision;
			lastReferenceRevision = lastAccessRevision;
			this.paths = new HashSet<String>();
		}

		public long getRevision() {
			return revision;
		}

		public Collection<String> getPaths() {
			return paths;
		}

		public void add(String copyFromPath) {
			paths.add(copyFromPath);
		}

		public long getLastReferenceRevision() {
			return lastReferenceRevision;
		}

		public void adjustLastReferenceRevision(long lastReferenceRevision) {
			if (lastReferenceRevision > this.lastReferenceRevision)
				this.lastReferenceRevision = lastReferenceRevision;
		}
		
		public void write(PrintStream stream) {
			stream.println(revision);
			stream.println(lastReferenceRevision);
			stream.println(paths.size());
			for (String path : paths) {
				stream.println(path);
			}
		}
		
	}
	
	public static CopyFromPreviousSubtrees loadSubtreesOfInterest(SVNDumpFileImportData importData) throws IOException {
		CopyFromPreviousSubtrees copies = new CopyFromPreviousSubtrees();
		if (importData.hasArchivedContentFor(SVNDump2ArchiverConverter.SVN_SUBTREES_OF_INTEREST_FILENAME)) {
			InputStream in = importData.getArchiveInputStream(SVNDump2ArchiverConverter.SVN_SUBTREES_OF_INTEREST_FILENAME);
			try {
				BufferedReader reader = new BufferedReader(new InputStreamReader(in));
				if (Activator.DEBUG_COPIES) {
					System.out.println("===Starting to read subtrees"); //$NON-NLS-1$
				}
				String line = reader.readLine();
				if (line != null) {
					int count = Integer.parseInt(line);
					for (int i = 0; i < count; i++) {
						SubtreeDescriptor desc = loadDescriptor(reader);
						copies.add(desc);
					}
				}
				if (Activator.DEBUG_COPIES) {
					System.out.println("===Subtrees read"); //$NON-NLS-1$
				}
			} finally {
				try {
					in.close();
				} catch (IOException e) {
					// Ignore
				}
			}
		}
		copies.updateForBaseline(importData);
		return copies;
	}

	private static SubtreeDescriptor loadDescriptor(BufferedReader reader) throws IOException {
		long revision = Long.parseLong(reader.readLine());
		long lastAccessRevision = Long.parseLong(reader.readLine());
		SubtreeDescriptor desc = new SubtreeDescriptor(revision, lastAccessRevision);
		int count = Integer.parseInt(reader.readLine());
		for (int i = 0; i < count; i++) {
			desc.add(reader.readLine());
		}
		return desc;
	}
	
	private void updateForBaseline(SVNDumpFileImportData importData) {
		long baseRevision = importData.getBaseRevision();
		if (importData.getStartingBaseline() != null && baseRevision > 0) {
			Set<String> paths = importData.getUnbaselinedCopies();
			SubtreeDescriptor desc = subtreesOfInterest.get(baseRevision);
			long lastReferenceRevision = getLastReferenceRevision(importData.getTargetTags());
			if (desc == null) {
				desc = new SubtreeDescriptor(baseRevision, lastReferenceRevision);
				add(desc);
			} else {
				desc.adjustLastReferenceRevision(lastReferenceRevision);
			}
			for (String path : paths) {
				desc.add(path);
			}
		}
	}

	private long getLastReferenceRevision(Collection<SVNBranchOrTag> targetTags) {
		long lastReferenceRevision = 0;
		for (SVNBranchOrTag bot : targetTags) {
			long rev = bot.getLastAccessRevision();
			if (rev > lastReferenceRevision)
				lastReferenceRevision = rev;
		}
		return lastReferenceRevision;
	}

	private void add(SubtreeDescriptor desc) {
		subtreesOfInterest.put(desc.getRevision(), desc);
	}

	public Collection<String> getPaths(long revision) {
		SubtreeDescriptor desc = subtreesOfInterest.get(revision);
		if (desc != null) {
			return desc.getPaths();
		}
		return Collections.emptyList();
	}

	public void write(PrintStream stream) {
		Set<Long> keySet = subtreesOfInterest.keySet();
		List<Long> list = new ArrayList<Long>();
		list.addAll(keySet);
		Collections.sort(list);
		stream.println(keySet.size());
		for (Long integer : list) {
			SubtreeDescriptor subtree = subtreesOfInterest.get(integer);
			subtree.write(stream);
		}
	}

	public boolean isEmpty() {
		return subtreesOfInterest.isEmpty();
	}

	public void recordCopy(long revision, String copyFromPath, long copyFromRevision) {
		// How do we use the revision to determine when we can purge the subtrees
		SubtreeDescriptor subtreeDescriptor = subtreesOfInterest.get(copyFromRevision);
		if (subtreeDescriptor == null) {
			subtreeDescriptor = new SubtreeDescriptor(copyFromRevision, revision);
			subtreesOfInterest.put(copyFromRevision, subtreeDescriptor);
		} else {
			subtreeDescriptor.adjustLastReferenceRevision(revision);
		}
		subtreeDescriptor.add(copyFromPath);
	}

	public long getLastReferenceRevision(long revision) {
		SubtreeDescriptor desc = subtreesOfInterest.get(revision);
		if (desc != null)
			return desc.getLastReferenceRevision();
		return -1L;
	}
}
